#include "ace/Log_Msg.h"
#include "ace/Service_Config.h"
#include "ace/Name_Options.h"
#include "ace/Local_Name_Space.h"

const char ACE_Local_Name_Space::NAME_SERVER_MAP[] = "Name Server Map";

wchar_t *
ACE_NS_String::fast_rep (void) const
{
  return this->rep_;
}

ACE_NS_String::operator ACE_WString () const
{
  return ACE_WString (this->rep_, (this->len_ / sizeof (wchar_t)) - 1);
}

char *
ACE_NS_String::char_rep (void) const
{
  ACE_WString w_string = ACE_WString (this->rep_, (this->len_ / sizeof (wchar_t)) - 1);
  return w_string.char_rep ();
}

ACE_NS_String::ACE_NS_String (void)
  : rep_ (0), len_ (0)
{
}

ACE_NS_String::ACE_NS_String (const ACE_WString &s)
  : rep_ (s.fast_rep ()), len_ ((s.length () + 1) * sizeof (wchar_t))
{
}

int
ACE_NS_String::operator == (const ACE_NS_String &s) const
{
  return this->len_ == s.len_
    && ACE_OS::memcmp ((void *) this->rep_, 
		       (void *) s.rep_, this->len_) == 0;
}

ACE_NS_String::ACE_NS_String (wchar_t *dst, 
			      const wchar_t *src, 
			      size_t bytes)
  : len_ (bytes), rep_ (dst)
{
  ACE_OS::memcpy (this->rep_, src, bytes);
}

int 
ACE_Local_Name_Space::shared_bind (const ACE_WString &name,
				   const ACE_WString &value, 
				   const char *type,
				   int rebind)
{
  // @ Note, we need to fix up the <type> stuff in here, as well...
  size_t name_len = (name.length () + 1) * sizeof (wchar_t);
  size_t value_len = (value.length () + 1) * sizeof (wchar_t);
  size_t total_len = name_len + value_len;
  char *ptr = (char *) this->allocator_->malloc (total_len);

  if (ptr == 0)
    return -1;
  else
    {
      // Note that the value_rep *must* come first to make sure we can
      // retrieve this pointer later on in unbind().
      wchar_t *value_rep = (wchar_t *) (ptr);
      wchar_t *name_rep = (wchar_t *) (ptr + value_len);
      ACE_NS_String new_name (name_rep, name.fast_rep (), name_len);
      ACE_NS_String new_value (value_rep, value.fast_rep (), value_len);
      int result = -1;

      if (rebind == 0)
        // Do a normal bind.  This will fail if there's already an
	// <new_name> with the same name.
	result = this->name_space_map_->bind (new_name, new_value);
      else
	{
	  // Do a rebind.  If there's already any entry, this will
	  // return the existing <new_name> and <new_value> and overwrite
	  // the existing name and value.
	  ACE_NS_String old_name, old_value;

	  result = this->name_space_map_->rebind (new_name, new_value,
						  old_name, old_value);

	  if (result == 1)
	    // Free up the memory we allocated in shared_bind().  Note that
	    // this assumes that the "value" pointer comes first and that
	    // the value and name are contiguously allocated (see
	    // above for details)
	    this->allocator_->free ((void *) old_value.fast_rep ());
	}

      if (result == -1)
	// Free our dynamically allocated memory.
	this->allocator_->free ((void *) ptr);
      else
	// If bind() or rebind() succeed, they will automatically sync
	// up the map manager entry.  However, we must sync up our
	// name/value memory. 
	this->allocator_->sync (ptr, total_len);

      return result;
    }
}

int  
ACE_Local_Name_Space::unbind (const ACE_WString &name)
{
  ACE_NS_String ns_name (name);
  ACE_NS_String ns_value;

  if (this->name_space_map_->unbind (ns_name, ns_value) != 0)
    return -1;
  else
    {
      // Free up the memory we allocated in shared_bind().  Note that
      // this assumes that the "value" pointer comes first and that
      // the value and name are contiguously allocated (see
      // shared_bind() for details)
      this->allocator_->free ((void *) ns_value.fast_rep ());
      return 0;
    }
}

int 
ACE_Local_Name_Space::bind (const ACE_WString &name,
			    const ACE_WString &value, 
			    const char *type)
{
  return this->shared_bind (name, value, type, 0);
}

int 
ACE_Local_Name_Space::rebind (const ACE_WString &name, 
			      const ACE_WString &value, 
			      const char *type)
{
  return this->shared_bind (name, value, type, 1);
}

int 
ACE_Local_Name_Space::resolve (const ACE_WString &name,
			       ACE_WString &value, 
			       char *&type)
{
  ACE_NS_String ns_value;
  ACE_NS_String ns_name (name);

  if (this->name_space_map_->find (ns_name, ns_value) != 0)
    return -1;
  else
    {
      // Calls conversion operator and then calls the ACE_WString
      // assignment operator to get a fresh copy.
      value = ns_value; 
      return 0;
    }
}

int
ACE_Local_Name_Space::open (ACE_Naming_Context::Context_Scope_Type scope_in) 
{
  this->ns_scope_ = scope_in;

  return this->create_manager ();
}

ACE_Local_Name_Space::ACE_Local_Name_Space (void)
{
}

ACE_Local_Name_Space::ACE_Local_Name_Space (ACE_Naming_Context::Context_Scope_Type scope_in)
{  
  if (this->open (scope_in) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n", "ACE_Local_Name_Space::ACE_Local_Name_Space"));
}

ACE_Local_Name_Space::~ACE_Local_Name_Space ()
{
}

int 
ACE_Local_Name_Space::create_manager (void)
{
  // get directory name
  char *dir = ACE_Name_Options::instance ()->namespace_dir ();

  switch (this->ns_scope_)
    {
    case ACE_Naming_Context::PROC_LOCAL:
      {
	// Use process name as the file name.
	size_t len = ACE_OS::strlen (dir);
	len += ACE_OS::strlen (ACE_Name_Options::instance ()->process_name ()) + 1;

	if (len >= MAXNAMELEN)
	  {
	    errno = ENAMETOOLONG;
	    return -1;
	  }

        ACE_OS::strcpy (this->context_file_, dir);
        ACE_OS::strcat (this->context_file_,"/");
        ACE_OS::strcat (this->context_file_, ACE_Name_Options::instance ()->process_name ());
	break;
       }
      
    case ACE_Naming_Context::NODE_LOCAL:
      {
	// use default local file name
	size_t len = ACE_OS::strlen (dir);
	len += ACE_OS::strlen (ACE_DEFAULT_LOCALNAME) + 1;

	if (len >= MAXNAMELEN)
	  {
	    errno = ENAMETOOLONG;
	    return -1;
	  }

        ACE_OS::strcpy (this->context_file_, dir);
        ACE_OS::strcat (this->context_file_, ACE_DEFAULT_LOCALNAME);
	break;
      }
    case ACE_Naming_Context::NET_LOCAL:
      {
	// use default global file name
	size_t len = ACE_OS::strlen (dir);
	len += ACE_OS::strlen (ACE_DEFAULT_GLOBALNAME) + 1;
	
	if (len >= MAXNAMELEN)
	  {
	    errno = ENAMETOOLONG;
	    return -1;
	  }

        ACE_OS::strcpy (this->context_file_, dir);
        ACE_OS::strcat (this->context_file_, ACE_DEFAULT_GLOBALNAME);
	break;
      }
    }
  
  ACE_DEBUG ((LM_DEBUG, "contextfile is %s\n", this->context_file_));

  this->allocator_ = new ALLOCATOR (this->context_file_);
  
  void *ns_map = 0;

  // This is the easy case since if we find the Name Server Map
  // Manager we know it's already initialized.
  if (this->allocator_->find (ACE_Local_Name_Space::NAME_SERVER_MAP, ns_map) == 0)
    {
      this->name_space_map_ = (MAP_MANAGER *) ns_map;
      ACE_DEBUG ((LM_DEBUG, "name_space_map_ = %d, ns_map = %d\n",
		  this->name_space_map_, ns_map));
    }

  // This is the hard part since we have to avoid potential race
  // conditions...
  else
    {
      ns_map = this->allocator_->malloc (sizeof (MAP_MANAGER));
      // Initialize the map into its memory location (e.g., shared memory).
      this->name_space_map_ = new (ns_map) MAP_MANAGER (this->allocator_);

      // Don't allow duplicates (atomically return existing int_id, if
      // there is one).
      if (this->allocator_->trybind (ACE_Local_Name_Space::NAME_SERVER_MAP, ns_map) == 1)
	{
	  // We're not the first one in, so free up the map and assign
	  // the map to the pointer that was allocated by the caller
	  // that was the first time in!
	  this->name_space_map_->close ();
	  // Note that we can't free <map> since that was overwritten
	  // in the call to bind()!
	  this->allocator_->free ((void *) this->name_space_map_);
	  this->name_space_map_ = (MAP_MANAGER *) ns_map;
	}
      ACE_DEBUG ((LM_DEBUG, "name_space_map_ = %d, ns_map = %d\n",
		  this->name_space_map_, ns_map));
    }

  return 0;
}

int 
ACE_Local_Name_Space::list_names (ACE_WSTRING_SET &set,
				  const ACE_WString &pattern)
{
  // Need to fix up...
  return 0;
}

void
ACE_Local_Name_Space::dump (void) const
{
  MAP_ITERATOR map_iterator(*name_space_map_);
  MAP_ENTRY *map_entry;
  
  // Dump out all the entries in the map.

  for (map_entry = 0;
       map_iterator.next (map_entry) != 0;
       map_iterator.advance())
    {
      ACE_DEBUG ((LM_DEBUG, "key=%s\nvalue=%s\ntype=\n",
		  map_entry->ext_id_.char_rep(), 
		  map_entry->int_id_.char_rep()));
      // Note: currently <type> is not implemented so the above
      // call simply adds an empty line for <type>
    }
}


#if defined (ACE_TEMPLATES_REQUIRE_SPECIALIZATION)
template class ACE_Unbounded_Set<ACE_WString>;
template class ACE_Unbounded_Set_Iterator<ACE_WString>;
template class ACE_Set_Node<ACE_WString>;
template class ACE_Map_Entry<ACE_NS_String, ACE_NS_String>;
#if defined(ACE_HAS_THREADS)
template class ACE_Map_Manager<ACE_NS_String, ACE_NS_String, ACE_RW_Process_Mutex>;
template class ACE_Allocator_Adapter<ACE_Malloc<ACE_MMAP_Memory_Pool, ACE_RW_Process_Mutex> 
>;
template class ACE_Malloc<ACE_MMAP_Memory_Pool, ACE_RW_Process_Mutex>;
template class ACE_Guard<ACE_RW_Process_Mutex>;
template class ACE_Write_Guard<ACE_RW_Process_Mutex>;
template class ACE_Read_Guard<ACE_RW_Process_Mutex>;
#else /* no ACE_HAS_THREADS */
template class ACE_Guard<ACE_SV_Semaphore_Complex>;
template class ACE_Read_Guard<ACE_SV_Semaphore_Complex>;
template class ACE_Write_Guard<ACE_SV_Semaphore_Complex>;
template class ACE_Malloc<ACE_MMAP_Memory_Pool, ACE_SV_Semaphore_Complex>;
template class ACE_Map_Manager<ACE_NS_String, ACE_NS_String, ACE_SV_Semaphore_Complex>;
template class ACE_Allocator_Adapter<ACE_Malloc<ACE_MMAP_Memory_Pool, 
ACE_SV_Semaphore_Complex> >;
#endif /* ACE_HAS_THREADS */
#endif /* ACE_TEMPLATES_REQUIRE_SPECIALIZATION */
