/*-*- C++ -*- */

// ============================================================================
//
// = LIBRARY
//    ACE
// 
// = FILENAME
//    ACE_Local_Name_Space
//
// = AUTHOR
//    Prashant Jain (pjain@cs.wustl.edu) and Irfan Pyarali (irfan@wuerl.wustl.edu)
// 
// ============================================================================

#if !defined (ACE_LOCAL_NAME_SPACE_H)
#define ACE_LOCAL_NAME_SPACE_H

#include "ace/ACE.h"
#include "ace/Set.h"
#include "ace/Name_Space.h"
#include "ace/Naming_Context.h"
#include "ace/Name_Proxy.h"
#include "ace/Malloc.h"
#include "ace/Map_Manager.h"
#include "ace/SString.h"

#if defined (ACE_HAS_THREADS)
typedef ACE_RW_Process_Mutex PROCESS_MUTEX;
#else
typedef ACE_SV_Semaphore_Complex PROCESS_MUTEX;
#endif /* ACE_HAS_THREADS */

typedef ACE_Unbounded_Set<ACE_WString> ACE_WSTRING_SET;

class ACE_NS_String
  // = TITLE
  //     This class is used as an Adapter to work with the Map_Manager.
  //
  // = DESCRIPTION
  //     In order to work correctly, this class must be able to
  //     convert back and forth with ACE_WStrings.
{
public:
  // = Initialization.
  ACE_NS_String (void);
  // Default "no-op" constructor.

  ACE_NS_String (wchar_t *dst, const wchar_t *src, size_t len);
  // Initialization method.

  ACE_NS_String (const ACE_WString &);
  // Converts an ACE_WString to an ACE_NS_String;

  operator ACE_WString () const;
  // Converts an ACE_NS_String to fresh copy of an ACE_WString;

  char *char_rep (void) const;
  // Return the ASCII character representation.

  int operator == (const ACE_NS_String &) const;
  // Compare an ACE_NS_String.

  wchar_t *fast_rep (void) const;
  // Returns the underlying representation.

private:
  size_t len_;
  // Length of the string.

  wchar_t *rep_;
  // This actually points into shared/persistent memory.
};

class ACE_Local_Name_Space : public ACE_Name_Space
  // = TITLE
  //     Maintaining accesses Local Name Server Database.  Allows to
  //     add NameBindings, change them, remove them and resolve
  //     NameBindings.
  //
  // = DESCRIPTION
  //     Manages a Naming Service for a local name space which includes
  //	 bindings for node_local and host_local naming contexts.
  //	 All strings are stored in wide character format.
  //     A Name Binding consists of a name (that's the key), a value 
  //	 string and an optional type string (no wide chars).
{
public:
  // = Initialization and termination methods.
  ACE_Local_Name_Space (void);
  // "Do-nothing" constructor.

  ACE_Local_Name_Space (ACE_Naming_Context::Context_Scope_Type scope_in);
  // Specifies the scope of this namespace, opens and memory-maps the
  // associated file (if accessible) or contacts the dedicated name
  // server process for NET_LOCAL namespace.

  int open (ACE_Naming_Context::Context_Scope_Type scope_in);
  // Specifies the scope of this namespace, opens and memory-maps the
  // associated file (if accessible) or contacts the dedicated name
  // server process for NET_LOCAL namespace.

  ~ACE_Local_Name_Space (void);
  // destructor, do some cleanup :TBD: last dtor should "compress"
  // file

  virtual int bind (const ACE_WString &name,
		    const ACE_WString &value,
		    const char *type = "");
  // Bind a new name to a naming context (Wide character strings).
  
  virtual int rebind (const ACE_WString &name, 
		      const ACE_WString &value,
		      const char *type = "");
  // Overwrite the value or type of an existing name in a
  // ACE_Local_Name_Space or bind a new name to the context, if it
  // didn't exist yet. (Wide charcter strings interface).
  
  virtual int unbind (const ACE_WString &name);
  // Delete a name from a ACE_Local_Name_Space (Wide charcter strings
  // Interface).
  
  virtual int resolve (const ACE_WString &name, 
		       ACE_WString &value, 
		       char *&type);
  // Get value and type of a given name binding (Wide chars).  The
  // caller is responsible for deleting <type>!
  
  virtual int list_names (ACE_WSTRING_SET &set, 
			  const ACE_WString &pattern);
  // Get a set of names matching a specified pattern (wchars) matching
  // means the names must begin with the pattern string.

  virtual void dump (void) const;
  // Dump the state of the object

private:
  int shared_bind (const ACE_WString &name, const ACE_WString &value,
		   const char *type, int rebind);
  // Factor out code from bind() and rebind().

  int create_manager (void);
  // Allocate the appropriate type of map manager that stores the
  // key/value binding.

  typedef ACE_Map_Manager <ACE_NS_String, ACE_NS_String, PROCESS_MUTEX> MAP_MANAGER;
  typedef ACE_Map_Iterator <ACE_NS_String, ACE_NS_String, PROCESS_MUTEX> MAP_ITERATOR;
  typedef ACE_Map_Entry <ACE_NS_String, ACE_NS_String> MAP_ENTRY;

  MAP_MANAGER *name_space_map_;
  // Pointer to the allocated map manager.

  ACE_Naming_Context::Context_Scope_Type ns_scope_;
  // Scope of this naming context (e.g., PROC_LOCAL, NODE_LOCAL, or NET_LOCAL).

  // = The following defines our persistent, shared memory allocator.
  typedef ACE_Malloc <ACE_MMAP_Memory_Pool, PROCESS_MUTEX> MALLOC;
  typedef ACE_Allocator_Adapter <MALLOC> ALLOCATOR;

  ALLOCATOR *allocator_;
  // This is our allocator, which stores things in persistent, shared
  // memory.

  char context_file_[MAXNAMELEN];
  // Name of the file used as the backing store.

  static const char NAME_SERVER_MAP[];
  // Name of the map that's stored in shared memory.
};

#endif /* ACE_LOCAL_NAME_SPACE_H */
