/* -*- C++ -*- */

// ============================================================================
//
// = LIBRARY
//    ace
// 
// = FILENAME
//    Local_Tokens
//
// = AUTHOR
//    Karl-Heinz Dorn (kdorn@erlh.siemens.de)
//    Douglas C. Schmidt (schmidt@cs.wustl.edu)
//    Tim Harrison (harrison@cs.wustl.edu)
//
// = DESCRIPTION
//   This file contains definitions for the following classes:
//
// public:
//   7. ACE_Token_Proxy
//   8. ACE_Null_Token : public ACE_Token_Proxy
//   9. ACE_Local_Mutex : public ACE_Token_Proxy
//   *. ACE_Local_RLock : public ACE_Local_Mutex
//   &. ACE_Local_WLock : public ACE_Local_Mutex
// private:
//   1. ACE_TOKEN_CONST
//   3. ACE_TPQ_Entry
//   11. ACE_TSS_TPQ_Entry
//   4. ACE_Token_Proxy_Queue
//   5. ACE_Tokens
//   6. ACE_Mutex_Token : public ACE_Tokens
//   #. ACE_RW_Token : public ACE_Tokens
//   
// ============================================================================

#if !defined (ACE_LOCAL_MUTEX_H)
#define ACE_LOCAL_MUTEX_H

#include "ace/Synch.h"
#include "ace/Synch_Options.h"
#include "ace/Local_Tokens_T.h"

// 1.
class ACE_TOKEN_CONST
{
  // = TITLE
  //   Not a public interface.
  //   
  // = DESCRIPTION
  //   Constant definitions and typdefs for Token library.  Mostly,
  //   this class is necessary to fight the compiler with order of
  //   declaration errors.
public:
#if defined (ACE_MT_SAFE) 
// ACE platform supports some form of threading.
  typedef ACE_Condition_Mutex COND_VAR;
  typedef ACE_Mutex MUTEX;
  typedef ACE_Mutex_Guard GUARD;
#else
  typedef ACE_Null_Condition<ACE_Null_Mutex> COND_VAR;
  typedef ACE_Null_Mutex MUTEX;
  typedef ACE_Guard<ACE_Null_Mutex> GUARD;
#endif /* ACE_HAS_THREADS */

  enum Constants
    {
      MAXCLIENTIDLEN  = MAXHOSTNAMELEN + 20,
      MAXTOKENNAMELEN = 40
    };
};

class ACE_Token_Proxy;

// 3..
class ACE_TPQ_Entry
  // = TITLE
  //   Not a public interface.
  //   
  // = DESCRIPTION
  //   Token Proxy Queue entry.
  //   Used in the ACE_Token_Proxy_Queue
{
friend class ACE_Token_Manager;
public:
  typedef void (*PTVF) (void *);
 
  ACE_TPQ_Entry (void);
  // null constructor

  ACE_TPQ_Entry (const ACE_Token_Proxy *proxy,
		 void *type,
		 const char *client_id);
  // construction

  ACE_TPQ_Entry (const ACE_TPQ_Entry &rhs);
  // copy constructor

  ~ACE_TPQ_Entry (void);
  // death

  void operator= (const ACE_TPQ_Entry &rhs);
  // copy operator use by the queue

  // = set/get top of the queue
  ACE_Token_Proxy *proxy (void) const;
  void proxy (ACE_Token_Proxy *);

  // = delta/get nesting level of the entry
  int nesting_level (void) const;
  void nesting_level (int delta);

  // = set/get client_id of the entry
  const char *client_id (void) const;
  void client_id (const char *);

  void set (void *arg,
	    void (*sleep_hook)(void *));
  // one method for arg and sleep_hook

  // = set/get sleep hook of the entry
  void sleep_hook (void (*sh)(void *));
  PTVF sleep_hook (void) const;

  void call_sleep_hook (void);
  // call the sleep hook function or method passing arg.

  // = set/get arg of the entry
  void *arg (void) const;
  void arg (void *);

  // = set/get proxy type
  // i.e. MUTEX, RLOCK, WLOCK.
  void type (void *t);
  void *type (void) const;

  ACE_TPQ_Entry *next_;
  // pointer to next in list.

private:

  ACE_Token_Proxy *proxy_;
  // proxy

  int nesting_level_;
  // nesting level

  void *arg_;
  // arg

  char client_id_[ACE_TOKEN_CONST::MAXCLIENTIDLEN];
  // client id

  void (*sleep_hook_)(void *);
  // sleep hook
  
  void *type_;
  // type
};

// ************************************************************

// 11..
class ACE_TSS_TPQ_Entry : public ACE_TSS<ACE_TPQ_Entry>
  // = TITLE
  //   Not a public interface.
  // = DESCRIPTION
{
public:
  ACE_TSS_TPQ_Entry (const ACE_Token_Proxy *proxy,
		     void *type,
		     const char *client_id);
  // construction

  virtual ACE_TPQ_Entry *make_TSS_TYPE (void) const;
  // allows us to pass args to the construction of the TSS object

  operator ACE_TPQ_Entry *(void);
  // operator overloading and inheritence don't mix

  typedef ACE_TSS_TPQ_Entry TSS_TPQ;

private:
  // = These are passed to the constructor of ACE_TPQ_Entry in
  // make_TSS_TYPE
  const ACE_Token_Proxy *proxy_;
  // proxy
  void *type_;
  // type
  const char *client_id_;
  // client_id
};

// ************************************************************

// 4..
class ACE_Token_Proxy_Queue
  // = TITLE
  //   Not a public interface.
  //   
  // = DESCRIPTION
  //   Token waiter list.
  //   This queue holds all the token proxies waiting for ownership of
  //   a token.  Along with the proxy reference, it also stores the
  //   nesting level, client id, and a magic cookie from the proxy.
  //   This queue stores the ACE_TPQ_Entries by pointer values.  It
  //   DOES NOT make copies.  Thus, the user is responsible to ensure
  //   that the TPQ's stick around.  This is motivated by the need to
  //   reduce dynamic memory allocation.
{
public:
  ACE_Token_Proxy_Queue (void);
  // construction

  void enqueue (ACE_TPQ_Entry* new_entry,
		int position);
  // enqueue a proxy, nesting level, client_id, and a magic cookie at
  // the given position in the list.  If the position is -1, we
  // enqueue at the end of the list (I think)

  const ACE_TPQ_Entry* head (void);
  // top of the queue

  int member (const char *id);
  // is this id in the waiter list?

  int is_owning_reader (const ACE_Token_Proxy *reader);
  // This should only be used by ACE_RW_Token.
  // is this proxy an owner.  searches the list, from the head until
  // finding a matching id, or finding a proxy of a different
  // ACE_Token_Proxy::type.  returns 0 and 1.

  void dequeue (void);
  // remove the top waiter

  void remove (const ACE_TPQ_Entry *remove_me);
  // remove the waiter whose proxy ref matches remove_me

  int size (void);
  // the number of waiters

protected:
  ACE_TPQ_Entry *head_;
  // head
  ACE_TPQ_Entry *tail_;
  // tail
  int size_;
  // size
};

// 5..
class ACE_Tokens
  // = TITLE
  //   Not a public interface.
  //   
  // = DESCRIPTION
  //   Abstract representation of ACE tokens.
  //   Currently, I don't see a reason for providing an abstract
  //   interface at this level of the library.  As of yet, no one uses
  //   ACE_Tokens derivatives through this abstract interface except
  //   for Token_Manager.  It only uses the statistical methods which
  //   are shared by all Tokens.  For that reason, it still makes
  //   since to have a common base class.  However, acquire, renew,
  //   and release do not need to have matching interfaces throughout
  //   all Tokens.
{
public:

  ACE_Tokens (void);
  // null constructor

  virtual int acquire (ACE_TPQ_Entry *caller,
		       int ignore_deadlock,
		       int notify) = 0;
  // no implementation

  virtual int tryacquire (ACE_TPQ_Entry *caller) = 0;
  // no implementation

  virtual int renew (ACE_TPQ_Entry *caller,
		     int requeue_position) = 0;
  // no implementation

  virtual int release (ACE_TPQ_Entry *caller) = 0;
  // no implementation

  // = Accessor methods.

  virtual ACE_Token_Proxy *current_owner (void);
  // Returns a pointer to the current owner of the token, or 0 if
  // there is no owner.

  virtual ACE_Token_Proxy_Queue *waiters (void);
  // return the queue of waiters.

  virtual int no_of_waiters (void);
  // Return the number of proxies that are currently waiting to get
  // the token.

  const char *owner_id (void);
  // the current owner

  const char* name (void);
  // token name

  enum TOKEN_TYPE { MUTEX = 0, RWLOCK = 1 };
  
  virtual TOKEN_TYPE type (void) const = 0;
  // type of

  // = reference counting 
  void inc_reference (void);
  int dec_reference (void);

protected:

  int reference_count_;
  // reference count
  ACE_TPQ_Entry *owner (void);
  // all the data of the current owner
  ACE_Token_Proxy_Queue waiters_;
  // list of client's owning and waiting.
  char token_name_[ACE_TOKEN_CONST::MAXTOKENNAMELEN];
  // name of token
};

class ACE_Local_Mutex;

// 6..
class ACE_Mutex_Token : public ACE_Tokens
  // = TITLE
  //    Not a public interface.
  //
  // = DESCRIPTION 
  //    Class that acquires, renews, and releases a process-local
  //    synchronization token.
  //    This class is a more general-purpose synchronization mechanism
  //    than SunOS 5.x mutexes.  For example, it implements "recursive
  //    mutex" semantics, where a thread that owns the token can
  //    reacquire it without deadlocking.  In addition, threads that are
  //    blocked awaiting the token are serviced in strict FIFO order as
  //    other threads release the token (SunOS 5.x mutexes don't strictly
  //    enforce an acquisition order).
{
public:
  ACE_Mutex_Token (const char* name);
  // life

  ~ACE_Mutex_Token (void);
  // death

  // = Synchronization operations.
  // With acquire, renew, and release, the caller must be specified so
  // that multiple proxies (e.g. ACE_Local_Mutex) can use the same
  // token.

  virtual int acquire (ACE_TPQ_Entry *caller,
		       int ignore_deadlock,
		       int notify);
  // Returns 0 on success, -1 on failure with ACE_LOG_MSG->errnum() as
  // the reason.  If errnum == EWOULDBLOCK, and notify == 1,
  // ACE_Token_Proxy::sleep_hook() has been called on the current owner
  // of the token.  If ignore_deadlock is passed as 1 and errnum ==
  // EDEADLK, then deadlock was detected via ace_token_manager.

  virtual int tryacquire (ACE_TPQ_Entry *caller);
  // same as acquire, but fails if would block

  virtual int renew (ACE_TPQ_Entry *caller,
		     int requeue_position);
  // An optimized method that efficiently reacquires the token if no
  // other threads are waiting.  This is useful for situations where
  // you don't want to degrade the quality of service if there are
  // other threads waiting to get the token.  If <requeue_position> ==
  // -1 and there are other threads waiting to obtain the token we are
  // queued at the end of the list of waiters.  If <requeue_position>
  // > -1 then it indicates how many entries to skip over before
  // inserting our thread into the list of waiters (e.g.,
  // <requeue_position> == 0 means "insert at front of the queue").
  // Renew has the rather odd semantics such that if there are other
  // waiting threads it will give up the token even if the
  // nesting_level_ > 1.  I'm not sure if this is really the right
  // thing to do (since it makes it possible for shared data to be
  // changed unexpectedly) so use with caution...
  // Returns 0 on success, -1 on failure with ACE_LOG_MSG->errnum() as
  // the reason.  If errnum == EWOULDBLOCK, and notify == 1,
  // ACE_Token_Proxy::sleep_hook() has been called on the current owner
  // of the token.

  virtual int release (ACE_TPQ_Entry *caller);
  // Relinquish the token.  If there are any waiters then the next one
  // in line gets it.  If the caller is not the owner, caller is
  // removed from the waiter list.

  ACE_Tokens::TOKEN_TYPE type (void) const;
  // type of the token

private:
  ACE_TOKEN_CONST::MUTEX lock_;
  // ACE_Mutex_Token used to lock internal data structures.
};

// #.
class ACE_RW_Token : public ACE_Tokens
  // = TITLE
  //    Not a public interface.
  //
  // = DESCRIPTION 
  //    Class that acquires, renews, and releases a process-local
  //    synchronization token.
  //    This class is a more general-purpose synchronization mechanism
  //    than SunOS 5.x mutexes.  For example, it implements "recursive
  //    mutex" semantics, where a thread that owns the token can
  //    reacquire it without deadlocking.  In addition, threads that are
  //    blocked awaiting the token are serviced in strict FIFO order as
  //    other threads release the token (SunOS 5.x mutexes don't strictly
  //    enforce an acquisition order).
{
public:
  ACE_RW_Token (const char* name);
  // life

  ~ACE_RW_Token (void);
  // death

  // = Synchronization operations.
  // With acquire, renew, and release, the caller must be specified so
  // that multiple proxies (e.g. ACE_Local_Mutex) can use the same
  // token.

  virtual int acquire (ACE_TPQ_Entry *caller,
		       int ignore_deadlock,
		       int notify);
  // Returns 0 on success, -1 on failure with ACE_LOG_MSG->errnum() as
  // the reason.  If errnum == EWOULDBLOCK, and notify == 1,
  // ACE_Token_Proxy::sleep_hook() has been called on the current owner
  // of the token.  If ignore_deadlock is passed as 1 and errnum ==
  // EDEADLK, then deadlock was detected via ace_token_manager.

  virtual int tryacquire (ACE_TPQ_Entry *caller);
  // same as acquire except fails on would block

  virtual int renew (ACE_TPQ_Entry *caller,
		     int requeue_position);
  // An optimized method that efficiently reacquires the token if no
  // other threads are waiting.  This is useful for situations where
  // you don't want to degrade the quality of service if there are
  // other threads waiting to get the token.  If <requeue_position> ==
  // -1 and there are other threads waiting to obtain the token we are
  // queued at the end of the list of waiters.  If <requeue_position>
  // > -1 then it indicates how many entries to skip over before
  // inserting our thread into the list of waiters (e.g.,
  // <requeue_position> == 0 means "insert at front of the queue").
  // Renew has the rather odd semantics such that if there are other
  // waiting threads it will give up the token even if the
  // nesting_level_ > 1.  I'm not sure if this is really the right
  // thing to do (since it makes it possible for shared data to be
  // changed unexpectedly) so use with caution...
  // Returns 0 on success, -1 on failure with ACE_LOG_MSG->errnum() as
  // the reason.  If errnum == EWOULDBLOCK, and notify == 1,
  // ACE_Token_Proxy::sleep_hook() has been called on the current owner
  // of the token.

  virtual int release (ACE_TPQ_Entry *caller);
  // Relinquish the token.  If there are any waiters then the next one
  // in line gets it.  If the caller is not the owner, caller is
  // removed from the waiter list.

  ACE_Tokens::TOKEN_TYPE type (void) const;
  // returns ACE_Tokens::RWLOCk

protected:
  int num_writers_;
  // the number of waiting writers.

  ACE_TOKEN_CONST::MUTEX lock_;
  // ACE_Mutex_Token used to lock internal data structures.

  void notify_new_owner (ACE_TPQ_Entry *caller);
  // Sets the new owner.
};

// 7..
class ACE_Token_Proxy
  // = TITLE
  //   Abstract representation of ACE tokens.
  //   
  // = DESCRIPTION
  //   Interface for all Tokens in ACE.  This class implements the
  //   synchronization needed for tokens (condition variables etc.)
  //   The algorithms for the operations (acquire, release, etc.)
  //   operate on the generic ACE_Tokens interface.  Thus, the _type_
  //   of token (mutex, rwlock) can be set at construction of
  //   ACE_Token_Proxy.  You can use all Tokens in ACE through the
  //   ACE_Token_Proxy by passing the proper values at construction.
  //   Alternatively, there are class definitions which "know" how to
  //   do this (ACE_Local_Mutex, ACE_Local_RLock, ACE_Local_WLock).

  // = RESTRICTIONS
  //   Tokens (e.g. ACE_Mutex_Token) assume that it can always call
  //   ACE_Token_Proxy::token_acquired () on a new token owner.  This
  //   is not a problem for synchronous use of token proxies (that is,
  //   when acquires block until successful.)  However, for
  //   implementations of the Token Server, which may use asynch
  //   operations, the proxy can not go away after an acquire until
  //   the token is acquired.  This is not really a problem, but
  //   should be understood.
{
  friend class ACE_Token_Manager;
public:

  enum PROXY_TYPE { MUTEX = 1, RLOCK = 2, WLOCK = 3, 
		    REMOTE_MUTEX = 3, 
		    REMOTE_RLOCK = 4, REMOTE_WLOCK = 5 };

  // Initialization and termination methods.
  ACE_Token_Proxy (ACE_Token_Proxy::PROXY_TYPE lock_type,
		   ACE_Tokens::TOKEN_TYPE token_type);
  // construction

  ACE_Token_Proxy (const char *token_name,
		   int ignore_deadlock = 0,
		   int debug = 0,
		   ACE_Token_Proxy::PROXY_TYPE lock_type = 
		   ACE_Token_Proxy::MUTEX,
		   ACE_Tokens::TOKEN_TYPE token_type = 
		   ACE_Tokens::MUTEX);
  // construction

  ACE_Token_Proxy (const ACE_Token_Proxy &rhs);
  // construction

  ~ACE_Token_Proxy (void);
  // death

  virtual int open (const char *token_name,
		    int ignore_deadlock = 0,
		    int debug = 0);
  // assign the name to the proxy


  // = The following methods have implementations which are
  // independent of the token semantics (mutex, rwlock, etc.)  They
  // forward operations to the underlying token and perform the
  // necessary blocking semantics for operations (condition variables
  // etc.)  This allows reuse of the blocking code as well as having
  // multiple proxies to the same token.

  virtual int acquire (int notify = 0,
		       void (*sleep_hook)(void *) = 0,
		       ACE_Synch_Options &options =
		       ACE_Synch_Options::defaults);
  // calls acquire on the token.  blocks the calling thread if would
  // block.

  virtual int renew (int requeue_position = 0,
		     ACE_Synch_Options &options =
		     ACE_Synch_Options::defaults);
  // calls renew on the token.  blocks the calling thread if would
  // block.

  virtual int tryacquire (void (*sleep_hook)(void *) = 0);
  // calls renew on the token.

  virtual int release (ACE_Synch_Options &options =
		       ACE_Synch_Options::defaults);
  // calls release on the token.

  virtual int remove (ACE_Synch_Options &options =
		      ACE_Synch_Options::defaults);
  // calls remove on the token.

  // = utility methods

  virtual const char *client_id (void) const;
  // Get the client id of the proxy.  This is implemented as
  // thread-specific data.

  virtual void client_id (const char *client_id);
  // Set the client_id for the calling thread.  I strongly recommend
  // that this not be used unless you really know what you're doing.
  // I use this in the Token Server, and it caused many headaches.

  virtual const char *name (void) const;
  // return the name of the token.  This is important for use within
  // the token servers (local and remote) as well as with token
  // collections.  So, all derivations of ACE_Token_Proxy must be able to
  // stringify some name.  The name must uniquely identify a token.
  // So, for instance, the token within the reactor should probably be
  // called "Reactor Token."

  virtual ACE_Token_Proxy *clone (void);
  // return a dynamically allocated clone of the derived class.

  virtual void sleep_hook (void);
  // This should really be called someone_waiting ().
  // This is called by ACE_Token_xx's when another proxy enters the
  // waiting list and requests that the current token holder be notified.

  virtual void token_acquired (void*);
  // This is called when a queued (waiting) proxy is removed from the
  // waiters list and given the token.

  virtual const char *owner_id (void);
  // the client id of the current token holder

  PROXY_TYPE type (void) const;
  // RTTI

protected:
  ACE_Token_Proxy::PROXY_TYPE proxy_type_;
  // proxy type

  ACE_Token_Name<ACE_Tokens::TOKEN_TYPE> token_name_;
  // token name

  int ignore_deadlock_;
  // if this is set, we ignore deadlock

  int debug_;
  // print a bunch of debug messages

  ACE_Tokens *mutex_;
  // Reference to the actual logical token.  Many ACE_Local_Mutex
  // proxies can reference the same ACE_Mutex_Token.

  int handle_options (ACE_Synch_Options &options, ACE_TOKEN_CONST::COND_VAR &cv);
  // handles cond_var waits.

  ACE_TSS_TPQ_Entry::TSS_TPQ waiter_;
  // Waiter info used for asynchronous transactions
};

// 8..
class ACE_Null_Token : public ACE_Token_Proxy
  // = TITLE
  //   No-op class for non-threaded platform protocols.
{
public:
  ACE_Null_Token (void)
    : ACE_Token_Proxy ((ACE_Token_Proxy::PROXY_TYPE) 0,
		       (ACE_Tokens::TOKEN_TYPE) 0) {};
  // construction

  virtual int acquire (int notify = 0,
		       void (*sleep_hook)(void *) = 0,
		       ACE_Synch_Options &options =
		       ACE_Synch_Options::defaults) { return 0; }
  // acquire

  virtual int renew (int requeue_position = 0,
		     ACE_Synch_Options &options =
		     ACE_Synch_Options::defaults) { return 0; }
  // renew

  virtual int tryacquire (void (*sleep_hook)(void *) = 0) { return 0; }
  // try acquire

  virtual int release (ACE_Synch_Options &options =
		       ACE_Synch_Options::defaults) { return 0; }
  // release

  virtual int remove (ACE_Synch_Options &options =
		      ACE_Synch_Options::defaults) { return 0; }
  // remove

  virtual ACE_Token_Proxy *clone (void) { return new ACE_Null_Token; }
  // return a dynamically allocated clone of the derived class.

  virtual ACE_Token_Proxy::PROXY_TYPE type (void) const 
    { return (ACE_Token_Proxy::PROXY_TYPE) 0; }
  // type
};

// 9..
class ACE_Local_Mutex : public ACE_Token_Proxy
  // = TITLE
  //    Class that acquires, renews, and releases a process-local
  //    synchronization token.
  //
  // = DESCRIPTION 
  //    This class is a more general-purpose synchronization mechanism
  //    than SunOS 5.x mutexes.  For example, it implements "recursive
  //    mutex" semantics, where a thread that owns the token can
  //    reacquire it without deadlocking.  In addition, threads that
  //    are blocked awaiting the token are serviced in strict FIFO
  //    order as other threads release the token (SunOS 5.x mutexes
  //    don't strictly enforce an acquisition order).  Lastly,
  //    ACE_Local_Mutex performs deadlock detection on acquire calls.
  //
  // = Synchronization operations.
  //   The interfaces for acquire, tryacquire, renew, release,
  //   etc. are defined in ACE_Token_Proxy.  The semantics for
  //   ACE_Local_Mutex are that of a mutex.
{
public:
  // = Initialization and termination.

  ACE_Local_Mutex (void);
  // construction

  ACE_Local_Mutex (const char *token_name,
		   int ignore_deadlock = 0,
		   int debug = 0);
  // construction
};

// *.
class ACE_Local_RLock : public ACE_Token_Proxy
  // = TITLE
  //    Class that acquires, renews, and releases a process-local
  //    readers locks.
  //
  // = DESCRIPTION 
  //    This class is a more general-purpose synchronization mechanism
  //    than SunOS 5.x RLockes.  For example, it implements "recursive
  //    RLock" semantics, where a thread that owns the token can
  //    reacquire it without deadlocking.  In addition, threads that are
  //    blocked awaiting the token are serviced in strict FIFO order as
  //    other threads release the token (SunOS 5.x RLockes don't strictly
  //    enforce an acquisition order).
  // 
  // = Synchronization operations.
  //   The interfaces for acquire, tryacquire, renew, release,
  //   etc. are defined in ACE_Token_Proxy.  The semantics for
  //   ACE_Local_RLock are that of a readers/writers lock.  Acquire
  //   for this class implies a reader acquisition.  That is, multiple
  //   clients may acquire a lock for read only.
{
public:
  // = Initialization and termination.

  ACE_Local_RLock (void);
  // construction

  ACE_Local_RLock (const char *token_name = 0,
		   int ignore_deadlock = 0,
		   int debug = 0);
  // construction
};

// *.
class ACE_Local_WLock : public ACE_Token_Proxy
  // = TITLE
  //    Class that acquires, renews, and releases a process-local
  //    writer locks.
  //
  // = DESCRIPTION 
  //    This class is a more general-purpose synchronization mechanism
  //    than SunOS 5.x WLock.  For example, it implements "recursive
  //    WLock" semantics, where a thread that owns the token can
  //    reacquire it without deadlocking.  In addition, threads that are
  //    blocked awaiting the token are serviced in strict FIFO order as
  //    other threads release the token (SunOS 5.x WLocks don't strictly
  //    enforce an acquisition order).
  //
  // = Synchronization operations.
  //   The interfaces for acquire, tryacquire, renew, release,
  //   etc. are defined in ACE_Token_Proxy.  The semantics for
  //   ACE_Local_WLock are that of a readers/writers lock.  Acquire
  //   for this class implies a writer acquisition.  That is, only one
  //   client may hold the lock for writing.
{
public:
  // = Initialization and termination.

  ACE_Local_WLock (void);
  // construction

  ACE_Local_WLock (const char *token_name = 0,
		   int ignore_deadlock = 0,
		   int debug = 0);
  // construction
};

#if defined (__INLINE__)
#include "ace/Local_Tokens.i"
#endif /* __INLINE__ */
#endif /* ACE_LOCAL_MUTEX_H */
