// Provides a useful variable-length argument error handling
// abstraction.

#include "ace/FIFO_Send_Msg.h"
#include "ace/Thread.h"
#include "ace/Log_Msg.h"

ACE_ALLOC_HOOK_DEFINE(ACE_Log_Msg)

#if !(defined (ACE_HAS_THREADS) && defined (ACE_HAS_THREAD_SPECIFIC_STORAGE))
typedef ACE_Log_Msg ACE_LOG_MSG_INSTANCE;
#define ACE_LOG_MSG_GET(INSTANCE) (INSTANCE)
#else
typedef ACE_TSS<ACE_Log_Msg> ACE_LOG_MSG_INSTANCE;
#define ACE_LOG_MSG_GET(INSTANCE) ((INSTANCE)->operator ACE_Log_Msg * ())
#endif /* !(defined (ACE_HAS_THREADS) && defined (ACE_HAS_THREAD_SPECIFIC_STORAGE)) */

void
ACE_Log_Msg::dump (void) const
{
  ACE_TRACE ("ACE_Log_Msg::dump");
}

ACE_Log_Msg *
ACE_Log_Msg::instance (void)
{
  static ACE_LOG_MSG_INSTANCE *instance_;

  if (instance_ == 0)
    {
      // This code uses the "Double-Check" pattern to prevent a race
      // condition if multiple threads try to initialize the Singleton
      // simultaneously.  The great thing about this is that we don't
      // pay the price of locking on every call, only the "first"
      // calls.  The pattern is also used in the
      // ACE_TSS::operator->().
      ACE_MT (ACE_Guard<ACE_Recursive_Mutex> m (ACE_Log_Msg::lock_));

      if (instance_ == 0)
	instance_ = new ACE_LOG_MSG_INSTANCE;
    }

  return ACE_LOG_MSG_GET (instance_);
}

// Records the program name. 
const char *ACE_Log_Msg::program_name_ = "<unknown>";            

// Default is to use stderr.
u_long ACE_Log_Msg::flags_ = ACE_Log_Msg::STDERR;

#if defined (ACE_MT_SAFE)
// Synchronize output operations.
ACE_Recursive_Mutex ACE_Log_Msg::lock_;
#endif /* ACE_MT_SAFE */

// Process id of the current process. 
pid_t ACE_Log_Msg::pid_ = 0;

// Message queue. 
ACE_FIFO_Send_Msg ACE_Log_Msg::message_queue_;

// Call after a fork to resynchronize the PID and PROGRAM_NAME
// variables.

void
ACE_Log_Msg::sync (const char *prog_name)
{
  ACE_TRACE ("ACE_Log_Msg::sync");
  ACE_Log_Msg::program_name_ = prog_name;
  ACE_Log_Msg::pid_ = ACE_OS::getpid ();
}

u_long
ACE_Log_Msg::flags (void)
{
  ACE_TRACE ("ACE_Log_Msg::flags");
  u_long result;
  ACE_MT (ACE_Guard<ACE_Recursive_Mutex> m (this->lock_));
  result = ACE_Log_Msg::flags_;
  return result;
}

void
ACE_Log_Msg::set_flags (u_long flgs)
{
  ACE_TRACE ("ACE_Log_Msg::set_flags");
  ACE_MT (ACE_Guard<ACE_Recursive_Mutex> m (this->lock_));
  ACE_SET_BITS (ACE_Log_Msg::flags_, flgs);
}

void
ACE_Log_Msg::clr_flags (u_long flgs)
{
  ACE_TRACE ("ACE_Log_Msg::clr_flags");
  ACE_MT (ACE_Guard<ACE_Recursive_Mutex> m (this->lock_));
  ACE_CLR_BITS (ACE_Log_Msg::flags_, flgs);
}

ACE_Log_Msg::ACE_Log_Msg (void)
  : status_ (0),
    errnum_ (0),
    linenum_ (0),
    restart_ (1) // Restart by default...
{
  ACE_TRACE ("ACE_Log_Msg::ACE_Log_Msg");
}

// Open the sender-side of the Message ACE_Queue.

int
ACE_Log_Msg::open (const char *prog_name, 
		   u_long flags, 
		   const char *logger_key)
{
  ACE_TRACE ("ACE_Log_Msg::open");
  ACE_MT (ACE_Guard<ACE_Recursive_Mutex> m (this->lock_));
  // Note, if we put a trace call here, we will lose!
  ACE_Log_Msg::program_name_ = prog_name;
  ACE_Log_Msg::pid_ = ACE_OS::getpid ();
  int status = 0;

  // Note that if we fail to open the message queue the default action
  // is to use stderr (set via static initialization in the
  // ACE_Log_Msg.C file).
  
  if (ACE_BIT_ENABLED (flags, ACE_Log_Msg::LOGGER))
    {
      status = ACE_Log_Msg::message_queue_.open (logger_key);
	
      if (status == -1)
	ACE_SET_BITS (ACE_Log_Msg::flags_, ACE_Log_Msg::STDERR);
      else
	ACE_SET_BITS (ACE_Log_Msg::flags_, ACE_Log_Msg::LOGGER);
    }

  // Remember, ACE_Log_Msg::STDERR bit is on by default... 
  if (status != -1 
      && ACE_BIT_ENABLED (flags, ACE_Log_Msg::STDERR) == 0)
    ACE_CLR_BITS (ACE_Log_Msg::flags_, ACE_Log_Msg::STDERR);

  if (ACE_BIT_ENABLED (flags, ACE_Log_Msg::VERBOSE))
    ACE_SET_BITS (ACE_Log_Msg::flags_, ACE_Log_Msg::VERBOSE);

  if (ACE_BIT_ENABLED (flags, ACE_Log_Msg::SILENT))
    ACE_SET_BITS (ACE_Log_Msg::flags_, ACE_Log_Msg::SILENT);

  return status;
}

// Valid Options (prefixed by '%', as in printf format strings) include:
//   'a': exit the program at this point (var-argument is the exit status!)
//   'c': print a character
//   'i', 'd': print a decimal number
//   'e', 'E', 'f', 'F', 'g', 'G': print a double
//   'l', print line number where an error occurred.
//   'N': print file name where the error occurred.
//   'n': print the name of the program (or "<unknown>" if not set)
//   'o': print as an octal number
//   'P': format the current process id
//   'p': format the appropriate errno value from sys_errlist
//   'r': call the function pointed to by the corresponding argument
//   'R': print return status
//   'S': format the appropriate _sys_siglist entry corresponding to var-argument.
//   's': format a character string
//   'T': print timestamp in hour:minute:sec:usec format.
//   't': print thread id (1 if single-threaded)
//   'u': print as unsigned int
//   'X', 'x': print as a hex number
//   '%': format a single percent sign, '%' 

ssize_t
ACE_Log_Msg::log (ACE_Log_Priority log_priority, 
		  char *format_str, ...)
{ 
  ACE_TRACE ("ACE_Log_Msg::log");
  // External decls. 
  extern int sys_nerr;
  typedef void (*PTF)(...);

  va_list    argp;		// Start of variable args section. 
  ACE_Log_Record log_record (log_priority, ACE_OS::time ((time_t *) 0), ACE_Log_Msg::pid_);
  char	     *bp	  = this->msg ();
  int	     abort_prog   = 0;
  int	     exit_value   = 0;
  int	     result       = 0;
  char	     *format      = ACE_OS::strdup (format_str);
  char	     *save_p      = format;	// remember for ACE_OS::free()

  if (format == 0)
    return -1;

  if (ACE_BIT_ENABLED (ACE_Log_Msg::flags_, ACE_Log_Msg::VERBOSE))
    {
      // Prepend the program name onto this message 

      for (const char *s = ACE_Log_Msg::program_name_; (*bp = *s) != '\0'; s++)
	bp++;

      *bp++ = '|';
    }

  va_start (argp, format_str);

  while (*format != '\0')
    {
      // Copy input to output until we encounter a %, however a
      // % followed by another % is not a format specification. 

      if (*format != '%')
	*bp++ = *format++;
      else if (format[1] == '%') // An "escaped" '%' (just print one '%').
	{
	  *bp++ = *format++;	// Store first % 
	  format++;	// but skip second % 
	}
      else
	{
	  char c;		// high use character	
	  char *fp;		// local format pointer	
	  int  wpc;		// width/precision cnt  
	  const int CONTINUE = 0;
	  const int SKIP_SPRINTF = -1;  // We must skip the sprintf phase 
	  const int SKIP_NUL_LOCATE = -2; // Skip locating the NUL character 
	  int type = CONTINUE;	// conversion type	
	  int  w[2];		// width/precision vals 

	  // % starts a format specification that ends with one of
	  // "arnPpSsdciouxXfFeEgG".  An optional width and/or precision
	  // (indicated by an "*") may be encountered prior to the
	  // nend of the specification, each consumes an int arg.
	  // A call to sprintf() does the actual conversion. 

	  fp = format++;	// Remember beginning of format.   
	  wpc = 0;		// Assume no width/precision seen. 

	  while (type == CONTINUE)
	    {
	      switch (*format++)
		{
		case 'a': // Abort program after handling all of format string. 
		  type = SKIP_SPRINTF;
		  abort_prog = 1; 
		  exit_value = va_arg (argp, int); 
		  break;
		case 'l':
		  ACE_OS::sprintf (bp, "%d", this->linenum ());
		  break;
		case 'N':
		  {
		    const char *file = this->file ();
		    ACE_OS::sprintf (bp, "%s", file ? file : "<unknown file>");
		    break;
		  }
		case 'n': // Print the name of the program. 
		  type = SKIP_SPRINTF;
		  ACE_OS::strcpy (bp, ACE_Log_Msg::program_name_ ? 
			    ACE_Log_Msg::program_name_ : "<unknown>");
		  break;
                case 'P': // format the current process id 
		  type = SKIP_SPRINTF;
		  ACE_OS::sprintf (bp, "%d", ACE_Log_Msg::pid_);
		  break;
		case 'p': // format the string assocated with the value of errno. 
		  {
		    type = SKIP_SPRINTF;
		    if (errno >= 0 && errno < sys_nerr) 
		      ACE_OS::sprintf (bp, "%s: %s", 
				 va_arg (argp, char *), strerror (errno));
		    else
		      ACE_OS::sprintf (bp, "%s: <unknown error> = %d", 
				 va_arg (argp, char *), errno);
		    break;
		  }
		case 'R': // format the return status of the operation. 
		  this->op_status (va_arg (argp, int));
		  ACE_OS::sprintf (bp, "%d", this->op_status ());
		  break;
		case 'r': // Run (invoke) this subroutine. 
		  type = SKIP_SPRINTF;
		  (*va_arg (argp, PTF))(); 
		  break;
		case 'S': // format the string associated with this signal number. 
		  {
		    int sig = va_arg (argp, int);
		    type = SKIP_SPRINTF;
#if defined (ACE_HAS_SYS_SIGLIST)
		    if (sig >= 0 && sig < NSIG)
		      ACE_OS::strcpy (bp, _sys_siglist[sig]);
		    else
		      ACE_OS::sprintf (bp, "<unknown signal> %d", sig);
#else
		    ACE_OS::sprintf (bp, "signal %d", sig);
#endif /* ACE_HAS_SYS_SIGLIST */
		    break;
		  }
		case 'T': /* format the timestamp*/
		  {
		    type = SKIP_SPRINTF;
		    char day_and_time[35];
		    ACE_OS::sprintf (bp, "%s", 
			       ACE::timestamp (day_and_time, sizeof day_and_time));
		    break;		  
		  }
		case 't': // format thread id. 
		  type = SKIP_SPRINTF;
		  hthread_t t_id;
		  ACE_Thread::self (t_id);
		  ACE_OS::sprintf (bp, "%d", t_id);
		  break;
		case 's':
		  type = 1 + wpc; // 1, 2, 3 
		  break;
		case 'd': case 'c': case 'i': case 'o': 
		case 'u': case 'x': case 'X': 
		  type = 4 + wpc; // 4, 5, 6 
		  break;
		case 'F': case 'f': case 'e': case 'E':
		case 'g': case 'G':
		  type = 7 + wpc; // 7, 8, 9 
		  break;
		case '*':	// consume width/precision 
		  w[wpc++] = va_arg (argp, int);
		  break;
		default:
		  // ? 
		  break;
		}
	    }

	  if (type != SKIP_SPRINTF)
	    {
	      c = *format;	// Remember char before we overwrite. 
	      *format = 0;	// Overwrite, terminating format. 

	      switch (type)
		{
		case 1:
		  ACE_OS::sprintf (bp, fp, va_arg (argp, char *));
		  break;
		case 2:
		  ACE_OS::sprintf (bp, fp, w[0], va_arg (argp, char *));
		  bp += w[0];
		  type = SKIP_NUL_LOCATE;
		  break;
		case 3:
		  ACE_OS::sprintf (bp, fp, w[0], w[1], va_arg (argp, char *));
		  bp += w[0];
		  type = SKIP_NUL_LOCATE;
		  break;
		case 4:
		  ACE_OS::sprintf (bp, fp, va_arg (argp, int));
		  break;
		case 5:
		  ACE_OS::sprintf (bp, fp, w[0], va_arg (argp, int));
		  break;
		case 6:
		  ACE_OS::sprintf (bp, fp, w[0], w[1], va_arg (argp, int));
		  break;
		case 7:
		  ACE_OS::sprintf (bp, fp, va_arg (argp, double));
		  break;
		case 8:
		  ACE_OS::sprintf (bp, fp, w[0], va_arg (argp, double));
		  break;
		case 9:
		  ACE_OS::sprintf (bp, fp, w[0], w[1], va_arg (argp, double));
		  break;
		}
	      *format = c;	// Restore char we overwrote. 
	    }

	  if (type != SKIP_NUL_LOCATE)
	    while (*bp != '\0')	// Locate end of bp. 
	      bp++;
	}
    }

  *bp++	= '\0';			// Terminate bp. 

  ACE_OS::free (ACE_MALLOC_T (save_p));

  // Only print the message if "SILENT" mode is disabled.
  if (!ACE_BIT_ENABLED (ACE_Log_Msg::flags_, ACE_Log_Msg::SILENT))
      {
	// Copy the message from thread-specific storage into the
	// transfer buffer (this can be optimized away by changing
	// other code...).
	log_record.msg_data (this->msg ());

	// Format the message and print it to stderr and/or ship it
	// off to the log_client daemon, and/or print it to the
	// ostream.

	if (ACE_BIT_ENABLED (ACE_Log_Msg::flags_, ACE_Log_Msg::OSTREAM)
	    && this->msg_ostream () != 0)
	  (*this->msg_ostream ()) << log_record;
	if (ACE_BIT_ENABLED (ACE_Log_Msg::flags_, ACE_Log_Msg::STDERR))
	  log_record.print ("<localhost>", 
			    ACE_BIT_ENABLED (ACE_Log_Msg::flags_, ACE_Log_Msg::VERBOSE));
	if (ACE_BIT_ENABLED (ACE_Log_Msg::flags_, ACE_Log_Msg::LOGGER))
	  {
	    ACE_Str_Buf log_msg ((void *) &log_record, 
				 int (log_record.length ()));
#if defined (ACE_HAS_STREAM_PIPES)
	    result = ACE_Log_Msg::message_queue_.send (int (log_record.type ()), 
						       &log_msg);
#else
	    result = ACE_Log_Msg::message_queue_.send (log_msg);
#endif /* ACE_HAS_STREAM_PIPES */
	  }
      }
  if (abort_prog)
    ACE_OS::exit (exit_value);

  va_end (argp);
  return result;
}

// Calls log to do the actual print, but formats first

int 
ACE_Log_Msg::log_hexdump (char *buffer, int size)
{
#if 0
  int result = 1;
  char obuf[MAXLOGMSGLEN];
  char *bp = obuf;

  ACE::format_hexdump (buffer, bp, size);

  ACE_OS::strcpy (this->msg (), obuf);

  // Now do the printing
  this->log (buf);
  return result;
#endif
  return 0;
}

void
ACE_Log_Msg::set (const char *filename, 
		  int line,
		  int status, 
		  int err, 
		  int rs,
		  ostream *os)
{
  ACE_TRACE ("ACE_Log_Msg::set");
  this->file (filename);
  this->linenum (line);
  this->op_status (status);
  this->errnum (err);
  this->restart (rs);
  this->msg_ostream (os);
}

void 
ACE_Log_Msg::op_status (int status)
{
  ACE_TRACE ("ACE_Log_Msg::op_status");
  this->status_ = status;
}

int 
ACE_Log_Msg::op_status (void)
{
  ACE_TRACE ("ACE_Log_Msg::op_status");
  return this->status_;
}

void 
ACE_Log_Msg::restart (int r)
{
  ACE_TRACE ("ACE_Log_Msg::restart");
  this->restart_ = r;
}

int 
ACE_Log_Msg::restart (void)
{
  ACE_TRACE ("ACE_Log_Msg::restart");
  return this->restart_;
}

int 
ACE_Log_Msg::errnum (void)
{
  ACE_TRACE ("ACE_Log_Msg::errnum");
  return this->errnum_;
}

void
ACE_Log_Msg::errnum (int e)
{
  ACE_TRACE ("ACE_Log_Msg::errnum");
  this->errnum_ = e;
}

int 
ACE_Log_Msg::linenum (void)
{
  ACE_TRACE ("ACE_Log_Msg::linenum");
  return this->linenum_;
}

void
ACE_Log_Msg::linenum (int l)
{
  ACE_TRACE ("ACE_Log_Msg::linenum");
  this->linenum_ = l;
}

const char *
ACE_Log_Msg::file (void)
{
  ACE_TRACE ("ACE_Log_Msg::file");
  return this->file_;
}

void
ACE_Log_Msg::file (const char *s)
{
  ACE_TRACE ("ACE_Log_Msg::file");
  ACE_OS::strncpy (this->file_, s, sizeof this->file_);
}

char *
ACE_Log_Msg::msg (void)
{
  ACE_TRACE ("ACE_Log_Msg::msg");
  return this->msg_;
}

void
ACE_Log_Msg::msg (char *m)
{
  ACE_TRACE ("ACE_Log_Msg::msg");
  ACE_OS::strncpy (this->msg_, m, sizeof this->msg_);
}

ostream *
ACE_Log_Msg::msg_ostream (void)
{
  ACE_TRACE ("ACE_Log_Msg::msg_ostream");
  return this->ostream_;
}

void
ACE_Log_Msg::msg_ostream (ostream *m)
{
  ACE_TRACE ("ACE_Log_Msg::msg_ostream");
  this->ostream_ = m;
}
