/* -*- C++ -*- */
// ============================================================================
//
// = LIBRARY
//    ace
// 
// = FILENAME
//    Log_Msg.h
//
// = AUTHOR
//    Doug Schmidt 
// 
// ============================================================================

#if !defined (ACE_LOG_MSG_H)
#define ACE_LOG_MSG_H

// This stuff must come first to avoid problems with circular
// headers...

#if defined (ACE_NLOGGING)
#define ACE_RETURN(Y) return (Y) 
#define ACE_ERROR_RETURN(X,Y) return (Y)
#define ACE_ERROR(X) 
#define ACE_DEBUG(X) 
#define ACE_ERROR_INIT(VALUE, FLAGS)
#else
#define ACE_RETURN(Y) \
  do { \
    ACE_Log_Msg::instance ()->set (__FILE__, __LINE__, Y, errno); \
    return Y; \
  } while (0)
#define ACE_ERROR_RETURN(X,Y) \
  do { \
    ACE_Log_Msg *ace___ = ACE_Log_Msg::instance (); \
    ace___->set (__FILE__, __LINE__, Y, errno, ace___->restart (), ace___->msg_ostream ()); \
    ace___->log X; \
    return Y; \
  } while (0)
#define ACE_ERROR(X) \
  do { \
    ACE_Log_Msg *ace___ = ACE_Log_Msg::instance (); \
    ace___->set (__FILE__, __LINE__, -1, errno, ace___->restart (), ace___->msg_ostream ()); \
    ace___->log X; \
  } while (0)
#define ACE_DEBUG(X) \
  do { \
    ACE_Log_Msg *ace___ = ACE_Log_Msg::instance (); \
    ace___->set (__FILE__, __LINE__, 0, errno, ace___->restart (), ace___->msg_ostream ()); \
    ace___->log X; \
  } while (0)
#define ACE_ERROR_INIT(VALUE, FLAGS) \
  do { \
    ACE_Log_Msg *ace___ = ACE_Log_Msg::instance (); \
    ace___->set_flags (FLAGS); ace___->op_status (VALUE); \
  } while (0)
#endif /* ACE_NLOGGING */

#include "ace/Log_Record.h"
#include "ace/Synch.h"

// Forward declarations.
class ACE_FIFO_Send_Msg;
class ACE_Log_Msg;

#define ACE_LOG_MSG ACE_Log_Msg::instance ()

class ACE_Log_Msg 
  // = TITLE
  //     Provides a variable-length argument message logging
  //     abstraction.
  //
  // = DESCRIPTION
  //     This class is very flexible since it allows formatted error
  //     messages to be printed in a thread-safe manner to stderr or a
  //     distributed logger.  Moreover, the message is kept in a
  //     thread-specific storage location, which can be used to
  //     communicate errors between framework methods and callers.
{
public:
  enum /* Logger Flags */
  {
    STDERR       = 01,  
    // Write messages to stderr. 
    LOGGER       = 02,  
    // Write messages to the local client logger deamon.
    OSTREAM     = 04,
    // Write messages to the ostream * stored in thread-specific
    // storage.
    VERBOSE	 = 010,
    // Display messages in a verbose manner 
    SILENT       = 020  
    // Do not print messages at all (just leave in thread-specific
    // storage for later inspection).
  };

  // = Initialization and termination routines.

  static ACE_Log_Msg *instance (void);
  // Returns a pointer to the Singleton.

  ACE_Log_Msg (void);
  // Initialize logger.

  int open (const char *prog_name, 
	    u_long options_flags = ACE_Log_Msg::STDERR, 
	    const char *logger_key = 0);
  // Initialize the ACE error handling facility.  <prog_name> is the
  // name of the executable program.  <flags> are a bitwise-or of
  // options flags passed to the Logger (see the enum above for the valid
  // values).  If the <LOGGER> bit in <flags> is enabled then
  // <logger_key> is the name of ACE_FIFO rendezvous point where the
  // local client logger daemon is listening for logging messages.

  // = Set/get the options flags.

  void set_flags (u_long f);
  // Enable the bits in the logger's options flags.
  void clr_flags (u_long f);
  // Disable the bits in the logger's options flags.
  u_long flags (void);
  // Return the bits in the logger's options flags.

  void sync (const char *program_name);
  // Call after doing a fork() to resynchronize the PID and
  // PROGRAM_NAME variables.

  // = Set/get methods.  Note that these are non-static and thus will
  // be thread-specific.

  void op_status (int status);
  // Set the result of the operation status (by convention, -1 means error).

  int op_status (void);
  // Get the result of the operation status (by convention, -1 means error).

  void errnum (int);
  // Set the value of the errnum (by convention this corresponds to errno).

  int errnum (void);
  // Get the value of the errnum (by convention this corresponds to errno).

  void linenum (int);
  // Set the line number where an error occurred.

  int linenum (void);
  // Get the line number where an error occurred.

  void file (const char *);
  // Set the file name where an error occurred.

  const char *file (void);
  // Get the file name where an error occurred.

  void msg (char *);
  // Set the message that describes what type of error occurred.

  char *msg (void);
  // Get the message that describes what type of error occurred.

  void restart (int);
  // Set the field that indicates whether interrupted calls should be
  // restarted.

  int restart (void);
  // Get the field that indicates whether interrupted calls should be
  // restarted.

  void msg_ostream (ostream *);
  // Set the ostream that is used to print error messages.

  ostream *msg_ostream (void);
  // Get the ostream that is used to print error messages.

  void set (const char *file, 
	    int line,
	    int op_status = -1,
	    int errnum = 0,
	    int restart = 1,
	    ostream *os = 0);
  // Set the line number, file name, operational status, error number,
  // restart flag, and ostream.  This combines all the other set
  // methods into a single method.

  ssize_t log (ACE_Log_Priority priority, char *format, ...);
  // Format a message to the ACE error logging mechanism.  Valid
  // options (prefixed by '%', as in printf format strings) include:
  // 'a': exit the program at this point (var-argument is the exit status!)
  // 'c': print a character
  // 'i', 'd': print a decimal number
  // 'e', 'E', 'f', 'F', 'g', 'G': print a double
  // 'l', print line number where an error occurred.
  // 'N': print file name where the error occurred.
  // 'n': print the name of the program (or "<unknown>" if not set)
  // 'o': print as an octal number
  // 'P': print out the current process id
  // 'p': print out the appropriate errno value from sys_errlist
  // 'r': call the function pointed to by the corresponding argument
  // 'R': print return status
  // 'S': print out the appropriate _sys_siglist entry corresponding to var-argument.
  // 's': print out a character string
  // 'T': print timestamp in hour:minute:sec:usec format.
  // 't': print thread id (1 if single-threaded)
  // 'u': print as unsigned int
  // 'X', 'x': print as a hex number
  // '%': print out a single percent sign, '%' 

  int log_hexdump (char *buffer, int size);
  // Method to log hex dump.  This is useful for debugging.  Calls
  // <log> to do the actual print, but formats first to make the chars
  // printable.
  
  void dump (void) const;
  // Dump the state of an object.

  ACE_ALLOC_HOOK_DECLARE;
  // Declare the dynamic allocation hooks.

private:
  int status_;
  // Status of operation (-1 means failure, >= 0 means success).

  int errnum_;
  // Type of error that occurred (see <sys/errno.h>).

  int linenum_;
  // Line number where the error occurred.

  char file_[MAXNAMELEN];
  // File where the error occurred.

  char msg_[ACE_Log_Record::MAXLOGMSGLEN];
  // The error message.

  int restart_;
  // Indicates whether we should restart system calls that are
  // interrupted.

  ostream *ostream_;
  // The ostream where logging messages can be written.

  // = The following fields are *not* kept in thread-specific storage
  // since we only want one instance for the entire process!
  
  static ACE_FIFO_Send_Msg message_queue_;
  // IPC conduit between sender and client daemon. 

  static const char *program_name_; 
  // Records the program name. 

  static pid_t pid_;
  // Process id of the current process. 

  static u_long flags_;
  // Options flags.

#if defined (ACE_MT_SAFE)  
  static ACE_Recursive_Mutex lock_;
  // Synchronize output operations.  
#endif /* ACE_MT_SAFE */
};

// #include "ace/Log_Msg.i"
#endif /* ACE_LOG_MSG_H */
