/* -*- C++ -*- */

// ============================================================================
//
// = LIBRARY
//    ace
// 
// = FILENAME
//     ACE_Memory_Pool.h
//
// = AUTHOR
//    Doug Schmidt 
// 
// ============================================================================

#if !defined (ACE_MEMORY_POOL_H)
#define ACE_MEMORY_POOL_H

#include "ace/ACE.h"
#include "ace/Event_Handler.h"
#include "ace/Signal.h"
#if !defined (ACE_WIN32)
#include "ace/SV_Semaphore_Complex.h"
#endif /* !ACE_WIN32 */

#if !defined (ACE_WIN32)
class ACE_Sbrk_Memory_Pool
  // = TITLE
  //     Make a memory pool that is based on <sbrk(2)>.
{
public:
  ACE_Sbrk_Memory_Pool (const char *pool_name = 0);
  // Initialization constructor.

  // = Implementor operations.
  void *init_acquire (size_t nbytes, 
		      size_t &rounded_bytes, 
		      int &first_time);
  // Ask system for initial chunk of local memory. 

  void init_release (void);
  // No-op.

  void *acquire (size_t nbytes, 
		 size_t &rounded_bytes);
  // Acquire at least NBYTES from the memory pool.  ROUNDED_BYTES is
  // the actual number of bytes allocated.

  int release (void);
  // Instruct the memory pool to release all of its resources.

  int sync (ssize_t len = -1, int flags = MS_SYNC);
  // Sync <len> bytes of the memory region to the backing store
  // starting at <this->base_addr_>.  If <len> == -1 then sync the
  // whole region.

  int sync (void *addr, size_t len, int flags = MS_SYNC);
  // Sync <len> bytes of the memory region to the backing store
  // starting at <addr_>.

  int protect (ssize_t len = -1, int prot = PROT_RDWR);
  // Change the protection of the pages of the mapped region to <prot>
  // starting at <this->base_addr_> up to <len> bytes.  If <len> == -1
  // then change protection of all pages in the mapped region.

  int protect (void *addr, size_t len, int prot = PROT_RDWR);
  // Change the protection of the pages of the mapped region to <prot>
  // starting at <addr> up to <len> bytes.

  void dump (void) const;
  // Dump the state of an object.

  ACE_ALLOC_HOOK_DECLARE;
  // Declare the dynamic allocation hooks.

protected:
  size_t round_up (size_t nbytes);
  // Implement the algorithm for rounding up the request to an
  // appropriate chunksize.
};

class ACE_Shared_Memory_Pool : public ACE_Event_Handler
  // = TITLE
  //     Make a memory pool that is based on System V shared memory
  //     (shmget(2) etc.).  This implementation allows memory to be
  //     shared between processes.
{
public:
  ACE_Shared_Memory_Pool (const char *pool_name = ACE_ITOA(ACE_DEFAULT_SHM_KEY));
  // Initialization constructor.

  void *init_acquire (size_t nbytes, 
		      size_t &rounded_bytes, 
		      int &first_time);
  // Ask system for initial chunk of local memory.

  void *acquire (size_t nbytes, 
		 size_t &rounded_bytes);
  // Acquire at least NBYTES from the memory pool.  ROUNDED_BYTES is
  // the actual number of bytes allocated.  Also acquires an internal
  // semaphore that ensures proper serialization of Memory_Pool
  // initialization across processes.

  void init_release (void);
  // Release the internal semaphore held by the Memory_Pool;

  int release (void);
  // Instruct the memory pool to release all of its resources.

  int sync (ssize_t len = -1, int flags = MS_SYNC);
  // Sync the memory region to the backing store starting at
  // <this->base_addr_>.

  int sync (void *addr, size_t len, int flags = MS_SYNC);
  // Sync the memory region to the backing store starting at <addr_>.

  int protect (ssize_t len = -1, int prot = PROT_RDWR);
  // Change the protection of the pages of the mapped region to <prot>
  // starting at <this->base_addr_> up to <len> bytes.  If <len> == -1
  // then change protection of all pages in the mapped region.

  int protect (void *addr, size_t len, int prot = PROT_RDWR);
  // Change the protection of the pages of the mapped region to <prot>
  // starting at <addr> up to <len> bytes.

  void dump (void) const;
  // Dump the state of an object.

  ACE_ALLOC_HOOK_DECLARE;
  // Declare the dynamic allocation hooks.

protected:
  size_t round_up (size_t nbytes);
  // Implement the algorithm for rounding up the request to an
  // appropriate chunksize.

  int commit_backing_store (size_t rounded_bytes, 
			    off_t &offset);
  // Commits a new shared memory segment if necessary after an
  // acquire() or a signal.  <offset> is set to the new offset into
  // the backing store.

  enum 
  { 
    BASE_ADDR = 16 * 1024 * 1024,
    SEGMENT_SIZE = 1024 * 128, // Each segment is 128K...
    MAX_SEGMENTS = 6 // What's the right number?
  };

  struct SHM_TABLE 
  {
    key_t key;
    int shmid;
    int used;
  };

  int in_use (off_t &offset, int &counter);
  // Determine how much memory is currently in use.

  ACE_Sig_Handler signal_handler_;
  // Handles SIGSEGV.

  virtual int handle_signal (int signum
#if defined (ACE_HAS_SIGINFO_T)
, siginfo_t *, ucontext_t *
#endif /* ACE_HAS_SIGINFO_T */
);
  // Handle SIGSEGV and SIGBUS signals to remap shared memory
  // properly.

  ACE_SV_Semaphore_Complex init_finished_;
  // Used to serialize initialization of the Memory_Pool and Malloc.
};
#endif /* !ACE_WIN32 */

class ACE_Local_Memory_Pool
  // = TITLE
  //   Make a memory pool that is based on C++ new/delete.  This is
  //   useful for integrating existing components that use new/delete
  //   into the ACE Malloc scheme...
{
public:
  ACE_Local_Memory_Pool (const char *pool_name = 0);
  // Initialization constructor.

  void *init_acquire (size_t nbytes, 
		      size_t &rounded_bytes, 
		      int &first_time);
  // Ask system for initial chunk of local memory. 

  void *acquire (size_t nbytes, 
		 size_t &rounded_bytes);
  // Acquire at least NBYTES from the memory pool.  ROUNDED_BYTES is
  // the actual number of bytes allocated.

  void init_release (void);
  // No-op.

  int release (void);
  // Instruct the memory pool to release all of its resources. 

  int sync (ssize_t len = -1, int flags = MS_SYNC);
  // Sync <len> bytes of the memory region to the backing store
  // starting at <this->base_addr_>.  If <len> == -1 then sync the
  // whole region.

  int sync (void *addr, size_t len, int flags = MS_SYNC);
  // Sync <len> bytes of the memory region to the backing store
  // starting at <addr_>.

  int protect (ssize_t len = -1, int prot = PROT_RDWR);
  // Change the protection of the pages of the mapped region to <prot>
  // starting at <this->base_addr_> up to <len> bytes.  If <len> == -1
  // then change protection of all pages in the mapped region.

  int protect (void *addr, size_t len, int prot = PROT_RDWR);
  // Change the protection of the pages of the mapped region to <prot>
  // starting at <addr> up to <len> bytes.

  void dump (void) const;
  // Dump the state of an object.

  ACE_ALLOC_HOOK_DECLARE;
  // Declare the dynamic allocation hooks.

protected:
  size_t round_up (size_t nbytes);

  // Implement the algorithm for rounding up the request to an
  // appropriate chunksize.
};

class ACE_MMAP_Memory_Pool : public ACE_Event_Handler
  // = TITLE
  //     Make a memory pool that is based on <mmap(2)>.  This
  //     implementation allows memory to be shared between processes.
{
public:
  ACE_MMAP_Memory_Pool (const char *pool_name = 0,
			int use_fixed_addr = 1, 
			int write_each_page = 1,
			char *base_addr = ACE_DEFAULT_BASE_ADDR,
			key_t sem_key = ACE_DEFAULT_SEM_KEY);
  // Initialization constructor.  

  void *init_acquire (size_t nbytes, 
		      size_t &rounded_bytes, 
		      int &first_time);
  // Ask system for initial chunk of shared memory. 

  void *acquire (size_t nbytes, 
		 size_t &rounded_bytes);
  // Acquire at least NBYTES from the memory pool.  ROUNDED_BYTES is
  // the actual number of bytes allocated.  Also acquires an internal
  // semaphore that ensures proper serialization of Memory_Pool
  // initialization across processes.

  void init_release (void);
  // Release the internal semaphore held by the Memory_Pool;

  int release (void);
  // Instruct the memory pool to release all of its resources. 

  int sync (ssize_t len = -1, int flags = MS_SYNC);
  // Sync the memory region to the backing store starting at
  // <this->base_addr_>.

  int sync (void *addr, size_t len, int flags = MS_SYNC);
  // Sync the memory region to the backing store starting at <addr_>.

  int protect (ssize_t len = -1, int prot = PROT_RDWR);
  // Change the protection of the pages of the mapped region to <prot>
  // starting at <this->base_addr_> up to <len> bytes.  If <len> == -1
  // then change protection of all pages in the mapped region.

  int protect (void *addr, size_t len, int prot = PROT_RDWR);
  // Change the protection of the pages of the mapped region to <prot>
  // starting at <addr> up to <len> bytes.

  void dump (void) const;
  // Dump the state of an object.

  ACE_ALLOC_HOOK_DECLARE;
  // Declare the dynamic allocation hooks.

protected:
  // = Implement the algorithm for rounding up the request to an
  // appropriate chunksize.

  size_t round_up (size_t nbytes);

  int commit_backing_store (size_t rounded_bytes, off_t &file_offset);
  // Compute the new file_offset of the backing store and commit the
  // memory.

  int map_file (off_t file_offset);
  // Memory map the file up to <file_offset> bytes.

  virtual int handle_signal (int signum
#if defined (ACE_HAS_SIGINFO_T)
, siginfo_t *, ucontext_t *
#endif /* ACE_HAS_SIGINFO_T */
);
  // Handle SIGSEGV and SIGBUS signals to remap shared memory
  // properly.

  ACE_Sig_Handler signal_handler_;
  // Handles SIGSEGV.

  ACE_HANDLE handle_;
  // Underlying HANDLE.

  void *base_addr_;
  // Base of mapped region.

  int flags_;
  // Flags passed into mmap(2).
  
  const int use_fixed_addr_;
  // Keeps track of whether we need to use a fixed address or not...

  size_t mapping_len_;
  // Keeps track of the current length of the mapping.

  const int write_each_page_;
  // Should we write a byte to each page to forceably allocate memory
  // for this backing store?

  char backing_store_[MAXPATHLEN];
  // Name of the backing store.

#if !defined(ACE_WIN32)
  ACE_SV_Semaphore_Complex init_finished_;   
  // Used to serialize initialization of the Memory_Pool and Malloc.
#else /* ACE_WIN32 */
  ACE_Process_Mutex init_finished_;
  // Win32 doesn't have the System V semaphores, we'll use global
  // named mutex's instead.
#endif /* !ACE_WIN32 */
};

#if defined (__INLINE__)
#include "ace/Memory_Pool.i"
#endif /* __INLINE__ */

#endif /* ACE_MEMORY_POOL_H */




