#include "ace/SString.h"
#include "ace/Set.h"
#include "ace/Naming_Context.h"
#include "ace/Name_Handler.h"

// Default constructor.
ACE_Name_Handler::ACE_Name_Handler (ACE_Reactor *reactor)
  : ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_NULL_SYNCH> (0, 0, reactor)
{
}

// Activate this instance of the ACE_Name_Handler (called by the
// ACE_Name_Acceptor).

/* VIRTUAL */ int 
ACE_Name_Handler::open (void *)
{
  ACE_INET_Addr addr;

  // Find out which client we are connected with.
  if (this->peer ().get_remote_addr (this->addr_) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "get_remote_addr"), -1);

  ACE_DEBUG ((LM_DEBUG, "%n: received connection from %s on handle %d\n", 
	      this->addr_.get_host_name (), this->get_handle ()));

  // Register this object with the ACE_Reactor.  We will then receive
  // upcalls when our client requests operations.
  if (this->reactor ()->register_handler (this, ACE_Event_Handler::READ_MASK) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "register_handler"), -1);

  // Instantiate our associated ACE_Naming_Context
  naming_context_ = new ACE_Naming_Context (ACE_Naming_Context::NET_LOCAL);

  return 0;
}

// Close down this instance of the ACE_Name_Handler.

/* VIRTUAL */ int 
ACE_Name_Handler::close (u_long)
{
  this->handle_close ();
  return 0;
}

// Create and send a reply to the client.

/* VIRTUAL */ int 
ACE_Name_Handler::send_reply (ACE_UINT32 status, ACE_UINT32 err)
{
  void *buf;
  this->name_reply_.msg_type (status);
  this->name_reply_.errnum (err);

  int len = this->name_reply_.encode (buf);

  if (len == -1)
    return -1;
    
  ssize_t n = this->peer ().send (buf, len);

  if (n != len)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n, expected len = %d, actual len = %d", 
		      "send failed", len, n), -1);
  else
    return 0;
}

ACE_Name_Handler::send_request (ACE_Name_Request &request)
{
  void    *buffer;
  ssize_t length;
 
  if ((length = request.encode (buffer)) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "encode failed"), -1);
 
  // Transmit request via a blocking send.
 
  if (this->peer ().send_n (buffer, length) != length)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "send_n failed"), -1);

  return 0;
}

// Give up waiting (e.g., when a timeout occurs or a client shuts down
// unexpectedly).
 
/* VIRTUAL */ int
ACE_Name_Handler::abandon ()
{
  int failure_reason = errno;
  return this->send_reply (ACE_Name_Reply::FAILURE, failure_reason);
}

// Enable clients to limit the amount of time they'll wait
 
/* VIRTUAL */ int
ACE_Name_Handler::handle_timeout (const ACE_Time_Value &, const void *)
{
  return this->abandon ();
}

// Return the underlying ACE_HANDLE.

/* VIRTUAL */ ACE_HANDLE 
ACE_Name_Handler::get_handle (void) const
{
  return this->peer ().get_handle ();
}

// Dispatch the appropriate operation to handle the client request.

/* VIRTUAL */ int 
ACE_Name_Handler::dispatch (void)
{
  // Dispatch the appropriate request.
  switch (this->name_request_.msg_type ())
    {
    case ACE_Name_Request::BIND:
      {
	ACE_DEBUG ((LM_DEBUG, "request for BIND \n"));
	// @@ Note, must we divide by sizeof (wchar_t)?
	ACE_WString a_name (this->name_request_.name (),
			    this->name_request_.name_len () / sizeof (wchar_t));
	ACE_WString a_value (this->name_request_.value (),
			     this->name_request_.value_len () / sizeof (wchar_t));
	if (this->naming_context_->bind (a_name, a_value, 
					 this->name_request_.type ()) == 0)
          return this->send_reply (ACE_Name_Reply::SUCCESS);
	else return this->send_reply (ACE_Name_Reply::FAILURE);
	/* NOTREACHED */
      }
    case ACE_Name_Request::REBIND:
      {
	ACE_DEBUG ((LM_DEBUG, "request for REBIND \n"));
	ACE_WString a_name (this->name_request_.name (),
			    this->name_request_.name_len () / sizeof (wchar_t));
	ACE_WString a_value (this->name_request_.value (),
			     this->name_request_.value_len () / sizeof (wchar_t));
	if (this->naming_context_->rebind (a_name, a_value, 
					this->name_request_.type ()) == 0)
          return this->send_reply (ACE_Name_Reply::SUCCESS);
	else return this->send_reply (ACE_Name_Reply::FAILURE);
	/* NOTREACHED */
      }
    case ACE_Name_Request::RESOLVE:
      {
	ACE_DEBUG ((LM_DEBUG, "request for RESOLVE \n"));
	ACE_WString a_name (this->name_request_.name (),
			    this->name_request_.name_len () / sizeof (wchar_t));

	// the following will deliver our reply back to client
	// we pre-suppose success (indicated by type RESOLVE)

	ACE_WString avalue;
	char *atype; 
	if (this->naming_context_->resolve (a_name, avalue, atype) == 0)
	  {
	    ACE_Name_Request nrq (ACE_Name_Request::RESOLVE,
				  NULL, 0,
				  avalue.rep (),
				  avalue.length () * sizeof (wchar_t),
				  "", strlen (""));
	    return this->send_request (nrq);
	  }

	ACE_Name_Request nrq (ACE_Name_Request::BIND, NULL, 0, NULL, 0, NULL, 0);
        
	this->send_request (nrq);
	return 0;
	/* NOTREACHED */
      }
    case ACE_Name_Request::UNBIND:
      {
	ACE_DEBUG ((LM_DEBUG, "request for UNBIND \n"));
	ACE_WString a_name (this->name_request_.name (),
			    this->name_request_.name_len () / sizeof (wchar_t));
	if (this->naming_context_->unbind (a_name) == 0)
          return this->send_reply (ACE_Name_Reply::SUCCESS);
	else return this->send_reply (ACE_Name_Reply::FAILURE);
	/* NOTREACHED */
      }
    case ACE_Name_Request::LISTNAMES:
      {
	ACE_DEBUG ((LM_DEBUG, "request for LISTNAMES \n"));
	ACE_Unbounded_Set<ACE_WString> myset;
	ACE_WString mypattern (this->name_request_.name (),
			       this->name_request_.name_len () / sizeof (wchar_t));
	if (naming_context_->list_names (myset, mypattern) == 0)
	  {
	    ACE_WString *one_name;

	    for (ACE_Unbounded_Set_Iterator<ACE_WString> setit (myset);
		 setit.next (one_name);
		 setit.advance ())
	      {
		ACE_Name_Request mynrq (ACE_Name_Request::LISTNAMES,
				    one_name->rep (),
				    one_name->length () * sizeof (wchar_t),
				    NULL, 0,
				    NULL, 0);
		if (this->send_request (mynrq) == -1)
		  return -1;
	      }
	  }                
	ACE_Name_Request nrq (ACE_Name_Request::LAST, NULL, 0, NULL, 0, NULL, 0);
        
	if (this->send_request (nrq) == -1)
	  return -1;
	return 0;

	/* NOTREACHED */
      }
    default:
      ACE_ERROR_RETURN ((LM_ERROR, "invalid msg_type = %d\n", 
			 this->name_request_.msg_type ()), -1);
      /* NOTREACHED */
    }
}
  
// Receive, frame, and decode the client's request.  Note, this method
// should use non-blocking I/O.

/* VIRTUAL */ int 
ACE_Name_Handler::recv_request (void)
{
  // Read the first 4 bytes to get the length of the message This
  // implementation assumes that the first 4 bytes are the length of
  // the message.
  ssize_t n = this->peer ().recv ((void *) &this->name_request_, sizeof (ACE_UINT32));

  switch (n)
    {
    case -1:
      /* FALLTHROUGH */
      ACE_DEBUG ((LM_DEBUG, "****************** recv_request returned -1\n"));
    default:
      ACE_ERROR ((LM_ERROR, "%p got %d bytes, expected %d bytes\n", 
		 "recv failed", n, sizeof (ACE_UINT32)));
      /* FALLTHROUGH */
    case 0:
      // We've shutdown unexpectedly, let's abandon the connection.
      this->abandon ();
      return -1;
      /* NOTREACHED */
    case sizeof (ACE_UINT32):
      {
	// Transform the length into host byte order.
	ssize_t length = ntohl (this->name_request_.length ());

	// Do a sanity check on the length of the message.
	if (length > sizeof this->name_request_)
	  {
	    ACE_ERROR ((LM_ERROR, "length %d too long\n", length));
	    return this->abandon ();
	  }

	// Receive the rest of the request message.
	// @@ beware of blocking read!!!.
	n = this->peer ().recv ((void *) (((char *) &this->name_request_) 
					+ sizeof (ACE_UINT32)), length);

	// Subtract off the size of the part we skipped over...
	if (n != (length - sizeof (ACE_UINT32)))
	  {
	    ACE_ERROR ((LM_ERROR, "%p expected %d, got %d\n", 
		       "invalid length", length, n));
	    return this->abandon ();
	  }

	// Decode the request into host byte order.
	if (this->name_request_.decode () == -1)
	  {
	    ACE_ERROR ((LM_ERROR, "%p\n", "decode failed"));
	    return this->abandon ();
	  }
      }
    }
  return 0;
}

// Callback method invoked by the ACE_Reactor when events arrive from
// the client.

/* VIRTUAL */ int 
ACE_Name_Handler::handle_input (ACE_HANDLE)
{

  if (this->recv_request () == -1)
    return -1;
  else
    return this->dispatch ();
}

ACE_Name_Handler::~ACE_Name_Handler (void)
{
  ACE_DEBUG ((LM_DEBUG, "closing down Handle  %d\n",
	      this->get_handle ()));

  // delete associated Naming Context
  delete this->naming_context_;

  if (this->peer ().close () == -1)
    ACE_DEBUG ((LM_DEBUG, "closing down Handle  failed\n"));
}
