/* -*- C++ -*- */
// ============================================================================
//
// = LIBRARY
//    ACE
// 
// = FILENAME
//    ACE_Name_Handler.h
//
// = DESCRIPTION 
//    Product object created by <ACE_Name_Acceptor>.  A <ACE_Name_Handler>
//    exchanges messages with a <ACE_Name_Proxy> object on the
//    client-side.
//
// = AUTHOR
//    Gerhard Lenzer and Douglas C. Schmidt
// 
// ============================================================================

#if !defined (ACE_NAME_HANDLER_H)
#define ACE_NAME_HANDLER_H

#include "ace/Svc_Handler.h"
#include "ace/SOCK_Stream.h"
#include "ace/Name_Request_Reply.h"

// Forward declaration.
class ACE_Naming_Context;

class ACE_Name_Handler : public ACE_Svc_Handler<ACE_SOCK_STREAM, ACE_NULL_SYNCH>
  // = TITLE
  //    Product object created by <ACE_Name_Acceptor>.  A <ACE_Name_Handler>
  //    exchanges messages with a <ACE_Name_Proxy> object on the
  //    client-side.
  //
  // = DESCRIPTION
  //   This class is the main workhorse of the <ACE_Name_Server>.  It
  //   handles client requests to bind, rebind, resolve, and unbind
  //   names.  It also schedules and handles timeouts that are used to
  //   support "timed waits."  Clients used timed waits to bound the
  //   amount of time they block trying to get a name.
{
public:
  // = Initialization and termination.

  ACE_Name_Handler (ACE_Reactor * = 0);
  // Default constructor.

  virtual int open (void *acceptor = 0);
  // Activate this instance of the <ACE_Name_Handler> (called by the
  // <ACE_Name_Acceptor>).  Note that the <void *acceptor> is *really*
  // an ACE_Acceptor but we need to keep the compiler happy with the
  // type signature of the open() method inherited from Svc_Handler...

  virtual int close (u_long = 0);
  // Close down this instance of the <ACE_Name_Handler>.

protected:
  // = Helper routines for the operations exported to clients.

  virtual int abandon (void);
  // Give up waiting (e.g., when a timeout occurs or a client shuts
  // down unexpectedly).

  // = Low level routines for framing requests, dispatching operations, 
  // and returning replies.

  virtual int recv_request (void);
  // Receive, frame, and decode the client's request.
  
  virtual int dispatch (void);
  // Dispatch the appropriate operation to handle the client's request.
  
  virtual int send_reply (ACE_UINT32 status, ACE_UINT32 errnum = 0);
  // Create and send a reply to the client.

  virtual int send_request(ACE_Name_Request&);
  // special kind of reply

  // = Demultiplexing hooks.
  virtual ACE_HANDLE get_handle (void) const;
  // Return the underlying <ACE_HANDLE>.

  virtual int handle_input (ACE_HANDLE);
  // Callback method invoked by the <ACE_Reactor> when client events arrive.

  // = Timer hook.
  virtual int handle_timeout (const ACE_Time_Value &tv, const void *arg);
  // Enable clients to limit the amount of time they wait for a name.

private:

  ACE_Naming_Context *naming_context_;
  // ACE_Naming_Context of this Handler

  ACE_Name_Request name_request_;
  // Cache request from the client.

  ACE_Name_Request name_request_back_;
  // special kind of reply for resolve and listnames

  ACE_Name_Reply name_reply_;
  // Cache reply to the client.

  ACE_INET_Addr addr_;
  // Address of client we are connected with.

  ~ACE_Name_Handler (void);
  // Ensure dynamic allocation...
};
#endif /* ACE_NAME_HANDLER_H */
