#include "ace/Read_Buffer.h"
#include "ace/Service_Config.h"

ACE_Read_Buffer::ACE_Read_Buffer (FILE *fp, 
				  ACE_Allocator *allocator)
  : stream_ (fp), 
    allocator_ (allocator)
{
  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();
}

ACE_Read_Buffer::ACE_Read_Buffer (int handle,
				  ACE_Allocator *allocator)
  : stream_ (ACE_OS::fdopen (handle, "r")),
    allocator_ (allocator)
{
  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();
}

ACE_Read_Buffer::~ACE_Read_Buffer (void)
{
  ACE_OS::fclose (this->stream_);
}

// Input: term        the character to terminate on
//        search      the character to search for
//	  replace     the character with which to replace search
// Output: a buffer containing the contents of stream
// Method: call the recursive helper function read_helper

char *
ACE_Read_Buffer::read (int term, int search, int replace)
{
  this->occurrences_ = 0;
  this->size_ = 0;
  return this->rec_read (term, search, replace);
}

// Accessor to the number of bytes in the buffer.

size_t
ACE_Read_Buffer::size (void) const
{
  return this->size_;
}

// The total number of characters replaced.

size_t
ACE_Read_Buffer::replaced (void) const
{
  return this->occurrences_;
}

// Input: term       the termination character
//        search     the character to search for
//        replace    the character with which to replace search
// Purpose: read in a file to a buffer using only a single dynamic
//          allocation.
// Method: read until the local buffer is full and then recurse.
//         Must continue until the termination character is reached. 
//         Allocate the final buffer based on the number of local
//         buffers read and as the recursive calls bottom out, 
//         copy them in reverse order into the allocated buffer.

char *
ACE_Read_Buffer::rec_read(int term, int search, int replace)
{
  // This is our temporary workspace.
  char buf[BUFSIZ];

  int c;
  size_t index = 0;
  int done = 0;
  
  // Read in the file char by char
  while (index < BUFSIZ)
    {
      c = getc (this->stream_);

      // Don't insert EOF into the buffer...
      if (c == EOF)
	{
	  if (index == 0)
	    return 0;
	  else
	    {
	      ungetc (c, this->stream_);
	      break;
	    }
	}
      else if (c == term)
	done = 1;

      // Check for possible substitutions.
      if (c == search) 
	{
	  this->occurrences_++;

	  if (replace >= 0)
	    c = replace;
	}

      buf[index++] = c;

      // Substitutions must be made before checking for termination. 
      if (done)
	break;
    }

  // Increment the number of bytes.
  this->size_ += index;

  char *result;

  // Recurse, when the recursion bottoms out, allocate the result
  // buffer.
  if (done || c == EOF)
    {
      // Use the allocator to acquire the memory.
      result = (char *) this->allocator_->malloc (this->size_ * sizeof (char));

      if (result == 0)
	{
	  errno = ENOMEM;
	  return 0;
	}
      result += this->size_;
    }
  else
    result = this->rec_read (term, search, replace);
  
  // Copy buf into the appropriate location starting from end of
  // buffer.

  for (size_t j = index; j > 0; j--)
    *--result = buf[j - 1];

  return result;
}
