/* -*- C++ -*- */

// ============================================================================
//
// = LIBRARY
//    ACE
// 
// = FILENAME
//    Remote_Tokens.h
//
// = AUTHOR
//    Douglas C. Schmidt (schmidt@cs.wustl.edu) and
//    Tim Harrison (harrison@cs.wustl.edu)
// 
// ============================================================================

#if !defined (ACE_Remote_Mutex_H)
#define ACE_Remote_Mutex_H

#include "ace/INET_Addr.h"
#include "ace/SOCK_Connector.h"
#include "ace/SOCK_Stream.h"
#include "ace/Synch_Options.h"
#include "ace/Local_Tokens.h"
#include "ace/Token_Request_Reply.h"

class ACE_Remote_Token_Proxy : public ACE_Token_Proxy
// = TITLE
//   Proxy for acquiring, renewing, and releasing a distributed
//   synchronization token.

// = DESCRIPTION
//    The Remote_Mutex class offers methods for acquiring, renewing,
//    and releasing a distributed synchronization mutex.  Similar to
//    ACE_Local_Mutex, ACE_Remote_Token_Proxy offers recursive acquisition,
//    FIFO waiter ordering, and deadlock detection.  All instances of
//    ACE_Remote_Token_Proxy use ACE_Remote_Token_Proxy::connection_repository_
//    for optimal connection reuse.

// = BUGS
//   Distributed sleep_hooks have not been implemented.  owner_id ()
//   is not implemented.
{
public:
  ACE_Remote_Token_Proxy (ACE_Token_Proxy::PROXY_TYPE proxy_type,
			  ACE_Tokens::TOKEN_TYPE token_type);
  // If you use this, you must call ACE_Remote_Token_Proxy::open()!!!
  // open is inherited from ACE_Token_Proxy

  ACE_Remote_Token_Proxy (const ACE_Remote_Token_Proxy &mutex);

  virtual ~ACE_Remote_Token_Proxy (void);
  // death

  ACE_Remote_Token_Proxy (const char *token_name,
			  int ignore_deadlock,
			  int debug,
			  ACE_Token_Proxy::PROXY_TYPE proxy_type,
			  ACE_Tokens::TOKEN_TYPE token_type);

  // = Operations exported by the ACE_Token Server to the client.

  int initiate_connection (void);
  // Open a connection with the token server.  This only need be used
  // when the user wishes to explicitly open a connection to check if
  // the server exists.  If a connection is established, it is
  // returned to the connection repository.  Currently, this initiates
  // a BLOCKING connection.  Calling initiate_connection guarantees
  // that the calling thread already has an open connection for future
  // token operations.

  virtual int acquire (int notify = 0,
		       void (*sleep_hook)(void *) = 0,
		       ACE_Synch_Options &options =
		       ACE_Synch_Options::synch);
  // Acquire the distributed mutex.  Connections are obtained on an
  // as-needed basis from the connection repository.  If notify is
  // specified and the token is already held, the owner is notified.
  // options contains the timeout value for the acquire call.  The
  // timer is kept at the token server.  Asynchronous operations are
  // not supported.  Returns 0 on success, -1 on failure with <errno>
  // == problem.

  virtual int tryacquire (void (*sleep_hook)(void *) = 0);
  // Try to acquire the distributed mutex.  If the token is already
  // held, the call returns without queueing the caller as a waiter.
  // Returns 0 on success (the token was acquired), and -1 with
  // EWOULDBLOCK if the token was already held.

  virtual int renew (int requeue_position = 0,
		     ACE_Synch_Options &options =
		     ACE_Synch_Options::synch);
  // Renew the token by offering to release it if there are any other
  // waiters, otherwise get the token back immediately.  This renew
  // has the same semantics as ACE_Local_Mutex release.  It is
  // semantically equivalent to <this->release()> followed by
  // <this->acquire()>, but it is faster.  options contains the
  // timeout value used if renew blocks.  As with acquire, the timer
  // is maintained at the token server.  If there are waiters and
  // requeue_position == -1, the caller is queued at the rear of the
  // waiter list.  Otherwise, requeue_position specifies the number of
  // waiters to "let by" before reacquiring the token (effectively,
  // the position in the waiter list.)

  virtual int release (ACE_Synch_Options &options =
		       ACE_Synch_Options::synch);
  // Release the distributed mutex. Similar to ACE_Local_Mutex, if the
  // caller is not the owner, it is removed from the waiter list (if
  // applicable.) Returns 0 on success, -1 on failure with <errno> ==
  // problem.

  virtual int remove (ACE_Synch_Options &options =
		      ACE_Synch_Options::synch);
  // Become interface compliant for ACE_Guard<>.  This has no
  // functionality.

  virtual ACE_Token_Proxy* clone (void);
  // Returns a dynamically allocated clone of the derived class.
  // Useful in ACE_Token_Collection for cloning a derived class from
  // the abstract interface of ACE_Token_Proxy.

  virtual void token_acquired (void*);
  // Override the default to do nothing.

  virtual const char* owner_id (void);
  // the client id of the current token holder

  static void set_server_address (const ACE_INET_Addr &server_address);
  // sets the server address for all instances of ACE_Remote_Token_Proxy
  // If this isn't called, the environment variable TOKEN_SERVER is
  // checked for the server address.  If that is not specified, all
  // ACE_Remote_** operations will fail.
protected:

  int ignore_shadow_deadlock_;
  // if shadows report deadlock, go remote anyway

  int request_reply (ACE_Token_Request &request,
		     ACE_Synch_Options &options);
  // Perform the request and wait for the reply.
};

class ACE_Remote_Mutex : public ACE_Remote_Token_Proxy
// = TITLE
//   Proxy for acquiring, renewing, and releasing a distributed
//   mutex.

// = DESCRIPTION
//    The Remote_Mutex class offers methods for acquiring, renewing,
//    and releasing a distributed synchronization mutex.  Similar to
//    ACE_Local_Mutex, ACE_Remote_Token_Proxy offers recursive acquisition,
//    FIFO waiter ordering, and deadlock detection.  All instances of
//    ACE_Remote_Token_Proxy use ACE_Remote_Token_Proxy::connection_repository_
//    for optimal connection reuse.
{
public:
  ACE_Remote_Mutex (void);

  ACE_Remote_Mutex (const char *token_name,
		    int ignore_deadlock = 0,
		    int debug = 0);

  ACE_Remote_Mutex (const ACE_Remote_Mutex &mutex);
};

class ACE_Remote_RLock : public ACE_Remote_Token_Proxy
// = TITLE
//   Proxy for acquiring, renewing, and releasing a distributed
//   readers lock.

// = DESCRIPTION
//   Shields applications from details of interacting with the
//   ACE_Token_Server.  The token_name_ is just the string that the
//   Token Server uses to identify the token.  The client_id_ (also
//   used by the Token Server,) identifies the owner of the token and
//   is used for deadlock detection.

{
public:
  ACE_Remote_RLock (void);

  ACE_Remote_RLock (const char *token_name,
		    int ignore_deadlock = 0,
		    int debug = 0);

  ACE_Remote_RLock (const ACE_Remote_RLock &mutex);
};

class ACE_Remote_WLock : public ACE_Remote_Token_Proxy
// = TITLE
//   Proxy for acquiring, renewing, and releasing a distributed
//   writers lock.

// = DESCRIPTION
//   Shields applications from details of interacting with the
//   ACE_Token_Server.  The token_name_ is just the string that the
//   Token Server uses to identify the token.  The client_id_ (also
//   used by the Token Server,) identifies the owner of the token and
//   is used for deadlock detection.

{
public:
  ACE_Remote_WLock (void);

  ACE_Remote_WLock (const char *token_name,
		    int ignore_deadlock = 0,
		    int debug = 0);

  ACE_Remote_WLock (const ACE_Remote_WLock &mutex);
};

#if defined (__INLINE__)
#include "ace/Remote_Tokens.i"
#endif /* __INLINE__ */

#endif /* ACE_Remote_Token_H */
