#include "ace/Malloc.h"
#include "ace/Service_Config.h"
#include "ace/SString.h"

#if !defined (__INLINE__)
#include "ace/SString.i"
#endif /* __INLINE__ */

ACE_ALLOC_HOOK_DEFINE(ACE_CString)

// Copy constructor.

ACE_CString::ACE_CString (const ACE_CString &s)
  : len_ (s.len_),
    allocator_ (s.allocator_)
     
{
  ACE_TRACE ("ACE_CString::ACE_CString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();

  this->rep_ = (char *) this->allocator_->malloc (s.len_ + 1);
  ACE_OS::memcpy ((void *) this->rep_, (const void *) s.rep_, this->len_);
  this->rep_[this->len_] = 0;
}

void
ACE_CString::dump (void) const
{
  ACE_TRACE ("ACE_CString::dump");
}

ACE_CString::~ACE_CString (void)
{
  ACE_TRACE ("ACE_CString::~ACE_CString");
  this->allocator_->free (this->rep_);
}

size_t
ACE_CString::length (void) const
{
  ACE_TRACE ("ACE_CString::length");
  return this->len_;
}

// Default constructor.

ACE_CString::ACE_CString (ACE_Allocator *allocator)
  : len_ (0), rep_ (0),
    allocator_ (allocator)
{
  ACE_TRACE ("ACE_CString::ACE_CString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();
}

// Constructor that actually copies memory.

ACE_CString::ACE_CString (const char *s, ACE_Allocator *allocator)
  : allocator_ (allocator)
{
  ACE_TRACE ("ACE_CString::ACE_CString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();

  if (s == 0)
    {
      this->len_ = 0;
      this->rep_ = 0;
    }
  else
    {
      this->len_ = ACE_OS::strlen (s);
      this->rep_ = (char *) this->allocator_->malloc (this->len_ + 1);
      ACE_OS::strcpy (this->rep_, s);
    }
}

// Constructor that actually copies memory.

ACE_CString::ACE_CString (const char *s, 
			    size_t len, 
			    ACE_Allocator *allocator)
  : allocator_ (allocator)
{
  ACE_TRACE ("ACE_CString::ACE_CString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();

  if (s == 0)
    {
      this->len_ = 0;
      this->rep_ = 0;
    }
  else
    {
      this->len_ = len;
      this->rep_ = (char *) this->allocator_->malloc (this->len_ + 1);
      ACE_OS::strncpy (this->rep_, s, len);
      this->rep_[len] = '\0'; // Make sure to NUL terminate this!
    }
}

// Assignment operator (does copy memory).

void 
ACE_CString::operator = (const ACE_CString &s)
{
  ACE_TRACE ("ACE_CString::operator =");
  // Check for identify.

  if (this != &s)
    {
      // Only reallocate if we don't have enough space...
      if (this->len_ < s.len_)
	{
	  this->allocator_->free (this->rep_);
	  this->rep_ = (char *) this->allocator_->malloc (s.len_ + 1);
	}
      this->len_ = s.len_;
      ACE_OS::strcpy (this->rep_, s.rep_);
    }
}

// Concat operator (does copy memory).

void 
ACE_CString::operator += (const ACE_CString &s)
{
  ACE_TRACE ("ACE_CString::operator +=");

  char *t = (char *) this->allocator_->malloc (this->len_ + s.len_ + 1);
  ACE_OS::memcpy (t, this->rep_, this->len_);
  ACE_OS::memcpy (t + this->len_, s.rep_, s.len_);
  this->len_ += s.len_;
  t[this->len_] = '\0';
  this->allocator_->free (this->rep_);
  this->rep_ = t;
}

ACE_ALLOC_HOOK_DEFINE(ACE_SString)

void
ACE_SString::dump (void) const
{
  ACE_TRACE ("ACE_SString::dump");
}

size_t
ACE_SString::length (void) const
{
  ACE_TRACE ("ACE_SString::length");
  return this->len_;
}

// Copy constructor.

ACE_SString::ACE_SString (const ACE_SString &s)
  : len_ (s.len_),
    allocator_ (s.allocator_)
{
  ACE_TRACE ("ACE_SString::ACE_SString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();

  this->rep_ = (char *) this->allocator_->malloc (s.len_ + 1);
  ACE_OS::memcpy ((void *) this->rep_, (const void *) s.rep_, this->len_);
  this->rep_[this->len_] = 0;
}

// Default constructor.

ACE_SString::ACE_SString (ACE_Allocator *allocator)
  : allocator_ (allocator), 
    len_ (0), rep_ (0)
{
  ACE_TRACE ("ACE_SString::ACE_SString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();
}

// Set the underlying pointer (does not copy memory).

void 
ACE_SString::rep (char *s)
{
  ACE_TRACE ("ACE_SString::rep");

  this->rep_ = s;

  if (s == 0)
    this->len_ = 0;
  else
    this->len_ = ACE_OS::strlen (s);
}

// Constructor that actually copies memory.

ACE_SString::ACE_SString (const char *s, 
			    ACE_Allocator *allocator)
  : allocator_ (allocator)
{
  ACE_TRACE ("ACE_SString::ACE_SString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();

  if (s == 0)
    {
      this->len_ = 0;
      this->rep_ = 0;
    }
  else
    {
      this->len_ = ACE_OS::strlen (s);
      this->rep_ = (char *) this->allocator_->malloc (this->len_ + 1);
      ACE_OS::strcpy (this->rep_, s);
    }
}

// Constructor that actually copies memory.

ACE_SString::ACE_SString (const char *s, 
			    size_t len, 
			    ACE_Allocator *allocator)
  : allocator_ (allocator)
{
  ACE_TRACE ("ACE_SString::ACE_SString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();

  if (s == 0)
    {
      this->len_ = 0;
      this->rep_ = 0;
    }
  else
    {
      this->len_ = len;
      this->rep_ = (char *) this->allocator_->malloc (this->len_ + 1);
      ACE_OS::strncpy (this->rep_, s, len);
      this->rep_[len] = '\0'; // Make sure to NUL terminate this!
    }
}

// Assignment operator (does copy memory).

void 
ACE_SString::operator = (const ACE_SString &s)
{
  ACE_TRACE ("ACE_SString::operator =");
  // Check for identify.

  if (this != &s)
    {
      // Only reallocate if we don't have enough space...
      if (this->len_ < s.len_)
	{
	  this->allocator_->free (this->rep_);
	  this->rep_ = (char *) this->allocator_->malloc (s.len_ + 1);
	}
      this->len_ = s.len_;
      ACE_OS::strcpy (this->rep_, s.rep_);
    }
}

ACE_ALLOC_HOOK_DEFINE(ACE_WString)

void
ACE_WString::dump (void) const
{
  ACE_TRACE ("ACE_WString::dump");
}

// Default constructor.

ACE_WString::ACE_WString (ACE_Allocator *allocator)
  : len_ (0), 
    rep_ (0),
    allocator_ (allocator)
{
  ACE_TRACE ("ACE_WString::ACE_WString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();
}

size_t
ACE_WString::wstrlen (const wchar_t *s)
{
  ACE_TRACE ("ACE_WString::wstrlen");
  int i;

  for (i = 0; s[i] != 0; i++)
    continue;
  
  return i;
}

// Get the underlying pointer as an ASCII char.

char *
ACE_WString::char_rep (void) const
{
  ACE_TRACE ("ACE_WString::char_rep");
  if (this->len_ <= 0)
    return 0;
  else
    {
      char *t = new char[this->len_ + 1];

      for (size_t i = 0; i < this->len_; i++)
	// Note that this cast may lose data if wide chars are
	// actually used!
	t[i] = char (this->rep_[i]);

      t[this->len_] = '\0';
      return t;
    }
}

// Constructor that actually copies memory.

ACE_WString::ACE_WString (const char *s, 
			  ACE_Allocator *allocator)
  : allocator_ (allocator)
{
  ACE_TRACE ("ACE_WString::ACE_WString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();

  if (s == 0)
    {
      this->len_ = 0;
      this->rep_ = 0;
    }
  else
    {
      this->len_ = ACE_OS::strlen (s);
      this->rep_ = (wchar_t *) this->allocator_->malloc ((this->len_ + 1) * sizeof (wchar_t));

      // Copy the char * string byte-by-byte into the wchar_t *
      // string. 
      for (size_t i = 0; i < this->len_; i++)
	this->rep_[i] = s[i];

      // null terminate 
      this->rep_[this->len_] = 0;
    }
}

// Constructor that actually copies memory.

ACE_WString::ACE_WString (const wchar_t *s, ACE_Allocator *allocator)
  : allocator_ (allocator)
{
  ACE_TRACE ("ACE_WString::ACE_WString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();

  if (s == 0)
    {
      this->len_ = 0;
      this->rep_ = 0;
    }
  else
    {
      this->len_ = this->wstrlen (s);
      this->rep_ = (wchar_t *) this->allocator_->malloc ((this->len_ + 1) * sizeof (wchar_t));

      ACE_OS::memcpy (this->rep_, s, this->len_ * sizeof (wchar_t));

      // null terminate 
      this->rep_[this->len_] = 0;
    }
}

// Constructor that actually copies memory.

ACE_WString::ACE_WString (const wchar_t *s, 
			  size_t len, 
			  ACE_Allocator *allocator)
  : allocator_ (allocator)
{
  ACE_TRACE ("ACE_WString::ACE_WString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();

  if (s == 0)
    {
      this->len_ = 0;
      this->rep_ = 0;
    }
  else
    {
      this->len_ = len;
      this->rep_ = (wchar_t *) this->allocator_->malloc ((this->len_ + 1) * sizeof (wchar_t));

      ACE_OS::memcpy (this->rep_, s, len * sizeof (wchar_t));

      // null terminate 
      this->rep_[this->len_] = 0;
    }
}

size_t
ACE_WString::length (void) const
{
  ACE_TRACE ("ACE_WString::length");
  return this->len_;
}

// Copy constructor.

ACE_WString::ACE_WString (const ACE_WString &s)
: len_ (s.len_),
  allocator_ (s.allocator_)
{
  ACE_TRACE ("ACE_WString::ACE_WString");

  if (this->allocator_ == 0)
    this->allocator_ = ACE_Service_Config::allocator ();

  this->rep_ = (wchar_t *) this->allocator_->malloc ((s.len_ + 1) * sizeof (wchar_t));
  ACE_OS::memcpy ((void *) this->rep_, (const void *) s.rep_,
		  this->len_ * sizeof (wchar_t));
  this->rep_[this->len_] = 0;
}

// Assignment operator (does copy memory).

void 
ACE_WString::operator = (const ACE_WString &s)
{
  ACE_TRACE ("ACE_WString::operator =");
  // Check for identify.

  if (this != &s)
    {
      // Only reallocate if we don't have enough space...
      if (this->len_ < s.len_)
	{
	  this->allocator_->free (this->rep_);
	  this->rep_ = (wchar_t *) this->allocator_->malloc ((s.len_ + 1) * sizeof (wchar_t));
	  // null terminate 
	  this->rep_[s.len_] = 0;
	}

      this->len_ = s.len_;
      ACE_OS::memcpy ((void *) this->rep_, (const void *) s.rep_,
		      this->len_ * sizeof (wchar_t));
    }
}

// Concat operator (does copy memory).

void 
ACE_WString::operator += (const ACE_WString &s)
{
  ACE_TRACE ("ACE_WString::operator +=");
  wchar_t *t = (wchar_t *) this->allocator_->malloc ((this->len_ + s.len_ + 1) * sizeof (wchar_t));

  ACE_OS::memcpy ((void *) t, (const void *) this->rep_, this->len_ * sizeof (wchar_t));
  ACE_OS::memcpy ((void *) (t + this->len_ * sizeof (wchar_t)), 
	    (const void *) s.rep_, s.len_ * sizeof (wchar_t));
  this->len_ += s.len_;

  // null terminate 
  t[this->len_] = 0;

  this->allocator_->free (this->rep_);
  this->rep_ = t;
}

ACE_WString::~ACE_WString (void)
{
  ACE_TRACE ("ACE_WString::~ACE_WString");
  this->allocator_->free (this->rep_);
}
