/* -*- C++ -*- */

// ============================================================================
//
// = LIBRARY
//    ace
// 
// = FILENAME
//    Service_Config.h
//
// = AUTHOR
//    Doug Schmidt 
// 
// ============================================================================

#if !defined (ACE_SERVICE_CONFIG_H)
#define ACE_SERVICE_CONFIG_H

#include "ace/Reactor.h"
#include "ace/Service_Object.h"
#include "ace/Thread_Manager.h"

// Forward decl.
class ACE_Service_Repository;
class ACE_Service_Record;
class ACE_Allocator;

struct ACE_Static_Svc_Descriptor
{
  char *name_;
  // Name of the service.
  
  int type_;
  // Type of service.

  ACE_Service_Object *(*alloc_)(void);
  // Factory function that allocates the service.
  
  u_int flags_;
  // Bitmask flags indicating how the framework should delete memory.

  int active_;
  // Flag indicating whether the service starts out active.

  void dump (void) const;
  // Dump the state of an object.

  ACE_ALLOC_HOOK_DECLARE;
  // Declare the dynamic allocation hooks.
};

class ACE_Service_Config : public ACE_Event_Handler
  // = TITLE
  //     Provide the abstract base class that supplies 
  // common server daemon operations.
  //
  // = DESCRIPTION
  //
{
public:
  enum { MAX_SERVICES = ACE_Reactor::DEFAULT_SIZE };

  // = Initialization and termination methods.

  ACE_Service_Config (int ignore_defaults = 0, 
		      size_t size = ACE_Service_Config::MAX_SERVICES, 
		      int signum = SIGHUP);
  // Initialize the Service Repository.

  ACE_Service_Config (const char program_name[]);
  // Performs an open without parsing command-line arguments.

  static int open (int argc, char *argv[]);
  // This is the primary entry point into the ACE_Service_Config (the
  // constructor just handles simple initializations).  It 
  // parses arguments passed in from the command-line.

  static int open (const char program_name[]);
  // Performs an open without parsing command-line arguments.

  virtual ~ACE_Service_Config (void);
  // Perform user-specified close activities and remove dynamic memory. 

  static int close (void);
  // Tidy up and perform last rites on a terminating ACE_Service_Config.

  // = Event loop management methods.
  static int run_event_loop (void);
  // Run the event loop until the <ACE_Reactor::handle_events>
  // method returns -1 or the <end_event_loop> method
  // is invoked.

  static int run_event_loop (ACE_Time_Value &tv);
  // Run the event loop until the <ACE_Reactor::handle_events>
  // method returns -1, the <end_event_loop> method
  // is invoked, or the <ACE_Time_Value> expires.

  static void end_event_loop (sig_atomic_t value = 1);
  // Instruct the <ACE_Service_Config> to terminate its event loop if
  // <value> != 0.  Else, if <value> == 0 then don't terminate.

  // = The following methods are static in order to enforce Singleton
  // semantics for the Reactor, Service_Repository, Thread_Manager,
  // and Acceptor/Connector Strategy factory.  Other portions of the
  // system may need to access them at some point or another...

  // = Accessors and mutators for process-wide Singletons.

  static ACE_Reactor *reactor (void);
  // Get pointer to a process-wide <ACE_Reactor>.

  static ACE_Reactor *reactor (ACE_Reactor *);
  // Set pointer to a process-wide <ACE_Reactor> and return existing
  // pointer.

  static ACE_Service_Repository *svc_rep (void);
  // Get pointer to a process-wide <ACE_Service_Repository>.

  static ACE_Service_Repository *svc_rep (ACE_Service_Repository *);
  // Set pointer to a process-wide <ACE_Service_Repository> and return
  // existing pointer.  

  static ACE_Thread_Manager *thr_mgr (void);
  // Get pointer to a process-wide <ACE_Thread_Manager>.

  static ACE_Thread_Manager *thr_mgr (ACE_Thread_Manager *);
  // Set pointer to a process-wide <ACE_Thread_Manager> and return
  // existing pointer.

  static ACE_Allocator *allocator (void);
  // Get pointer to a default <ACE_Allocator>.

  static ACE_Allocator *allocator (ACE_Allocator *);
  // Set pointer to a process-wide <ACE_Allocator> and return existing
  // pointer.

  // = Member functions used by various other parts
  //   of the Service Configurator class category.
  static int initialize (const ACE_Service_Record *, char parameters[]);
  // Dynamically link the shared object file and retrieve
  // a pointer to the designated shared object in this file. 

  static int initialize (const char svc_name[], char parameters[]);
  // Initialize and activate a statically <svc_name> service. 

  static int resume (const char svc_name[]);
  // Resume a <svc_name> that was previously suspended or has not yet
  // been resumed (e.g., a static service). 

  static int suspend (const char svc_name[]);
  // Suspend <svc_name>.  Note that this will not unlink the service 
  // from the daemon if it was dynamically linked, it will mark it 
  // as being suspended in the Service Repository and call the
  // suspend() member function on the appropriate ACE_Service_Object.  
  // A service can be resumed later on by calling the RESUME() 
  // member function... 

  static int remove (const char svc_name[]);
  // Totally remove <svc_name> from the daemon by removing it 
  // from the ACE_Reactor, and unlinking it if necessary.

  void dump (void) const;
  // Dump the state of an object.

  ACE_ALLOC_HOOK_DECLARE;
  // Declare the dynamic allocation hooks.

protected:
  static int process_directives (void);
  // Process service configuration requests as indicated 
  // in the <service_config_file>.

  static void parse_args (int, char *[]);
  // Handle the command-line options intended for the <ACE_Service_Config>.

  static void reconfigure (void);
  // Perform the reconfiguration process.

  static int start_daemon (void);
  // Become a daemon. 

  static int load_defaults (void);
  // Add the default statically-linked services to the <ACE_Service_Repository>.

  virtual int handle_signal (int signum
#if defined (ACE_HAS_SIGINFO_T)
, siginfo_t * = 0, ucontext_t * = 0
#endif /* ACE_HAS_SIGINFO_T */
  );
  // Signal handling API to trigger dynamic reconfiguration.

private:
  static ACE_Reactor *reactor_;
  // Pointer to a process-wide <ACE_Reactor>.

  static ACE_Service_Repository *svc_rep_;
  // Pointer to a process-wide <ACE_Service_Repository>.

  static ACE_Thread_Manager *thr_mgr_;
  // Pointer to a process-wide <ACE_Thread_Manager>.

  static ACE_Allocator *allocator_;
  // Pointer to a process-wide <ACE_Allocator> instance.

  static const char *service_config_file_;
  // Name of service configuration file.

  static const char *logger_key_;
  // Where to write the logging output.

  static ACE_Static_Svc_Descriptor service_list_[];
  // List of statically linked services.

  static sig_atomic_t end_event_loop_;
  // Terminate the event loop.

  static sig_atomic_t reconfig_occurred_;
  // True if reconfiguration occurred.

  // = Set by command-line options. 
  static char debug_;
  static char be_a_daemon_;
  static char no_defaults_;

  static int signum_;
  // Number of the signal used to trigger reconfiguration.
};

#if defined (__INLINE__)
#include "ace/Service_Config.i"
#endif /* __INLINE__ */

#endif /* ACE_SERVICE_CONFIG_H */
