#include "ace/Log_Msg.h"
#include "ace/Token_Manager.h"

#if !defined (__INLINE__)
#include "ace/Token_Manager.i"
#endif /* __INLINE__ */

// singleton token manager
ACE_Token_Manager *ACE_Token_Manager::token_manager_ = 0;

ACE_TOKEN_CONST::MUTEX ACE_Token_Manager::creation_lock_;

ACE_Token_Manager::ACE_Token_Manager ()
{
  ACE_TRACE ("ACE_Token_Manager::ACE_Token_Manager");
}

ACE_Token_Manager::~ACE_Token_Manager ()
{
  ACE_TRACE ("ACE_Token_Manager::~ACE_Token_Manager");
  COLLECTION_ITERATOR iterator (collection_);
  for (COLLECTION_ENTRY *temp = 0;
       iterator.next (temp) != 0;
       iterator.advance ())
    {
      // @ should I be doing an unbind here?
      delete temp->int_id_;
      // The ext_id_'s delete themselves when the array of
      // COLLECTION_ENTRYs goes away.
    }
}

ACE_Token_Manager *
ACE_Token_Manager::instance (void)
{
  ACE_TRACE ("ACE_Token_Manager::token_manager");

  // This first check is to avoid acquiring the mutex in the common
  // case.  Double-Check pattern rules.
  if (token_manager_ == 0)
    {
      ACE_TOKEN_CONST::GUARD m (ACE_Token_Manager::creation_lock_);
      if (token_manager_ == 0)
	token_manager_ = new ACE_Token_Manager;
    }  

  return token_manager_;
}

void
ACE_Token_Manager::get_mutex (const TOKEN_NAME &name,
			      ACE_Tokens *&token)
{
  ACE_TRACE ("ACE_Token_Manager::get");
  // Hmm.  I think this makes sense.  We perform our own locking here
  // (see safe_acquire.)  We have to make sure that only one thread
  // uses the collection at a time.
  ACE_TOKEN_CONST::GUARD m (this->lock_);
  
  if (collection_.find (name, token) == -1)
    // we did not find one in the collection
    {
      // make one
      token = this->make_mutex (name);
      // put it in the collection
      if (collection_.bind (name, token) == -1)
	{
	  delete token;
	  token = 0;
	}
    }

  if (token != 0)
    token->inc_reference ();
  // we may be returning token == 0 if new failed, caller must check. 
}

ACE_Tokens *
ACE_Token_Manager::make_mutex (const TOKEN_NAME &name)
{
  ACE_Tokens *token;
  switch (name.type ())
    {
    case ACE_Tokens::MUTEX :
      token = (ACE_Tokens*) new ACE_Mutex_Token (name.name ());
      break;			  
    case ACE_Tokens::RWLOCK :
      token = (ACE_Tokens*) new ACE_RW_Token (name.name ());
      break;
    }

  return token;
}

// returns 0 if the acquire will _not_ cause deadlock.
// returns 1 if the acquire _will_ cause deadlock.
//
// If a Client tries to acquire a token which is already held, before
// entering the waiter list and blocking, the following algorithm is
// run:
// 1. OWNER = owner of token.
// 2. TOKEN = OWNER->token_waiting_for ().
// 3. if TOKEN == null
//       Become waiter and exit algorithm.
// 4. OWNER = TOKEN->owner ()
// 5. if OWNER == Client
//	 Report deadlock situation and exit algorithm.
// 6. goto step 2.
//
// Assumptions:  Clients can be in only one waiter list.
int 
ACE_Token_Manager::check_deadlock (ACE_Token_Proxy *proxy)
{
  ACE_TRACE ("ACE_Token_Manager::safe_acquire");
  // We perform no locking here.  It is up to the caller to use our
  // mutex to ensure atomicity!!!

  const char *requestor_id = proxy->client_id ();
  // 1. OWNER = owner of token.
  const char *owner_id = proxy->mutex_->owner_id ();

  assert (owner_id != 0);

  ACE_Tokens *TOKEN;

  while (1)
    {
      // 2. TOKEN = OWNER->token_waiting_for ().
      TOKEN = this->token_waiting_for (owner_id);

      // 3. if TOKEN == null Become waiter and exit algorithm.
      if (TOKEN == 0)
	// no deadlock, return.
	return 0;

      // 4. OWNER = TOKEN->owner ()
      owner_id = TOKEN->owner_id ();

      // 5. if OWNER == Client Report deadlock situation and exit
      //		algorithm.
      if (ACE_OS::strcmp (owner_id, requestor_id) == 0)
	// deadlock situation.
	return 1;
      // 6. goto step 2.
    }

}

ACE_Tokens *
ACE_Token_Manager::token_waiting_for (const char *client_id)
{
  COLLECTION_ITERATOR iterator (collection_);
  for (COLLECTION_ENTRY *temp = 0;
       iterator.next (temp) != 0;
       iterator.advance ())
    {
      if (ACE_OS::strcmp (client_id, temp->int_id_->owner_id ()) == 0)
	continue;
      else
	{
	  if (temp->int_id_->waiters ()->member (client_id) == 1)
	    // client_id is waiting for this token.
	    return temp->int_id_;
	}
    }
  // nothing was found, return NULL.
  return 0;
}


// notify the token manager that a token is has been released.  If
// as a result, there is no owner of the token, the token is
// deleted.
void
ACE_Token_Manager::release_mutex (ACE_Tokens *&token)
{
  ACE_TRACE ("ACE_Token_Manager::release");
  // again, let's perform our own locking here.
  ACE_TOKEN_CONST::GUARD m (this->lock_);


  if (token->dec_reference () == 0)
    {
      // no one has the token, so we can delete it and remove it from
      // our collection.  First, let's get it from the collection.
      
      TOKEN_NAME token_name (token->name (), token->type ());

      ACE_Tokens *temp;

      if (collection_.unbind (token_name, temp) == -1)
	// we did not find one in the collection
	{
	  errno = ENOENT;
	  ACE_ERROR ((LM_ERROR, "Token Manager could not release %s:%d\n",
		      token->name (), token->type ()));
	  // @@ bad
	}
      else
	// we found it
	{
	  // sanity pointer comparison.  The token referenced by the
	  // proxy better be the one we found in the list.
	  assert (token == temp);
	  delete token;  // or delete temp
	  // we set their token to zero.  if the calling proxy is
	  // still going to be used, it had better check it's token
	  // value before calling a method on it!
	  token = 0;
	}
    }
  // else
  // someone is still interested in the token, so keep it around.
}

#if defined (ACE_TEMPLATES_REQUIRE_SPECIALIZATION)
 template class ACE_Token_Name <ACE_Tokens::TOKEN_TYPE>;
 template class ACE_Map_Manager <TOKEN_NAME, ACE_Tokens *, ACE_Null_Mutex>;
 template ACE_Map_Iterator <TOKEN_NAME, ACE_Tokens *, ACE_Null_Mutex>;
 template ACE_Map_Entry <TOKEN_NAME, ACE_Tokens *>;
#endif /* ACE_TEMPLATES_REQUIRE_SPECIALIZATION */
