/* -*- C++ -*- */

// ============================================================================
//
// = LIBRARY
//    ace
// 
// = FILENAME
//    Token_Manager
//
// = AUTHOR
//    Tim Harrison (harrison@cs.wustl.edu)
// 
// ============================================================================

#if !defined (ACE_TOKEN_MANAGER_H)
#define ACE_TOKEN_MANAGER_H

#include "ace/Synch.h"
#include "ace/Map_Manager.h"
#include "ace/Local_Tokens.h"

class ACE_Local_Mutex;
class ACE_Mutex_Token;

class ACE_Token_Manager
  // = TITLE
  //   Manages all tokens in a process space.
  //
  // = DESCRIPTION
  //   Factory:  Proxies use the token manager to obtain token
  //   references.  This allows multiple proxies to reference the same
  //   logical token.
  //   Deadlock detection:  Tokens use the manager to check for
  //   deadlock situations during acquires.
{
public:
  ACE_Token_Manager (void);
  ~ACE_Token_Manager (void);

  static ACE_Token_Manager *instance (void);
  // return pointer to token manager singleton.

  typedef ACE_Token_Name<ACE_Tokens::TOKEN_TYPE> TOKEN_NAME;

  void get_mutex (const TOKEN_NAME &name,
		  ACE_Tokens *&token);
  // The token manager uses ACE_Token_Proxy::type () to find out what
  // type of token the caller is asking for.  If token == 0, then new
  // failed.  Caller must check for this.

  int check_deadlock (ACE_Token_Proxy *proxy);
  // returns 1 if the acquire will _not_ cause deadlock.
  // returns 0 if the acquire _will_ cause deadlock.
  // this method ignores recursive acquisition.  That is, it will not
  // report deadlock if the client holding the token requests the
  // token again.  Thus, it assumes recursive mutexes.

  void release_mutex (ACE_Tokens *&token);
  // notify the token manager that a token has been released.  If as a
  // result, there is no owner of the token, the token is deleted.

  ACE_TOKEN_CONST::MUTEX &mutex (void);
  // This is to allow Tokens to perform atomic transactions.  The
  // typical usage is to acquire this mutex, check for a safe_acquire,
  // perform some queueing (if need be) and then release the lock.
  // This is necessary since safe_acquire is implemented in terms of
  // the Token queues.

  virtual ACE_Tokens *make_mutex (const TOKEN_NAME &name);
  // This can be redefined if you need more lock types.

private:
  static ACE_Token_Manager *token_manager_;
  // pointer to singleton token manager.

  static ACE_TOKEN_CONST::MUTEX creation_lock_;
  // make sure that token_manager_ is created atomically (only once!)

  ACE_Tokens *token_waiting_for (const char *client_id);
  // return the token that the given client_id is waiting for, if any

  ACE_TOKEN_CONST::MUTEX lock_;
  // ACE_Mutex_Token used to lock internal data structures.

  typedef ACE_Map_Manager<TOKEN_NAME, ACE_Tokens *, ACE_Null_Mutex>
    COLLECTION;
  // COLLECTION maintains a mapping from token names to ACE_Tokens*

  typedef ACE_Map_Iterator<TOKEN_NAME, ACE_Tokens *, ACE_Null_Mutex>
    COLLECTION_ITERATOR;
  // Allows iterations through collection_

  typedef ACE_Map_Entry<TOKEN_NAME, ACE_Tokens *>
    COLLECTION_ENTRY;
  // Allows iterations through collection_

  COLLECTION collection_;
  // COLLECTION maintains a mapping from token names to ACE_Tokens*.
};


extern ACE_Token_Manager ace_token_manager;

#if defined (__INLINE__)
#include "ace/Token_Manager.i"
#endif /* __INLINE__ */

#endif /* ACE_TOKEN_MANAGER_H */
