/* Defines the classes that handle the server logging daemon. */

#if !defined (ACE_SERVER_LOGGER_C)
#define ACE_SERVER_LOGGER_C

#include "ace/Log_Record.h"
#include "ace/Log_Msg.h"
#include "ace/Get_Opt.h"
#include "Server_Logger.h"

#if !defined (__INLINE__)
#include "Server_Logger.i"
#endif /* __INLINE__ */

extern "C" ACE_Service_Object *_alloc_logger (void);

#if !defined (ACE_LACKS_STATIC_DATA_MEMBER_TEMPLATES)
// Track number of requests.
template <ACE_PEER_STREAM_1, class COUNTER, ACE_SYNCH_1>
COUNTER Logging_Handler<ACE_PEER_STREAM_2, COUNTER, ACE_SYNCH_2>::request_count_ = 0;
#endif /* ACE_LACKS_STATIC_DATA_MEMBER_TEMPLATES */

/* Dynamically linked function that dynamically allocates
   a new Logging_Acceptor object */

ACE_Service_Object *
_alloc_logger (void)
{
  return new Logging_Acceptor;
}

template <ACE_PEER_STREAM_1, class COUNTER, ACE_SYNCH_1>
Logging_Handler<ACE_PEER_STREAM_2, COUNTER, ACE_SYNCH_2>::Logging_Handler (ACE_Thread_Manager *)
{
  this->host_name_[0] = '\0'; // Initialize to a known state.
}

template <ACE_PEER_STREAM_1, class COUNTER, ACE_SYNCH_1> int
Logging_Handler<ACE_PEER_STREAM_2, COUNTER, ACE_SYNCH_2>::handle_logging_record (void)
{
  ssize_t n;
  size_t  len;
  // Lock used to serialize access to std output 
  // (this should be in the class, but the SunC++ compiler is broken...)
  static ACE_SYNCH_MUTEX lock;

  // Perform two recv's to emulate record-oriented semantiCLS.
  // Note that this code is not entirely portable since it
  // relies on the fact that sizeof (ssize_t) is the same
  // on both the sender and receiver side.  To correctly
  // handle this is painful, and we leave it as an exercise
  // for the reader ;-).

  switch (n = this->peer ().recv (&len, sizeof len))
    {
    case -1:
      ACE_ERROR_RETURN ((LM_ERROR, "%p at host %s\n",
			"server logger", this->host_name_), -1);      
      /* NOTREACHED */
    case 0:
      ACE_ERROR_RETURN ((LM_ERROR, "closing log daemon at host %s\n",
			this->host_name_), -1);	
      /* NOTREACHED */
    case sizeof (ssize_t):
      {
	ACE_Log_Record lp;
	
#if !defined (ACE_LACKS_STATIC_DATA_MEMBERS)
	u_long count = ++this->request_count_;
	ACE_DEBUG ((LM_DEBUG, "request count = %d\n", count));
#endif /* ACE_LACKS_STATIC_DATA_MEMBERS */

        len = ntohl (len);
	if ((n = this->peer ().recv_n ((void *) &lp, len)) != len)
	  ACE_ERROR_RETURN ((LM_ERROR, "len = %d, %p at host %s\n",
			    n, "server logger", this->host_name_), -1);
	/* NOTREACHED */
	  
	lp.decode ();

	if (lp.length () == n)
	  {
	    // Serialize output, if necessary (i.e., if we are running in separate threads).
	    ACE_Guard <ACE_SYNCH_MUTEX> monitor (lock);

	    lp.print (this->host_name_, 0, stderr);
	  }
	else
	  ACE_ERROR ((LM_ERROR, "error, lp.length = %d, n = %d\n",
		     lp.length (), n));
	break;
      }
    default:
      ACE_ERROR_RETURN ((LM_ERROR, "%p at host %s\n",
			"server logger", this->host_name_), -1);
      /* NOTREACHED */
    }
  return n;
}

/* Callback routine for handling the reception of remote logging transmissions */

template <ACE_PEER_STREAM_1, class COUNTER, ACE_SYNCH_1> int
Logging_Handler<ACE_PEER_STREAM_2, COUNTER, ACE_SYNCH_2>::handle_input (ACE_HANDLE)
{
  // Check if this is the first time in and properly initialize the client hostname.
  if (this->host_name_[0] == '\0')
    {
      ACE_PEER_STREAM_ADDR client_addr;

      if (this->peer ().get_remote_addr (client_addr) == -1)
	ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "get_remote_addr"), -1);

      ACE_OS::strncpy (this->host_name_, client_addr.get_host_name (), MAXHOSTNAMELEN + 1);

      ACE_DEBUG ((LM_DEBUG, "(%t) handling input from host %s on fd %d\n",
		 client_addr.get_host_name (), this->peer ().get_handle ()));

      // Shut off non-blocking IO!
      if (this->peer ().disable (ACE_NONBLOCK) == -1)
	ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "disable"), -1);
    }

  return this->handle_logging_record ();
}

template <ACE_PEER_STREAM_1, class COUNTER, ACE_SYNCH_1> int
Logging_Handler<ACE_PEER_STREAM_2, COUNTER, ACE_SYNCH_2>::put (ACE_Message_Block *, 
							       ACE_Time_Value *)
{
  return -1;
}

int
Logging_Acceptor::handle_signal (int signum
#if defined (ACE_HAS_SIGINFO_T)
, siginfo_t *, ucontext_t *
#endif /* ACE_HAS_SIGINFO_T */
)
{
  ACE_DEBUG ((LM_DEBUG, "%S\n", signum));

  // Shut down the main event loop.
  ACE_Service_Config::end_event_loop ();
  return 0;
}

Logging_Acceptor::Logging_Acceptor (void)
{
  // Register ourselves to receive SIGINT so that
  // we shut down gracefully when user signals us.

  if (ACE_Service_Config::reactor ()->register_handler (SIGINT, this) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n", "register_handler"));
}

/* Returns information on the currently active service */

int
Logging_Acceptor::info (char **strp, size_t length) const
{
  char buf[BUFSIZ];
  char addr_str[BUFSIZ];
  ACE_INET_Addr addr;

  if (this->acceptor ().get_local_addr (addr) == -1)
    return -1;
  else if (addr.addr_to_string (addr_str, sizeof addr) == -1)
    return -1;

  ACE_OS::sprintf (buf, "%s\t %s/%s %s",
	     "Logger", addr_str, "tcp",
	     "#single-threaded server logger\n");

  if (*strp == 0 && (*strp = ACE_OS::strdup (buf)) == 0)
    return -1;
  else
    ACE_OS::strncpy (*strp, buf, length);
  return ACE_OS::strlen (buf);
}

int
Logging_Acceptor::init (int argc, char *argv[])
{
  ACE_Get_Opt get_opt (argc, argv, "dp:", 0);
  ACE_INET_Addr addr;

  for (int c; (c = get_opt ()) != -1; )
     switch (c)
       {
       case 'p': 
	 addr.set (ACE_OS::atoi (get_opt.optarg));
	 break;
       case 'd':
	 break;
       default:
	 break;
       }
  
  if (this->open (addr, ACE_Service_Config::reactor ()) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "open"), -1);
  return 0;
}

#endif /* ACE_SERVER_LOGGER_C */
