#include "ace/Log_Msg.h"
#include "ace/Get_Opt.h"
#include "Deadlock_Client.h"

Deadlock_Client::Deadlock_Client (int argc, char *argv[])
  : serverHost_ (ACE_DEFAULT_SERVER_HOST),
    serverPort_ (ACE_DEFAULT_SERVER_PORT),
    token_name_ ("tokenA"),
    deadlock_token_name_ ("tokenB"),
    timeout_ (0,0),
    repeat_ (1)
{
  // Allow command line arguments to override default settings. 
  this->parse_args (argc, argv);

  // Set up the server's address.
  ACE_Remote_Mutex::set_server_address (ACE_INET_Addr (serverPort_, serverHost_));

  ACE_Synch_Options::defaults.timeout (timeout_);
}

// Run the simple repeating test client.
int
Deadlock_Client::run (void)
{
  ACE_DEBUG ((LM_DEBUG, "starting up client test\n"));

  if (remote_mutex_.open (token_name_) == -1)
    ACE_ERROR ((LM_ERROR, "remote_mutex_.open failed"));

  if (deadlock_remote_mutex_.open (deadlock_token_name_) == -1)
    {
      ACE_ERROR ((LM_ERROR, "deadlock_remote_mutex_.open failed"));
    }

  // Repeat the acquire/renew/release repeat_ number of times
  while (this->repeat_ > 0)
  {
    if (do_acquire (remote_mutex_) == -1)
      {
	// I don't think this can ever happen
	assert (errno != EDEADLK);
	ACE_ERROR_RETURN ((LM_ERROR, "first acquire bombed\n"), -1);
      }

    // Acquire the deadlock token.
    if (do_acquire (deadlock_remote_mutex_) == -1)
      {
	if (errno == EDEADLK)
	  goto DEADLOCK_RELEASE;
	else
	  // serious error
	  ACE_ERROR_RETURN ((LM_ERROR, "second acquire bombed\n"), -1);
      }

    // Release the token.
    if (do_release (deadlock_remote_mutex_) == -1)
      ACE_ERROR_RETURN ((LM_ERROR, "first release bombed\n"), -1);

DEADLOCK_RELEASE:

    // Release the token.
    if (do_release (remote_mutex_) == -1)
      ACE_ERROR_RETURN ((LM_ERROR, "second release bombed\n"), -1);

    --this->repeat_;
  }
  
  ACE_DEBUG ((LM_DEBUG, "shutting down client test\n"));
  return 0;
}

int
Deadlock_Client::parse_args (int argc, char *argv[])
{
  ACE_LOG_MSG->open (argv[0]);

  ACE_Get_Opt get_opt (argc, argv, "c:h:p:r:a:b:w:", 1);

  for (int c; (c = get_opt ()) != -1; )
    {
      switch (c)
	{
	case 'h':  // specify the host machine on which the server is running
	  this->serverHost_ = get_opt.optarg;
	  break;
	case 'p':  // specify the port on which the server is running
	  this->serverPort_ = ACE_OS::atoi (get_opt.optarg);
	  break;
	case 'r':  // specify the number of repetitions of the
                   // acquire/renew/release cycle to perform
	  this->repeat_ = ACE_OS::atoi (get_opt.optarg);
	  break;
	case 'a':  // specify the tokenname to use
	  this->token_name_ = get_opt.optarg;
	  break;
	case 'b':  // specify the tokenname to use
	  this->deadlock_token_name_ = get_opt.optarg;
	  break;
	case 'w':  // specify the timeout Window for acquires and renews
	  {
	    this->timeout_.sec (ACE_OS::atoi (get_opt.optarg));

	    // Try to locate the ':' that separates the secs from the usecs.
	    char *usecs = ACE_OS::strchr (get_opt.optarg, ':');

	    if (usecs != 0)
	      this->timeout_.usec (ACE_OS::atoi (&usecs[1])); // Skip past ':';
	    break;
	  }
	default:
	  ACE_ERROR_RETURN ((LM_ERROR, 
			    "%n:\n"
			     "[-h server-host]\n[-p server-port]\n"
			     "[-r repetitions]\n"
			     "[-t token A name]\n"
			     "[-d token B name]\n"
			     "[-w secs:usecs\n%a", 1), -1);
	  break;
	}
    }

  return 0;
}

int
Deadlock_Client::do_acquire (ACE_Token_Proxy &mutex)
{
  // Try to acquire the token.
  if (mutex.acquire () == -1)
    {
      switch (errno)
      {
	case ETIME:
  	case EWOULDBLOCK:
	  ACE_ERROR ((LM_ERROR, "%p %s\n", 
	             "acquire timed out", mutex.name ()));
	  break;
	case ENOLCK:
	  ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
          	            "token lock was removed unexpectedly", 
		            mutex.name ()), -1);
  	case EDEADLK:
	  ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
			     "deadlock detected", mutex.name ()), -1);
	  break;
	default:
	  ACE_ERROR_RETURN ((LM_ERROR, "unexpected error %d\n",
			    errno), -1);
      }
    } 
  else
    ACE_DEBUG ((LM_DEBUG, "got token %s\n", mutex.name ()));
  return 0;
}
    
int
Deadlock_Client::do_renew (ACE_Token_Proxy &mutex)
{
  // Try to renew the token.
  if (mutex.renew () == -1)
    {
      switch (errno)
      {
	case ETIME:
	case EWOULDBLOCK:
	  ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
			    "renew timed out", mutex.name ()), -1);
	  /* NOTREACHED */
	case EACCES:
	  ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
			    "not owner of token", mutex.name ()), -1);
	  /* NOTREACHED */
  	case ENOLCK:
	  ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
			    "token lock was removed unexpectedly", 
			    mutex.name ()), -1);
  	case EDEADLK:
	  ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
			     "deadlock detected", mutex.name ()), -1);
	  break;
	default:
	  ACE_ERROR_RETURN ((LM_ERROR, "unexpected error %d\n",
			    errno), -1);
      }
    }
  else
    ACE_DEBUG ((LM_DEBUG, "renewed token %s\n", mutex.name ()));
  return 0;
}
    
int
Deadlock_Client::do_release (ACE_Token_Proxy &mutex)
{
  // Try to release the token.
  if (mutex.release () == -1)
    if (errno == EACCES)
        ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
			   "not owner of token", mutex.name ()), -1);
    else
        ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", "release failed", mutex.name ()), -1);
  else
    ACE_DEBUG ((LM_DEBUG, "released token %s\n", mutex.name ()));
  return 0;
}
