#include "ace/Log_Msg.h"
#include "ace/Get_Opt.h"
#include "Simple_Token_Client.h"

Simple_Token_Client::Simple_Token_Client (int argc, char *argv[])
  : serverHost_ (ACE_DEFAULT_SERVER_HOST),
    serverPort_ (ACE_DEFAULT_SERVER_PORT),
    tokenName_ ("depreciation"),
    timeout_ (0,0),
    repeat_ (1)
{
  // Allow command line arguments to override default settings. 
  this->parse_args (argc, argv);

  // Set up the server's address.
  ACE_Remote_Mutex::set_server_address (ACE_INET_Addr (serverPort_, serverHost_));

  ACE_Synch_Options::defaults.timeout (timeout_);
}

// Run the simple repeating test client.
int
Simple_Token_Client::run (void)
{
  if (remote_mutex_.open (tokenName_) == -1)
    ACE_ERROR ((LM_ERROR, "remote_mutex_.open failed"));

  ACE_DEBUG ((LM_DEBUG, "starting up client test\n"));

  // Repeat the acquire/renew/release repeat_ number of times
  while (this->repeat_ > 0)
  {
    // Acquire the token.
    do_acquire();
    
    // Renew the token.
    do_renew();

    // Release the token.
    do_release();

    --this->repeat_;
  }
  
  ACE_DEBUG ((LM_DEBUG, "shutting down client test\n"));
  return 0;
}

int
Simple_Token_Client::parse_args (int argc, char *argv[])
{
  ACE_LOG_MSG->open (argv[0]);

  ACE_Get_Opt get_opt (argc, argv, "c:dh:p:r:t:w:", 1);

  for (int c; (c = get_opt ()) != -1; )
    {
      switch (c)
	{
	case 'h':  // specify the host machine on which the server is running
	  this->serverHost_ = get_opt.optarg;
	  break;
	case 'p':  // specify the port on which the server is running
	  this->serverPort_ = ACE_OS::atoi (get_opt.optarg);
	  break;
	case 'r':  // specify the number of repetitions of the
                   // acquire/renew/release cycle to perform
	  this->repeat_ = ACE_OS::atoi (get_opt.optarg);
	  break;
	case 't':  // specify the tokenname to use
	  this->tokenName_ = get_opt.optarg;
	  break;
	case 'w':  // specify the timeout Window for acquires and renews
	  {
	    this->timeout_.sec (ACE_OS::atoi (get_opt.optarg));

	    // Try to locate the ':' that separates the secs from the usecs.
	    char *usecs = ACE_OS::strchr (get_opt.optarg, ':');

	    if (usecs != 0)
	      this->timeout_.usec (ACE_OS::atoi (&usecs[1])); // Skip past ':';
	    break;
	  }
	default:
	  ACE_ERROR_RETURN ((LM_ERROR, 
			    "%n:\n"
			    "[-h server-host]\n[-p server-port]\n"
			    "[-r repetitions]"
			    "[-t token-name]\n[-w secs:usecs\n%a", 1), -1);
	  break;
	}
    }

  return 0;
}

int
Simple_Token_Client::do_acquire(void)
{
  // Try to acquire the token.
  if (this->remote_mutex_.acquire () == -1)
    {
      switch (errno)
      {
	case ETIME:
  	case EWOULDBLOCK:
	  ACE_ERROR ((LM_ERROR, "%p %s\n", 
	             "acquire timed out", this->tokenName_));
	  break;
	case ENOLCK:
	  ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
          	            "token lock was removed unexpectedly", 
		            this->tokenName_), -1);
	default:
	  ACE_ERROR_RETURN ((LM_ERROR, "unexpected error %d\n",
			    errno), -1);
      }
    } 
  else
    ACE_DEBUG ((LM_DEBUG, "got token %s\n", this->tokenName_));
}
    
int
Simple_Token_Client::do_renew(void)
{
  // Try to renew the token.
  if (this->remote_mutex_.renew () == -1)
    {
      switch (errno)
      {
	case ETIME:
	case EWOULDBLOCK:
	  ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
			    "renew timed out", this->tokenName_), -1);
	  /* NOTREACHED */
	case EACCES:
	  ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
			    "not owner of token", this->tokenName_), -1);
	  /* NOTREACHED */
  	case ENOLCK:
	  ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
			    "token lock was removed unexpectedly", 
			    this->tokenName_), -1);
	default:
	  ACE_ERROR_RETURN ((LM_ERROR, "unexpected error %d\n",
			    errno), -1);
      }
    }
  else
    ACE_DEBUG ((LM_DEBUG, "renewed token %s\n", this->tokenName_));
}
    
int
Simple_Token_Client::do_release(void)
{
  // Try to release the token.
  if (this->remote_mutex_.release () == -1)
    if (errno == EACCES)
        ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", 
			   "not owner of token", this->tokenName_), -1);
    else
        ACE_ERROR_RETURN ((LM_ERROR, "%p %s\n", "release failed", this->tokenName_), -1);
  else
    ACE_DEBUG ((LM_DEBUG, "released token %s\n", this->tokenName_));
}
