/* -*- C++ -*- */

// ============================================================================
//
// = LIBRARY
//    ace
// 
// = FILENAME
//    Simple_Token_Client.h
//
// = DESCRIPTION 
// = BEGIN<NOFILL>
//    simpletokenclient performs a simple acquire-renew-release cycle on a 
//    token a specified number of times.  Basic program parameters are passed 
//    as command-line arguments.
//
//    The success or failure of the connection to the Token_Server, the success 
//    or failure of the requested operations, and the termination of the 
//    program are reported as the event happens.  Timeouts are also reported.
//
//    Command line arguments are as follows:
//
//     		-c client-id	Allows an explicit set of the client id.  If
//				this option is not used, a default id is
// 				assigned.
//
//		-h host-name	Specifies the host on which the Token_Server is
//				running.  Required if the Token_Server is
//				not running as a local background process.
//
// 		-p port-number	Specifies the port on the Token_Server's host
//				machine to use for communication.  
//		-r repetitions	Specifies the number of times to repeat the
//				acquire-renew-release cycle on the token.
//				Defaults to 1.
//
//		-t token-name	Specifies the token name to use.  Defaults to
//				'appreciation'.
//
//		-w secs:usecs	Specifies the timeout to use for acquire and
//				renew operations.  Without this parameter, an
//				acquire or renew can block forever.
//
//    simpletokenclient reports any errors in an acquistion, renewal, or
//    release of the token.
//
// = EXAMPLES
// = BEGIN<NOFILL>
//    simpletokenclient
//  		Performs a single acquire-renew-release cycle with the default
//		token ('appreciation').  Acquires and renews may block forever.
//		The Token_Server is running on the local machine at the 
//		default port.
//
//    simpletokenclient -h ritz -p 10004 -r 100 -w 10:00
//		Tells the client that the Token_Server is on machine 'ritz' in 
//		the local domain at port 10004.  Repeats the
//		acquire-renew-release cycle 100 times, with a timeout of 10
//		seconds on acquires and renews.
// = END<NOFILL>
//
// = AUTHOR
//    Steve DeLassus
// 
// ============================================================================

#if !defined (_SIMPLE_TOKEN_CLIENT_H)
#define _SIMPLE_TOKEN_CLIENT_H

#include "ace/Event_Handler.h"
#include "ace/Remote_Tokens.h"

class Simple_Token_Client : public ACE_Event_Handler
{
public:
  Simple_Token_Client (int argc, char *argv[]);
  // Initialize the test client.

  int run (void);
  // Run the test client.

private:
  int parse_args (int argc, char *argv[]);
  // Parse the command line arguments.

  int do_acquire(void);
  // Attempt an acquisition of the token.

  int do_renew(void);
  // Attempt to renew the token.

  int do_release(void);
  // Attempt to release the token.

  ACE_Remote_Mutex remote_mutex_;
  // Proxy for the token server.

  char *serverHost_;
  // Name of the server.

  u_short serverPort_;
  // Port number of the server.

  char *tokenName_;
  // Name of the token.

  ACE_INET_Addr addr_;
  // Address of the server.

  ACE_Time_Value timeout_;
  // Timeout value for <this->acquire()>.

  int repeat_;
  // Number of times to repeat test.

};

#endif /* _SIMPLE_TOKEN_CLIENT_H */
