#include "ace/Log_Msg.h"
#include "ace/Token_Request_Reply.h"
#include "Token_Acceptor.h"
#include "Token_Handler.h"

// Default constructor.
Token_Handler::Token_Handler (void)
 :  reactor_ (0),
    collection_ (1),
    timeout_id_ (0)
{
}

// Activate this instance of the Token_Handler (called by the Token_Acceptor). 

/* VIRTUAL */ int 
Token_Handler::open (Token_Acceptor *ta)
{
  ACE_INET_Addr addr;

  // Find out which client we are connected with.
  if (this->peer_.get_remote_addr (this->addr_) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "get_remote_addr"), -1);

  ACE_DEBUG ((LM_DEBUG, "%n: received connection from %s on handle %d\n", 
	     this->addr_.get_host_name (), this->get_handle ()));

  // Keep a pointer to the ACE_Reactor so that timers work correctly and efficiently.
  this->reactor_ = ta->reactor_;

  // Register this object with the ACE_Reactor.
  // We will then receive upcalls when our client requests operations.
  if (this->reactor_->register_handler (this, ACE_Event_Handler::READ_MASK) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "register_handler"), -1);

  return 0;
}

// Close down this instance of the Token_Handler.

/* VIRTUAL */ int 
Token_Handler::close (u_long)
{
  this->handle_close ();
  return 0;
}

// Conversion operator required by the Token_Acceptor.
Token_Handler::operator ACE_SOCK_Stream & (void)
{
  return this->peer_;
}

// Create and send a reply to the client.

/* VIRTUAL */ int 
Token_Handler::send_reply (ACE_UINT32 err)
{
  void    *buf;
  size_t  len;
  ssize_t n;

  this->token_reply_.errnum (err);

  len = this->token_reply_.encode (buf);

  if (len == -1)
    return -1;
    
  n = this->peer_.send (buf, len);

  if (n != len)
    ACE_ERROR_RETURN ((LM_ERROR, "%p, expected len = %d, actual len = %d\n",
		      "send failed", len, n), -1);
  else
    return 0;
}

// Return the underlying ACE_HANDLE.

/* VIRTUAL */ ACE_HANDLE 
Token_Handler::get_handle (void) const
{
  return this->peer_.get_handle ();
}

// Acquire the token.

/* VIRTUAL */ int 
Token_Handler::acquire (ACE_Token_Proxy *proxy)
{
  ACE_DEBUG ((LM_DEBUG, "in acquire for client id = %s\n", 
	     proxy->client_id ()));

  // @@ add notify in token request reply
  if (proxy->acquire (0, 0, ACE_Synch_Options::asynch) == -1)
    {
      if (errno != EWOULDBLOCK)
	// bad bad bad
	return this->send_reply (errno);

      // acquire would block
      if (request_options_[ACE_Synch_Options::USE_TIMEOUT] == 1)
	{
	  // check for polling
	  if ((request_options_.timeout ().sec () == 0) &&
	      (request_options_.timeout ().usec () == 0))
	    return this->send_reply (EWOULDBLOCK);

	  // schedule a timer
	  timeout_id_ = this->reactor_->schedule_timer (this, (void *) proxy,
							request_options_.timeout ());
	  if (timeout_id_ == -1)
	    {
	      ACE_ERROR ((LM_ERROR, "%p\n", "schedule_timer"));
	      return this->send_reply (errno);
	    }
	}
      // send no reply.  wait until we acquire it or until the timer
      // goes off.
      return 0;
    }
  else
    // success
    {
      return this->send_reply (ACE_Token_Reply::SUCCESS);
    }
}

// Try to acquire the token.  Never block.

/* VIRTUAL */ int 
Token_Handler::try_acquire (ACE_Token_Proxy *proxy)
{
  ACE_DEBUG ((LM_DEBUG, "in try_acquire for client id = %s\n",
	     proxy->client_id ()));

  // @@ add notify in token request reply
  if (proxy->tryacquire () == -1)
    return this->send_reply (errno);
  else
    return this->send_reply (ACE_Token_Reply::SUCCESS);
}

// Release the token and allow the next client that is waiting to proceed.

/* VIRTUAL */ int 
Token_Handler::release (ACE_Token_Proxy *proxy)
{
  ACE_DEBUG ((LM_DEBUG, "in release for client id = %s\n", proxy->client_id ()));

  if (proxy->release (ACE_Synch_Options::asynch) == -1)
    // oops, it failed
    return this->send_reply (ACE_LOG_MSG->errnum ());

  // success
  if (timeout_id_ != 0)
    {
      reactor_->cancel_timer (timeout_id_);
      timeout_id_ = 0;
    }
  
  return this->send_reply (ACE_Token_Reply::SUCCESS);
}

// Yield the token if any clients are waiting, otherwise keep the token.

/* VIRTUAL */ int 
Token_Handler::renew (ACE_Token_Proxy *proxy)
{
  ACE_DEBUG ((LM_DEBUG, "in renew for client id = %s\n", 
	     proxy->client_id ()));


  if (proxy->renew (token_request_.requeue_position (), ACE_Synch_Options::asynch) == -1)
    {
      int result = ACE_LOG_MSG->errnum ();
      if (result != EWOULDBLOCK)
	// bad bad bad
	return this->send_reply (result);

      // acquire would block
      if (request_options_[ACE_Synch_Options::USE_TIMEOUT] == 1)
	{
	  timeout_id_ = this->reactor_->schedule_timer (this, 0,
							request_options_.timeout ());
	  if (timeout_id_ == -1)
	    {
	      ACE_ERROR ((LM_ERROR, "%p\n", "schedule_timer"));
	      return this->send_reply (ACE_LOG_MSG->errnum ());
	    }
	}
      // send no reply.  wait until we acquire it or until the timer
      // goes off.
      return 0;
    }
  else
    // success, we still hold the token
    {
      return this->send_reply (ACE_Token_Reply::SUCCESS);
    }
}

/* VIRTUAL */ int 
Token_Handler::remove (ACE_Token_Proxy *proxy)
{
  ACE_DEBUG ((LM_DEBUG, "in remove for client id = %s\n", 
	     proxy->client_id ()));
  ACE_ERROR ((LM_ERROR, "sorry: Token_Handler::remove() is not implemented"));
  return this->send_reply (ENOTSUP);
}

// Enable clients to limit the amount of time they'll wait for a token.

/* VIRTUAL */ int 
Token_Handler::handle_timeout (const ACE_Time_Value &, const void *tp)
{
  ACE_TRACE ("Token_Handler::handle_timeout");

  ACE_Token_Proxy *proxy = (ACE_Token_Proxy *) tp;

  ACE_DEBUG ((LM_DEBUG, "in handle_timeout for client id = %s\n", 
	     proxy->client_id ()));

  timeout_id_ = 0;

  // @@ add a try acquire here!
  // Try to acquire the token, but if we can't get it immediately
  // then abandon the wait.
//  if (this->try_acquire (&token_entry) == -1)
//    return this->abandon (token_entry);

  // remove ourselves from the waiter list.
  proxy->release ();

  this->send_reply (ETIME);
  return 0;
}

// Dispatch the appropriate operation to handle the client request.

ACE_Token_Proxy *
Token_Handler::get_proxy (void)
{
  ACE_Token_Name<ACE_Token_Proxy::PROXY_TYPE> name (token_request_.token_name (), 
						    token_request_.proxy_type ());
  // make sure that the proxy exists in the collection
  ACE_Token_Proxy *proxy = collection_.is_member (name);
  if (proxy == 0)
    {
      // create a new proxy for this client_id
      ACE_TS_Mutex new_proxy (token_request_.token_name (),1,this,
			      token_request_.proxy_type (), 
			      token_request_.token_type ());
      // set the client_id (it was set to 1 since we're single-threaded.
      new_proxy.client_id (token_request_.client_id ());

      // put the new_proxy in this client_id's collection
      if (collection_.insert (new_proxy) == -1)
	ACE_ERROR_RETURN ((LM_ERROR, "insert failed\n"), 0);

      // let's try again.  we have to get the proxy created by the
      // collection since it made a new copy.
      proxy = collection_.is_member (name);
      if (proxy == 0)
	ACE_ERROR_RETURN ((LM_ERROR, "is_member failed\n"), 0);
    }

  return proxy;
}

/* VIRTUAL */ int 
Token_Handler::dispatch (void)
{
  ACE_Token_Proxy *proxy = this->get_proxy ();

  if (proxy == 0)
    return -1;

  // Dispatch the appropriate request.
  switch (this->token_request_.operation_type ())
    {
    case ACE_Token_Request::ACQUIRE:
      return this->acquire (proxy);
    case ACE_Token_Request::TRY_ACQUIRE:
      return this->try_acquire (proxy);
    case ACE_Token_Request::RELEASE:
      return this->release (proxy);
    case ACE_Token_Request::RENEW:
      return this->renew (proxy);
    case ACE_Token_Request::REMOVE:
      return this->remove (proxy);
    default:
      ACE_ERROR_RETURN ((LM_ERROR, "invalid type = %d\n", 
			this->token_request_.operation_type ()), -1);
      /* NOTREACHED */
    }
}
  
// Receive, frame, and decode the client's request.
// Note, this method should use non-blocking I/O.

/* VIRTUAL */ int 
Token_Handler::recv_request (void)
{
  ssize_t n;

  // Read the first 4 bytes to get the length of the message
  // This implementation assumes that the first 4 bytes are
  // the length of the message.
  n = this->peer_.recv ((void *) &this->token_request_, sizeof (ACE_UINT32));

  switch (n)
    {
    case -1:
      /* FALLTHROUGH */
    default:
      ACE_ERROR ((LM_ERROR, "%p got %d bytes, expected %d bytes\n", 
		 "recv failed", n, sizeof (ACE_UINT32)));
      /* FALLTHROUGH */
    case 0:
      // We've shutdown unexpectedly, let's abandon the connection.
      this->abandon (0);
      return -1;
      /* NOTREACHED */
    case sizeof (ACE_UINT32):
      {
	// Transform the length into host byte order.
	ssize_t length = ntohl (this->token_request_.length ());

	// Do a sanity check on the length of the message.
	if (length > sizeof this->token_request_)
	  {
	    ACE_ERROR ((LM_ERROR, "length %d too long\n", length));
	    return this->abandon (1);
	  }

	// Receive the rest of the request message.
	// @@ beware of blocking read!!!.
	n = this->peer_.recv ((void *) (((char *) &this->token_request_)
					+ sizeof (ACE_UINT32)), length);

	// Subtract off the size of the part we skipped over...
	if (n != (length - sizeof (ACE_UINT32)))
	  {
	    ACE_ERROR ((LM_ERROR, "%p expected %d, got %d\n", 
		       "invalid length", length, n));
	    return this->abandon (1);
	  }

	// Decode the request into host byte order.
	if (this->token_request_.decode () == -1)
	  {
	    ACE_ERROR ((LM_ERROR, "%p\n", "decode failed"));
	    return this->abandon (1);
	  }

	// if (OS::debug)
	this->token_request_.dump ();
      }
    }
  return 0;
}

// Callback method invoked by the ACE_Reactor when 
// events arrive from the client.

/* VIRTUAL */ int 
Token_Handler::handle_input (ACE_HANDLE)
{
  ACE_DEBUG ((LM_DEBUG, "****************** in handle_input\n"));

  if (this->recv_request () == -1)
    return -1;
  else
    return this->dispatch ();
}

// Close down the Token_Handler 
// (e.g., when a client disconnectors or releases a token).

/* VIRTUAL */ int 
Token_Handler::handle_close (ACE_HANDLE, ACE_Reactor_Mask)
{
  ACE_DEBUG ((LM_DEBUG, "removing Token_Handler on handle %d\n",
	      this->get_handle ()));

//  this->abandon (0);

  // *Must* be allocated dynamically!
  delete this;
  return 0;
}

Token_Handler::~Token_Handler (void)
{
  this->peer_.close ();
}

void
Token_Handler::sleep_hook (void)
{
  // @@ what should we do?
  return;
}

void
Token_Handler::token_acquired (void *)
{
  if (timeout_id_ != 0)
    {
      reactor_->cancel_timer (timeout_id_);
      timeout_id_ = 0;
    }
  
  this->send_reply (ACE_Token_Reply::SUCCESS);
}

int
Token_Handler::abandon (int send_error)
{
  // release ownership or remove us from the waiter list.
  if (timeout_id_ != 0)
    {
      reactor_->cancel_timer (timeout_id_);
      timeout_id_ = 0;
    }

  // @@ release all tokens
  collection_.release ();
  
  if (send_error)
    return this->send_reply (EIO);
  else
    return -1;
}
