/* -*- C++ -*- */

// ============================================================================
//
// = LIBRARY
//    ACE
// 
// = FILENAME
//    Token_Handler.h
//
// = DESCRIPTION 
//    Product object created by <Token_Acceptor>.  A <Token_Handler> exchanges
//    messages with a <Token_Proxy> object on the client-side.
//
// = AUTHOR
//    Douglas C. Schmidt (schmidt@cs.wustl.edu)
// 
// ============================================================================

#if !defined (_TOKEN_HANDLER_H)
#define _TOKEN_HANDLER_H

#include "ace/Reactor.h"
#include "ace/SOCK_Stream.h"
#include "ace/INET_Addr.h"
#include "ace/SString.h"
#include "ace/Token_Request_Reply.h"
#include "ace/Token_Collection.h"
#include "TS_Mutex.h"

// = Forward declarations.
class Token_Acceptor; // ACE_Acceptor factory that creates TokenHandlers.
class Token_Map;      // Maps client requests to list of active token holders.
class Token_Entry;    // Keeps track of current token holder and list of waiters.

class Token_Handler : public ACE_Event_Handler
  // = TITLE
  //    Product object created by <Token_Acceptor>.  A <Token_Handler> exchanges
  //    messages with a <Token_Proxy> object on the client-side.
  //
  // = DESCRIPTION
  //   This class is the main workhorse of the <Token_Server>.  It handles
  //   client requests to acquire, release, renew, and remove tokens.  It also
  //   schedules and handles timeouts that are used to support "timed waits."
  //   Clients used timed waits to bound the amount of time they block 
  //   trying to get a token.
{
public:
  // = Initialization and termination.

  Token_Handler (void);
  // Default constructor.

  virtual int open (Token_Acceptor *ta = 0);
  // Activate this instance of the <Token_Handler> 
  // (called by the <Token_Acceptor>).

  virtual int close (u_long = 0);
  // Close down this instance of the <Token_Handler>.

  // = Accessor and mutator methods.

  operator ACE_SOCK_Stream & (void);
  // Conversion operator required by the Token_Acceptor.

  // = Remote operations "exported" to a client.
  virtual int acquire (ACE_Token_Proxy *proxy);
  // Try to acquire the token.
  // Precondition: client *may* hold the token already (i.e.,
  // supports recursive acquisitions).

  virtual int try_acquire (ACE_Token_Proxy *proxy);
  // try

  virtual int release (ACE_Token_Proxy *proxy);
  // Release the token and allow the next client that is waiting to proceed.
  // Preconditions: client must hold the token.

  virtual int renew (ACE_Token_Proxy *proxy);
  // Yield the token if any clients are waiting, otherwise keep the token.
  // Preconditions: client must hold the token.

  virtual int remove (ACE_Token_Proxy *proxy);
  // Remove the specified token from the Token_Map.
  // Preconditions: ACE_Token must exist.
  // @@ Any other preconditions, e.g., must client hold token, must there 
  // be no waiters, etc.?

  void sleep_hook (void);
  // called by TS_[Mutex,RLock,WLock] when we hold the mutex and
  // someone wants it.

  void token_acquired (void* arg);
  // called by TS_[Mutex,RLock,WLock] when we are waiting and acquire
  // the mutex.

protected:
  // = Low level routines for framing requests, dispatching operations, 
  // and returning replies.

  virtual int abandon (int send_error);
  // our connection has been closed

  virtual int recv_request (void);
  // Receive, frame, and decode the client's request.
  
  virtual int dispatch (void);
  // Dispatch the appropriate operation to handle the client's request.
  
  virtual int send_reply (ACE_UINT32 errnum);
  // Create and send a reply to the client.

  // = Demultiplexing hooks.
  virtual ACE_HANDLE get_handle (void) const;
  // Return the underlying <ACE_HANDLE>.

  virtual int handle_input (ACE_HANDLE);
  // Callback method invoked by the <ACE_Reactor> when client events arrive.

  virtual int handle_close (ACE_HANDLE = -1, 
                            ACE_Reactor_Mask = ACE_Event_Handler::RWE_MASK);
  // Close down the <Token_Handler> (e.g., when a client 
  // disconnectors or releases a token.

  // = Timer hook.
  virtual int handle_timeout (const ACE_Time_Value &tv, const void *arg);
  // Enable clients to limit the amount of time they wait for a token.

  ACE_Token_Proxy *get_proxy (void);
  // return a proxy for the calling client_id and token name.

private:

  ACE_Synch_Options request_options_;

  int timeout_id_;

  ACE_Token_Collection collection_;
  // collection of the client's token proxies.

  ACE_Token_Request token_request_;
  // Cache request from the client.

  ACE_Token_Reply token_reply_;
  // Cache reply to the client.

  ACE_SOCK_Stream peer_;
  // Maintains connection to client peer.

  ACE_INET_Addr addr_;
  // Address of client we are connected with.

  ACE_Reactor *reactor_;
  // Pointer to the <ACE_Reactor> (owned by the <Token_Server>).
  // This is needed to make timers work correctly and efficiently.
  
  ~Token_Handler (void);
  // Ensure dynamic allocation...
};

#endif /* _TOKEN_HANDLER_H */
