/* Test the event server */
#include "ace/Log_Msg.h"
#include "ace/Stream.h"
#include "ace/Service_Config.h"
#include "Options.h"
#include "Consumer_Router.h"
#include "Event_Analyzer.h"
#include "Supplier_Router.h"

#if defined (ACE_HAS_THREADS)

typedef ACE_Stream<ACE_MT_SYNCH> MT_Stream;
typedef ACE_Module<ACE_MT_SYNCH> MT_Module;

/* Handle SIGINT and terminate the entire application */

class Quit_Handler : public ACE_Event_Handler
{
public:
  Quit_Handler (int fd, MT_Stream &es, ACE_Thread_Manager &thr_manager);
  virtual int handle_input (int fd);
  virtual int get_handle (void) const;
  int     done (void);

private:
  sig_atomic_t    done_;
  MT_Stream       &event_server_;
  ACE_Thread_Manager  &thr_manager_;
  int	          handle_;
};

Quit_Handler::Quit_Handler (int fd, MT_Stream &es, ACE_Thread_Manager &thr_manager)
  : event_server_ (es), 
    thr_manager_ (thr_manager), 
    handle_ (fd), 
    done_ (0)
{  
}

inline int
Quit_Handler::done (void)
{
  return this->done_;
}

int
Quit_Handler::get_handle (void) const
{
  return this->handle_;
}

int
Quit_Handler::handle_input (int)
{
  options.stop_timer ();
  ACE_DEBUG ((LM_INFO, "(%t) closing down the test\n"));
  options.print_results ();

  this->thr_manager_.suspend_all ();
  this->event_server_.close ();
  this->done_ = 1;
  return 0;
}

int
main (int argc, char *argv[])
{
  ACE_Service_Config daemon;
  
  options.parse_args (argc, argv);
  options.start_timer ();

  /* Primary ACE_Stream for EVENT_SERVER application */
  MT_Stream event_server; 

  /* ACE_Thread pool manager */
  ACE_Thread_Manager thr_manager (ACE_Thread_Manager::DEFAULT_SIZE);

  /* Enable graceful shutdowns... */
  Quit_Handler quit_handler (0, event_server, thr_manager);

  if (ACE_Service_Config::reactor ()->register_handler (&quit_handler, ACE_Event_Handler::READ_MASK) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "register_handler"), -1);    

  /* Create the modules. */

  MT_Module *sr = new MT_Module ("Supplier_Router", new Supplier_Router (&thr_manager));
  MT_Module *ea = new MT_Module ("Event_Analyzer", new Event_Analyzer, new Event_Analyzer);
  MT_Module *cr = new MT_Module ("Consumer_Router", 
				 0, // 0 triggers the creation of a ACE_Thru_Task...
				 new Consumer_Router (&thr_manager));

  /* Push the modules onto the event_server stream */

  if (event_server.push (sr) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "push (Supplier_Router)"), -1);
					
  if (event_server.push (ea) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "push (Event_Analyzer)"), -1);

  if (event_server.push (cr) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "push (Consumer_Router)"), -1);

  /* Set the high and low water marks appropriately */

  int wm = options.low_water_mark ();

  if (event_server.control (ACE_IO_Cntl_Msg::SET_LWM, &wm) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "push (setting low watermark)"), -1);

  wm = options.high_water_mark ();
  if (event_server.control (ACE_IO_Cntl_Msg::SET_HWM, &wm) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "push (setting high watermark)"), -1);

  /* Perform the main event loop */

  while (!quit_handler.done ())
    ACE_Service_Config::reactor ()->handle_events ();

  return 0;
}
#else
int 
main (void)
{
  ACE_ERROR_RETURN ((LM_ERROR, "test not defined for this platform\n"), -1);
}
#endif /* ACE_HAS_THREADS */
