/* Test the event server */
#include "ace/Log_Msg.h"
#include "ace/Stream.h"
#include "ace/Service_Config.h"
#include "Options.h"
#include "Consumer_Router.h"
#include "Event_Analyzer.h"
#include "Supplier_Router.h"
#include "ace/UPIPE_Acceptor.h"
#include "ace/UPIPE_Connector.h"
#include "ace/UPIPE_Stream.h"

#if defined (ACE_HAS_THREADS)

typedef ACE_Stream<ACE_MT_SYNCH> MT_Stream;
typedef ACE_Module<ACE_MT_SYNCH> MT_Module;

/* ACE_Thread pool manager */
ACE_Thread_Manager thr_mgr (ACE_Thread_Manager::DEFAULT_SIZE);

/* Handle SIGINT and terminate the entire application */

class Quit_Handler : public ACE_Event_Handler
{
public:
  Quit_Handler (int fd, MT_Stream &es, ACE_Thread_Manager &thr_mgr);
  virtual int handle_input (int fd);
  virtual int get_handle (void) const;
  int     done (void);

private:
  sig_atomic_t    done_;
  MT_Stream       &event_server_;
  ACE_Thread_Manager  &thr_mgr_;
  int	          fd_;
};

Quit_Handler::Quit_Handler (int fd, MT_Stream &es, 
			    ACE_Thread_Manager &thr_mgr)
  : event_server_ (es), 
    thr_mgr_ (thr_mgr), 
    fd_ (fd), 
    done_ (0)
{  
}

inline int
Quit_Handler::done (void)
{
  return this->done_;
}

int
Quit_Handler::get_handle (void) const
{
  return this->fd_;
}

int
Quit_Handler::handle_input (int)
{
  options.stop_timer ();
  ACE_DEBUG ((LM_INFO, " (%t) closing down the test\n"));
  options.print_results ();

  this->thr_mgr_.suspend_all ();
  this->event_server_.close ();
  this->done_ = 1;
  return 0;
}

static void *
consumer (void * dummy)
{
  ACE_UPIPE_Stream c_stream;
  ACE_UPIPE_Addr c_addr ("/tmp/conupipe");
  int iter = options.iterations ();
  int verb = options.verbose ();
  int msiz = options.message_size ();
  int secs,par1,par2,i;
  time_t currsec;

  if (verb)
    cout << "consumer starting connect" << endl;
  ACE_UPIPE_Connector con;

  if (con.connect (c_stream, c_addr) == -1)
    ACE_DEBUG ((LM_INFO, " (%t) connect failed\n"));
  else
    cout << "consumer :we're connected" << endl;

  char	   buf[BUFSIZ];
  int	   n;
  ACE_Message_Block * mb_p;

  int done = 0;
  int cnt = 0;
  time (&currsec);
  par1= (int)currsec;

  while (done == 0 && ((n = c_stream.recv_msg (mb_p)) != -1))
  {
    if (mb_p->length () > 1) 
    {
      cnt++;
      if (verb)
      cout << " consumer received message !!!!!! " <<
           mb_p->rd_ptr () << endl;
    }
    else
    {
      if (verb)
        cout << "consumer got last mb" << (char) * (mb_p->rd_ptr ()) << endl;
      c_stream.close ();
      done = 1;
    }
  }
  ACE_OS::time (&currsec);
  par2= (int)currsec;

  secs=par2-par1;
  if (secs<=0)secs=1;
  cout << "consumer got " << cnt << " messages of size " << msiz 
       << "within " << secs << " seconds" << endl;

  sleep (2);
  cout << "consumer terminating " << endl;
  return 0;
}

static void *
supplier (void *dummy)
{
  ACE_UPIPE_Stream s_stream;
  ACE_UPIPE_Addr serv_addr ("/tmp/supupipe");
  ACE_UPIPE_Connector con;
  int iter = options.iterations ();
  int verb = options.verbose ();
  int msiz = options.message_size ();
  cout << "supplier starting connect" << endl;

  if (con.connect (s_stream, serv_addr) == -1)
    ACE_DEBUG ((LM_INFO, " (%t) connect failed\n"));

  cout << "supplier : we're connected" << endl;
  int	   n;
  n = 0;
  ACE_Message_Block * mb_p;

  while (n < iter)
    {
      mb_p = new ACE_Message_Block (msiz);
      strcpy (mb_p->rd_ptr (), (char *) dummy);
      mb_p->length (msiz);
      if (verb)
        cout << "supplier sending 1 message_block" << endl;
      if (s_stream.send_msg (mb_p) == -1)
	{
	  cout << "supplier send failed" << endl;
	  return (void *) -1;
	}
      n++;
    }
  mb_p = new ACE_Message_Block (10);
  mb_p->length (1);
  *mb_p->rd_ptr () = 'g';

  cout << "supplier sending last message_block" << endl;

  if (s_stream.send_msg (mb_p) == -1)
    {
      cout << "supplier send last mb failed" << endl;
      return (void *) -1;
    }
  mb_p = new ACE_Message_Block (10);
  mb_p->length (0);
  if (verb)
    cout << "supplier sending very last message_block" << endl;
  if (s_stream.send_msg (mb_p) == -1)
    {
      cout << "supplier send very last mb failed" << endl;
      return (void *) -1;
    }

  sleep (2);
  cout << "supplier terminating" << endl;
  return 0;
}

int
main (int argc, char *argv[])
{
  ACE_Service_Config daemon;
  
  options.parse_args (argc, argv);
  options.start_timer ();

  /* Primary ACE_Stream for EVENT_SERVER application */
  MT_Stream event_server; 

  /* Enable graceful shutdowns... */
  Quit_Handler quit_handler (0, event_server, thr_mgr);

  if (ACE_Service_Config::reactor ()->register_handler 
      (&quit_handler, ACE_Event_Handler::READ_MASK) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "register_handler"), -1);    

  /* Create the modules. */

  MT_Module *sr = new MT_Module ("Supplier_Router", 
				 new Supplier_Router (&thr_mgr));
  MT_Module *ea = new MT_Module ("Event_Analyzer", 
				 new Event_Analyzer, 
				 new Event_Analyzer);
  MT_Module *cr = new MT_Module ("Consumer_Router", 
				 0, // 0 triggers the creation of a ACE_Thru_Task...
				 new Consumer_Router (&thr_mgr));

  /* Push the modules onto the event_server stream */

  if (event_server.push (sr) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "push (Supplier_Router)"), -1);
					
  if (event_server.push (ea) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "push (Event_Analyzer)"), -1);

  if (event_server.push (cr) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "push (Consumer_Router)"), -1);

  /* Set the high and low water marks appropriately */

  int wm = options.low_water_mark ();

  if (event_server.control (ACE_IO_Cntl_Msg::SET_LWM, &wm) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "push (setting low watermark)"), -1);

  wm = options.high_water_mark ();
  if (event_server.control (ACE_IO_Cntl_Msg::SET_HWM, &wm) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "push (setting high watermark)"), -1);

  /* spawn the two threads */

  if (thr_mgr.spawn (ACE_THR_FUNC (consumer), (void *) 0,
		     THR_NEW_LWP | THR_DETACHED) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "spawn"), 1);

  else if (thr_mgr.spawn (ACE_THR_FUNC (supplier), (void *) "hello",
		     THR_NEW_LWP | THR_DETACHED) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "spawn"), 1);
 
  /* Perform the main event loop */

  while (!quit_handler.done ())
    ACE_Service_Config::reactor ()->handle_events ();

  cout << "main exiting" << endl;
}
#else
int 
main (void)
{
  ACE_ERROR_RETURN ((LM_ERROR, "test not defined for this platform\n"), -1);
}
#endif /* ACE_HAS_THREADS */
