// Illustrates how the ACE_Reactor's thread-safe event notification
// mechanism works.

#include "ace/Log_Msg.h"
#include "ace/Service_Config.h"
#include "ace/Thread.h"

#if defined (ACE_HAS_THREADS)

// Define a class that handles timeouts in the main thread via the
// ACE_Reactor and I/O events in a separate thread.  Just before the
// separate I/O thread exits it notifies the ACE_Reactor in the main
// thread using the ACE_Reactor's notification mechanism.

class Thread_Handler : public ACE_Event_Handler
{
public:
  static Thread_Handler *instance (void);
  
  int run (ACE_Time_Value &delay, 
	   ACE_Time_Value &interval,
	   size_t n_threads);

  virtual int handle_signal (ACE_HANDLE signum
#if defined (ACE_HAS_SIGINFO_T)
    , siginfo_t * = 0, ucontext_t * = 0
#endif /* ACE_HAS_SIGINFO_T */
    );
  // Handle signals.

  virtual int handle_input (ACE_HANDLE);

  virtual int handle_exception (ACE_HANDLE);
  // Print data from main thread.

  virtual int handle_output (ACE_HANDLE);
  // Print data from main thread.

  virtual int handle_timeout (const ACE_Time_Value &, 
			      const void *);
  // Handle timeout events in the main thread.

  virtual int svc (void);
  // Handle I/O events in a separate thread.

protected:
  Thread_Handler (void) {}
  Thread_Handler (size_t id): id_ (id) {}

private:
  static sig_atomic_t done_;
  // Used to indicate when we are finished.

  static Thread_Handler *instance_;
  // Singleton...

  size_t id_;
  // ID passed in by Thread_Handler constructor.

  ACE_Time_Value delay_;
  // Delay factor for timer-driven I/O.

  static void *svc_run (void *);
  // Glues C++ to C thread library functions.
};

// Static definitions
Thread_Handler *Thread_Handler::instance_ = 0;
sig_atomic_t Thread_Handler::done_ = 0;

Thread_Handler *
Thread_Handler::instance (void)
{
  if (Thread_Handler::instance_ == 0)
    Thread_Handler::instance_ = new Thread_Handler;

  return Thread_Handler::instance_;
}
// Test stdin handling.

int
Thread_Handler::handle_input (ACE_HANDLE handle)
{
  assert (handle == 0);
  char buf[BUFSIZ];
  ssize_t n;

  if ((n = read (handle, buf, sizeof buf)) > 0)
    {
      ACE_DEBUG ((LM_DEBUG, "(%t) %*s", n, buf));

      ACE_Event_Handler *eh = 0;

      // Just do something to test the ACE_Reactor's multi-thread
      // capabilities...

      if (ACE_Service_Config::reactor ()->handler 
	  (0, ACE_Event_Handler::READ_MASK, &eh) == -1 || eh != this)
	ACE_ERROR_RETURN ((LM_ERROR, "(%t) %p\n", "handler"), -1);

      if (ACE_Service_Config::reactor ()->notify 
	  (this, ACE_Event_Handler::EXCEPT_MASK) == -1)
	ACE_ERROR ((LM_ERROR, "(%t) %p\n", "notify"));
      else if (ACE_Service_Config::reactor ()->notify 
	       (this, ACE_Event_Handler::WRITE_MASK) == -1)
	ACE_ERROR ((LM_ERROR, "(%t) %p\n", "notify"));	
      return 0;
    }
  else
    return -1;
}

// Test signal handling.

int 
Thread_Handler::handle_signal (ACE_HANDLE signum
#if defined (ACE_HAS_SIGINFO_T)
    , siginfo_t *, ucontext_t *
#endif /* ACE_HAS_SIGINFO_T */
  )
{
  ACE_DEBUG ((LM_DEBUG, "(%t) received signal %S\n", signum));

  switch (signum)
    {
    case SIGINT:
    case SIGQUIT:
      ACE_ERROR ((LM_ERROR, "shutting down %n\n"));
      Thread_Handler::done_ = 1;
      /* NOTREACHED */
    }
  return 0;
}

int
Thread_Handler::run (ACE_Time_Value &delay,
		     ACE_Time_Value &interval,
		     size_t n_threads)
{
  ACE_Sig_Set sig_set;
  
  sig_set.sig_add (SIGQUIT);
  sig_set.sig_add (SIGINT);

  this->delay_ = delay;
  this->id_ = 0;

  // Register for stdin events...
  if (ACE_Service_Config::reactor ()->register_handler 
      (0, this, ACE_Event_Handler::READ_MASK) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "(%t) %p\n", "register_handler"), -1);

  if (ACE_Service_Config::reactor ()->register_handler (sig_set, this) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "(%t) %p\n", "register_handler"), -1);

  if (ACE_Service_Config::reactor ()->schedule_timer 
      (this, 0, delay, interval) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "(%t) %p\n", "schedule_timer"), -1);
   
  // Set up this thread's signal mask, which is inherited by the
  // threads it spawns.

  ACE_Thread::sigsetmask (SIG_BLOCK, sig_set);

  // Create N new threads of control running this object.
  for (size_t i = 0; i < n_threads; i++)
    {
      if (ACE_Thread::spawn (&Thread_Handler::svc_run, 
			     new Thread_Handler (i + 1),
			     THR_NEW_LWP | THR_DETACHED) != 0)
	ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "ACE_Thread::spawn"), -1);
    }

  // Unblock signal set so that only this thread receives them!
  ACE_Thread::sigsetmask (SIG_UNBLOCK, sig_set);

  while (!Thread_Handler::done_)
    ACE_Service_Config::reactor ()->handle_events ();

  ACE_OS::exit (1);
  /* NOTREACHED */
  return 0;
}

int 
Thread_Handler::handle_timeout (const ACE_Time_Value &, 
				const void *)
{
  ACE_DEBUG ((LM_DEBUG, "(%t) received timeout\n"));
  return 0;
}

// Perform a task that will test the ACE_Reactor's multi-threading
// capabilities in separate threads.

int
Thread_Handler::svc (void)
{
  for (;;)
    {
      if (this->delay_.sec () > 0)
	// Block for (delay.secs () * 1000) milliseconds, then notify
	// ACE_Reactor.
	::poll (0, 0, this->delay_.sec () * 1000);

      ACE_Event_Handler *eh = 0;

      // Just do something to test the ACE_Reactor's multi-thread capabilities...
      if (ACE_Service_Config::reactor ()->handler 
	  (0, ACE_Event_Handler::READ_MASK, &eh) == -1)
	ACE_ERROR_RETURN ((LM_ERROR, "(%t) %p\n", "handler"), -1);

      if (ACE_Service_Config::reactor ()->notify 
	  (this, ACE_Event_Handler::EXCEPT_MASK) == -1)
	ACE_ERROR_RETURN ((LM_ERROR, "(%t) %p\n", 
			   "notify"), -1);
      else if (ACE_Service_Config::reactor ()->notify 
	       (this, ACE_Event_Handler::WRITE_MASK) == -1)
	ACE_ERROR_RETURN ((LM_ERROR, "(%t) %p\n", "notify"), -1);
    }
  return 0;
}

// Called by the ACE_Reactor when it receives a notification.

int
Thread_Handler::handle_exception (ACE_HANDLE)
{
  ACE_DEBUG ((LM_DEBUG, 
	      "(%t) handle_exception received notification from id %d\n", 
	      this->id_));
  return 0;
}

// Called by the ACE_Reactor when it receives a notification.

int
Thread_Handler::handle_output (ACE_HANDLE)
{
  ACE_DEBUG ((LM_DEBUG, 
	      "(%t) handle_output received notification from id %d\n", 
	      this->id_));
  return 0;
}

// "Shim" function that integrates C thread API with C++.

void *
Thread_Handler::svc_run (void *eh)
{
  Thread_Handler *this_handler = (Thread_Handler *) eh;

  return this_handler->svc () == 0 ? 0 : (void *) -1;
}

int 
main (int argc, char *argv[])
{
  // Required to initialize the Service Configurator.
  ACE_Service_Config daemon (argv[0]);

  if (argc != 4)
    ACE_ERROR_RETURN ((LM_ERROR, "usage: %s delay interval n_threads\n", 
		       argv[0]), 1);

  ACE_Time_Value delay    = ACE_OS::atoi (argv[1]);
  ACE_Time_Value interval = ACE_OS::atoi (argv[2]);
  size_t n_threads    = ACE_OS::atoi (argv[3]);

  if (Thread_Handler::instance ()->run (delay, interval, n_threads) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "run"), -1);

  return 0;
}
/* End of Program */
#else
int 
main (int, char *[])
{
  ACE_ERROR_RETURN ((LM_ERROR, 
		     "threads not supported on this platform\n"), -1);
}
#endif /* ACE_HAS_THREADS */
