/* Simple program that illustrates all the features of the ACE_Reactor:

   1. I/O event demultiplexing
   2. Signal-based demultiplexing
   3. Timer-based demultiplexing

   To test this program, compile it and then execute it as follows:

   % ./pingpong hello

   You should see lots of the following output:
   
   writing <4> [7860]
   writing <4> [7860]
   writing <4> [7860]
   writing <4> [7860]
   reading <5> (7860) [1] = hello
   writing <4> [7860]
   writing <5> [7861]
   reading <4> (7861) [2] = hello
   reading <5> (7860) [2] = hello
   writing <4> [7860]
   writing <5> [7861]
   reading <4> (7861) [3] = hello
   reading <5> (7860) [3] = hello

   After 10 seconds you'll see the following:

   ./pingpong: shutting down tester (pid = 7861)
   ./pingpong: shutting down tester (pid = 7860)

   and the program will stop.  If you'd like to
   stop it earlier, just hit the control-C sequence
   and you'll see the same messages. */

#include "ace/Log_Msg.h"
#include "ace/Reactor.h"

class Ping_Pong : public ACE_Event_Handler
{
public:
  Ping_Pong (char b[], ACE_HANDLE f);
  virtual ACE_HANDLE get_handle (void) const;
  virtual int handle_input (ACE_HANDLE);
  virtual int handle_output (ACE_HANDLE);
  virtual int handle_signal (ACE_HANDLE signum
#if defined (ACE_HAS_SIGINFO_T)
, siginfo_t * = 0, ucontext_t * = 0
#endif /* ACE_HAS_SIGINFO_T */
);
  virtual int handle_timeout (const ACE_Time_Value &, const void *);
  int         done (void);
  void	      done (int);

private:
  char   buf[BUFSIZ];
  size_t buflen;
  pid_t  pid;
  ACE_HANDLE fd_;
  sig_atomic_t finished; 
  // Flag that indicates when the program should terminate
};

Ping_Pong::Ping_Pong (char b[], ACE_HANDLE f)
		     : fd_ (f),
		       buflen (ACE_OS::strlen (b) + 1 + (2 * sizeof (int))),
		       pid (ACE_OS::getpid ()),
		       finished (0)
{
  *((pid_t *) this->buf) = this->pid;
  *((int *) (this->buf + sizeof (int))) = 0;
  ACE_OS::strcpy (this->buf + (2 * sizeof (int)), b); 
  this->buf[this->buflen - 1] = '\n';
  this->buf[this->buflen] = '\0';
}

inline int
Ping_Pong::done (void)
{
  return this->finished;
}

inline void
Ping_Pong::done (int d)
{
  this->finished = d;
}

ACE_HANDLE
Ping_Pong::get_handle (void) const
{
  return this->fd_;
}

int 
Ping_Pong::handle_input (ACE_HANDLE)
{
  if (ACE_OS::read (this->fd_, this->buf, sizeof this->buf) == -1)
    ACE_ERROR_RETURN ((LM_ERROR, "[%d] %p\n", fd_, "read"), -1);
  ACE_DEBUG ((LM_DEBUG, "reading <%d> (%d) [%d] = %s",
	      this->fd_, *(int *) buf, 
	      *(int *) (buf + sizeof (int)), 
	      buf + (2 * sizeof (int))));
  return 0;
}

int 
Ping_Pong::handle_output (ACE_HANDLE)
{
  (*(pid_t *) (this->buf)) = this->pid;
  (*(int *) (this->buf + sizeof (int)))++;
  if (ACE_OS::write (this->fd_, this->buf, this->buflen) == -1)
    return -1;
  else
    {
      ACE_DEBUG ((LM_DEBUG, 
		  "writing <%d> [%d]\n", 
		  this->fd_, this->pid));
      return 0;
    }
}

int
Ping_Pong::handle_signal (ACE_HANDLE
#if defined (ACE_HAS_SIGINFO_T)
, siginfo_t *, ucontext_t *
#endif /* ACE_HAS_SIGINFO_T */
)
{
  this->done (1);
  return 0;
}

int 
Ping_Pong::handle_timeout (const ACE_Time_Value &, 
			   const void *)
{
  this->done (1);
  return 0;
}

static void
init_handles (int fds[])
{
#if defined (ACE_HAS_STREAM_PIPES)
  if (ACE_OS::pipe (fds) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n%a", "pipe", 1));

  int arg = RMSGN;

  if (ACE_OS::ioctl (fds[0], I_SRDOPT, (void *) arg) == -1
      || ACE_OS::ioctl (fds[1], I_SRDOPT, (void *) arg) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n%a", "ioctl", 1));
#else
  if (ACE_OS::socketpair (AF_UNIX, SOCK_DGRAM, 0, fds) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n%a", "socketpair", 1));
#endif /* ACE_HAS_STREAM_PIPES */
}

int 
main (int argc, char *argv[])
{
  ACE_LOG_MSG->open (argv[0]);

  ACE_Reactor reactor;
  pid_t pid;
  int fds[2];

  if (argc != 2)
    ACE_ERROR ((LM_ERROR, "usage: %n string\n%a", 1));

  init_handles (fds);

  if ((pid = ACE_OS::fork ()) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n%a", "fork", 1));

  // Resync the ACE_Log_Msg notion of process id and program name
  ACE_LOG_MSG->sync (argv[0]); 

  Ping_Pong callback (argv[1], fds[pid == 0]);

  // Register the callback object for the various I/O, signal, and
  // timer-based events.

  if (reactor.register_handler (&callback,
				ACE_Event_Handler::READ_MASK 
				| ACE_Event_Handler::WRITE_MASK) == -1
      || reactor.register_handler (SIGINT, &callback) == -1
      || reactor.schedule_timer (&callback, 0, 10) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n%a", "reactor", 1));

  // Main event loop (one per process).

  while (callback.done () == 0)
    reactor.handle_events ();

  ACE_DEBUG ((LM_DEBUG, "%n: shutting down tester (pid = %P)\n%a", 0));
  return 0;
}
