// Perform an extensive test of the ACE_Reactor's event dispatching
// mechanisms.  These mechanisms illustrate how signals, I/O, and
// timeout events can all be handled within the same framework.  In
// addition, this example illustrates how to use the ACE_Reactor for
// devices that perform I/O via signals (such as SVR4 message queues).

#include "ace/Log_Msg.h"
#include "ace/Event_Handler.h"
#include "ace/Event_Handler_T.h"
#include "ace/Reactor.h"

// Used to shut down the event loop.
static sig_atomic_t done = 0;

// This class illustrates how to handle signal-driven I/O using the
// ACE_Reactor framework.  Note that signals may be caught and
// processed without requiring the use of global signal handler
// functions or global signal handler data.

class Sig_Handler : public ACE_Event_Handler
{
public:
  Sig_Handler (ACE_Reactor &);
  virtual ACE_HANDLE get_handle (void) const;
  virtual int handle_output (ACE_HANDLE);
  virtual int shutdown (ACE_HANDLE, ACE_Reactor_Mask);
  virtual int handle_signal (ACE_HANDLE signum
#if defined (ACE_HAS_SIGINFO_T)
    , siginfo_t * = 0, ucontext_t * = 0
#endif /* ACE_HAS_SIGINFO_T */
    );

private:
  ACE_Reactor &reactor_;
  int handle_;
};

// A dummy_handle is required to reserve a slot in the ACE_Reactor's
// descriptor table.

Sig_Handler::Sig_Handler (ACE_Reactor &reactor)
  : reactor_ (reactor)
{
  // Assign the Sig_Handler a dummy I/O descriptor.
  // Note that even though we open this file "Write Only"
  // that we need to use the ACE_Event_Handler::NULL_MASK when
  // registering this with the ACE_Reactor (see below).
  this->handle_ = ACE_OS::open ("/dev/null", O_WRONLY);
  assert (this->handle_ != -1);
}

// Called by the ACE_Reactor to extract the fd.

ACE_HANDLE
Sig_Handler::get_handle (void) const
{
  return this->handle_;
}

// In a real application, this method would be where the read on the
// signal-driven I/O device would occur asynchronously.  For now we'll
// just print a greeting to let you know that everything is working
// properly!

int 
Sig_Handler::handle_output (ACE_HANDLE)
{
  ACE_DEBUG ((LM_DEBUG, "handling asynchonrous input...\n"));
  return 0;
}

// In a real application, this method would do any cleanup activities
// required when shutting down the I/O device.

int 
Sig_Handler::shutdown (ACE_HANDLE, ACE_Reactor_Mask)
{
  ACE_DEBUG ((LM_DEBUG, "closing down Sig_Handler...\n"));
  return 0;
}

// This method handles all the signals that are being caught by this
// object.  In our simple example, we are simply catching SIGALRM,
// SIGINT, and SIGQUIT.  Anything else is logged and ignored.
//
// There are several advantages to using this approach.  First, 
// the behavior triggered by the signal is handled in the main event
// loop, rather than in the signal handler.  Second, the ACE_Reactor's 
// signal handling mechanism eliminates the need to use global signal 
// handler functions and data. 

int
Sig_Handler::handle_signal (int signum
#if defined (ACE_HAS_SIGINFO_T)
, siginfo_t *, ucontext_t *
#endif /* ACE_HAS_SIGINFO_T */
)
{
  ACE_DEBUG ((LM_DEBUG, "received signal %d\n", signum));
  switch (signum)
    {
    case SIGALRM:
      // Rearm the alarm.
      ACE_OS::alarm (4);
      break;
    case SIGINT:
      // Tell the ACE_Reactor to enable the ready bit for this->handle_.
      // The ACE_Reactor will subsequently call the Sig_Handler::handle_input
      // method from within its event loop.
      return this->reactor_.ready_ops (this->handle_, 
				       ACE_Event_Handler::READ_MASK, 
				       ACE_Reactor::ADD_MASK);
    case SIGQUIT:
      ACE_DEBUG ((LM_DEBUG, "%S: shutting down signal tester\n", signum));
      done = 1;
      break;
    default: 
      ACE_DEBUG ((LM_DEBUG, "%S: not understood, returning to program\n", signum));
      break;
    }
  return 0;
}

// This class illustrates that the ACE_Reactor can handle signals, I/O,
// and timeouts using the same mechanisms.

class IO_Handler : public ACE_Event_Handler
{
public:
  IO_Handler (int handle);
  virtual ACE_HANDLE get_handle (void) const;
  virtual int handle_input (ACE_HANDLE);
  virtual int handle_timeout (const ACE_Time_Value &, const void *arg);

private:
  int handle_;
};

IO_Handler::IO_Handler (int handle)
  : handle_ (handle) 
{
}

int 
IO_Handler::handle_timeout (const ACE_Time_Value &tv, const void *arg)
{
  ACE_DEBUG ((LM_DEBUG, "timeout occurred at %d sec, %d usec\n",
	      tv.sec (), tv.usec ()));
  return 0;
}

// Called by the ACE_Reactor to extract the underlying I/O descriptor.

ACE_HANDLE
IO_Handler::get_handle (void) const
{
  return this->handle_;
}

// Read from input descriptor and write to stdout descriptor.

int 
IO_Handler::handle_input (ACE_HANDLE)
{
  ssize_t n;
  char    buf[BUFSIZ];

  switch (n = ACE_OS::read (this->handle_, buf, sizeof buf))
    {
    case -1:
      if (errno == EINTR)
	return 0;
        /* NOTREACHED */
      else
	ACE_ERROR ((LM_ERROR, "%p\n", "read"));
      /* FALLTHROUGH */
    case 0:
      done = 1;
      break;
    default:
      {
	int result = result = ACE_OS::write (1, buf, n);

	if (result != n)
	  ACE_ERROR_RETURN ((LM_ERROR, "%p\n", "write"), 
			   result == -1 && errno == EINTR ? 0 : -1);
      }
    }
  return 0;
}

int
main (int argc, char *argv[])
{
  ACE_LOG_MSG->open (argv[0]);
  ACE_DEBUG ((LM_DEBUG, "starting up program %n (pid = %P)\n"));
  
  // Event demultiplexer.
  ACE_Reactor reactor;

  // Tie the ACE_Event_Handler_T together with the methods from ACE_Sig_Handler.
  Sig_Handler sh (reactor);

  // Define an I/O handler object (passing it the stdin descriptor == 0).
  IO_Handler ioh (0);

  // Register I/O handler object.
  if (reactor.register_handler (&ioh, ACE_Event_Handler::READ_MASK) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n%a", "register_handler", 1));

  // Register the IO_Handler to be dispatched every second.
  if (reactor.schedule_timer (&ioh, 0, ACE_Time_Value (1), ACE_Time_Value (1)) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n%a", "schedule_timer", 1));

  // Register signal handler object.  Note that NULL_MASK is used
  // to keep the ACE_Reactor from calling us back on the "/dev/null"
  // descriptor.
  if (reactor.register_handler (&sh, ACE_Event_Handler::NULL_MASK) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n%a", "register_handler", 1));

  // Create a sigset_t corresponding to the signals we want to catch.
  ACE_Sig_Set sig_set;

  sig_set.sig_add (SIGINT);
  sig_set.sig_add (SIGQUIT);
  sig_set.sig_add (SIGALRM);  

  // Register the signal handler object to catch the signals.
  if (reactor.register_handler (sig_set, &sh) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n%a", "register_handler", 1));

  // Optionally start the alarm.
  if (argc > 1)
    ACE_OS::alarm (4);

  // Loop forever handling signals and I/O events.

  while (done == 0)
    reactor.handle_events ();

  return 0;
}
