// Test the persistence capabilities of the ACE shared memory manager.

#include "ace/Log_Msg.h"
#include "ace/Malloc.h"

typedef ACE_Malloc <ACE_MMAP_Memory_Pool, ACE_Null_Mutex> MALLOC;

// Shared memory manager.
static MALLOC shmem_manager ("/tmp/malloc-test");

class Employee
{
public:
  Employee (void): name_ (0), id_ (0) {}

  Employee (char* name, u_long id) : id_ (id)
    {
      this->name_ = (char*) shmem_manager.malloc (ACE_OS::strlen (name) + 1);
      ACE_OS::strcpy (this->name_, name );
    }

  ~Employee (void)
    {
      shmem_manager.free (this->name_);      
    }

  char *name (void) const
    {
      return this->name_;
    }

  void name (char* name)
    {
      if (this->name_)
	shmem_manager.free (this->name_);
      this->name_ = (char*) shmem_manager.malloc (ACE_OS::strlen (name) + 1);
      ACE_OS::strcpy (this->name_, name);
    }

  u_long id (void) const
    {
      return id_;
    }

  void id (u_long id)
    {
      id_ = id;
    }
  
  friend ostream &operator<<(ostream &stream, const Employee &employee)
    {
      stream << endl;
      stream << "Employee name: " << employee.name() << endl;
      stream << "Employee id:   " << employee.id()   << endl;
      stream << endl;

      return stream;
    }
  
  void *operator new (size_t size)
    {
      return shmem_manager.malloc (sizeof (Employee));
    }

  void operator delete (void *pointer)
    {
      shmem_manager.free (pointer);
    }

private:
  char *name_;
  // Employee name.

  u_long id_;
  // Employee ID.
};

class GUI_Handler 
{
public:
  GUI_Handler (void) { menu(); }

  ~GUI_Handler (void) 
    { 
      MALLOC::MEMORY_POOL &pool = shmem_manager.memory_pool();
      pool.sync ();
    }

  int service(void)
    {
      char option[BUFSIZ];
      char buf1[BUFSIZ];
      char buf2[BUFSIZ];

      if (::scanf ("%s", option) <= 0)
	{
	  cerr << "try again" << endl;
	  return 0;
	}
      
      int result = 0;
      switch (option[0])
	{
	case 'I' :
	case 'i' :
	  if (::scanf ("%s %s", buf1, buf2) <= 0)
	    break;
	  result = insert_employee (buf1, ACE_OS::atoi (buf2));
	  break;
	case 'F' :
	case 'f' :
	  if (::scanf ("%s", buf1) <= 0)
	    break;
	  result = find_employee (buf1);
	  break;
	case 'D' :
	case 'd' :
	  if (::scanf ("%s", buf1) <= 0)
	    break;
	  result = delete_employee (buf1);
	  break;
	case 'L' :
	case 'l' :
	  result = list_employees ();
	  break;
	case 'Q' :
	case 'q' :
	  return -1;
	  break;
	  default :
	  cout << "unrecognized command" << endl;
	}
      if (result == 0)
	cout << "Last operation was successful!!" << endl;
      else
	cout << "Last operation failed!! " << endl;
      
      menu ();
      
      return 0;
    }

  void menu(void)
    {
      cout << endl;
      cout << "\t**************************                                  " << endl;
      cout << "\tThe employee database menu                                  " << endl;
      cout << endl;
      cout << "\t<I> Insert    <name>   <id>                                 " << endl;
      cout << "\t<D> Delete    <name>                                        " << endl;
      cout << "\t<F> Find      <name>                                        " << endl;
      cout << endl;
      cout << "\t<L> List all employees                                      " << endl;
      cout << endl;
      cout << "\t<Q> Quit " << endl;
      cout << "\t**************************                                  " << endl;
    }

private:
  int insert_employee (char* name, u_long id)
    {
      if (find_employee (name) == 0)
	
	{
	  cerr << "Employee already exists" << endl << endl;
	  return -1;
	}

      Employee* new_employee = new Employee (name, id);
      shmem_manager.bind (name, new_employee);
      return 0;
    }

  int find_employee (char* name)
    {
      void *temp;
      if (shmem_manager.find (name, temp) == 0)
	{
	  Employee* employee = (Employee*) temp;
	  
	  cerr << "The following employee was found......." << endl << endl;
	  cerr << *employee << endl;
	  return 0;
	}

      return -1;
    }

  int list_employees (void)
    {
      ACE_Malloc_Iterator <ACE_MMAP_Memory_Pool, ACE_Null_Mutex> iterator (shmem_manager);
      
      cerr << "The following employees were found......." << endl << endl;
      
      for( void* temp = 0;
	  iterator.next( temp ) != 0; 
	  iterator.advance() )
	{
	  Employee* employee = (Employee*) temp;
	  cerr << *employee << endl;
	}

      return 0;

    }

  int delete_employee (char* name)
    {
      void *temp;
      if (shmem_manager.unbind (name, temp) == 0)
	{
	  Employee* employee = (Employee*) temp;
	  
	  cerr << "The following employee was found and deleted......." << endl << endl;
	  cerr << *employee << endl;

	  delete employee;
	  return 0;
	}

      cerr << "There is no employee with name " << name << endl;
      return -1;
    }      
};

void
parse_args (int argc, char** argv)
{
}  

int 
main (int argc, char** argv)
{
  parse_args (argc, argv);
  
  GUI_Handler handler;

  for(;;)
    if (handler.service() == -1)
      {
	cerr << "closing down ...." << endl;
	break;
      }
  
  return 0;
}


