#include "ace/Thread.h"
#include "ace/Synch.h"
#include "ace/Log_Msg.h"

// Define a class that will be stored in thread-specific data.  Note
// that as far as this class is concerned it's just a regular C++
// class.  The ACE_TSS wrapper transparently ensures that
// objects of this class will be placed in thread-specific storage.
// All calls on ACE_TSS::operator->() are delegated to the
// appropriate method in the Errno class.

class Errno
{
public:
  int error (void) { return this->errno_; }
  void error (int i) { this->errno_ = i; }
  
  int line (void) { return this->lineno_; }
  void line (int l) { this->lineno_ = l; }

  // Errno::flags_ is a static variable, so we've got to protect it
  // with a mutex since it isn't kept in thread-specific storage.
  int flags (void) { 
    ACE_MT (ACE_Guard<ACE_Mutex> m (Errno::lock_));
    return Errno::flags_;
  }
  void flags (int f)
  {
    ACE_MT (ACE_Guard<ACE_Mutex> m (Errno::lock_));
    Errno::flags_ = f;
  }

private:
  // = errno_ and lineno_ will be thread-specific data so they don't
  // need a lock.  
  int errno_;
  int lineno_;

  static int flags_;
#if defined (ACE_HAS_THREADS)
  // flags_ needs a lock. 
  static ACE_Mutex lock_;
#endif /* ACE_HAS_THREADS */
};

// Static variables.
ACE_MT (ACE_Mutex Errno::lock_);
int Errno::flags_;

// This is our thread-specific error handler...
static ACE_TSS<Errno> TSS_Error;

#if defined (ACE_HAS_THREADS)
// Serializes output via cout.
static ACE_Mutex lock;

typedef ACE_TSS_Guard<ACE_Mutex> GUARD;
#else
// Serializes output via cout.
static ACE_Null_Mutex lock;

typedef ACE_Guard<ACE_Null_Mutex> GUARD;
#endif /* ACE_HAS_THREADS */

// This worker function is the entry point for each thread.

static void *
worker (void *c)
{
  int count = int (c);

  for (int i = 0; i < count; i++)
    {
      // Cause an error.
      ACE_OS::read (-1, 0, 0);

      // The following two lines set the thread-specific state.
      TSS_Error->error (errno);
      TSS_Error->line (__LINE__);

      // This sets the static state (note how C++ makes it easy to do
      // both).
      TSS_Error->flags (ACE_Thread::self ());

      // Use the guard to serialize access to cout...
      GUARD m (lock);

      cout << "thread = " << ACE_Thread::self ()
           << "errno = "  << TSS_Error->error () 
	   << ", lineno = " << TSS_Error->line ()
	   << ", flags = " << TSS_Error->flags () 
	   << endl;
    }
  return 0;
}

static void 
handler (int signum)
{
  ACE_DEBUG ((LM_DEBUG, "signal = %S\n", signum));
  ACE_Thread::exit (0);
}

int 
main (int argc, char *argv[])
{
  int threads = argc > 1 ? ACE_OS::atoi (argv[1]) : 4;
  int count = argc > 2 ? ACE_OS::atoi (argv[2]) : 10000;

  signal (SIGINT, ACE_SignalHandler (handler));
  
#if defined (ACE_HAS_THREADS)
  if (ACE_Thread::spawn_n (threads, ACE_THR_FUNC (&worker), 
			   (void *) count,
			   THR_BOUND | THR_DETACHED) != threads)
    ACE_OS::perror ("thr_create");

  ACE_Thread::exit (0);
#else
  worker ((void *) count);
#endif /* ACE_HAS_THREADS */
  return 0;
}

