// Synch_Test.C

#include "ace/ACE.h"
#include "ace/Synch.h"
#include "ace/Thread.h"
#include "ace/SOCK_Acceptor.h"

// Define a class that will be stored in thread-specific data.  Note
// that as far as this class is concerned it's just a regular C++
// class.  The ACE_TSS wrapper transparently ensures that objects of
// this class will be placed in thread-specific storage.  All calls on
// ACE_TSS::operator->() are delegated to the appropriate method in
// the Errno class.

class Errno
{
public:
 int error (void) { return this->errno_; }
 void error (int i) { this->errno_ = i; }
 
 int line (void) { return this->lineno_; }
 void line (int l) { this->lineno_ = l; }
 
 ~Errno() { cerr << "Destroying Errno." << endl; }
 
 // Errno::flags_ is a static variable, so we've got to protect it
 // with a mutex since it isn't kept in thread-specific storage.
 int flags (void) { 
		    ACE_MT (ACE_Guard<ACE_Mutex> m (Errno::lock_));
		    return Errno::flags_;
		  }
 void flags (int f)
   {
    ACE_MT (ACE_Guard<ACE_Mutex> m (Errno::lock_));
    Errno::flags_ = f;
  }
 
private:
 // = errno_ and lineno_ will be thread-specific data so they don't
 // need a lock.  
 int errno_;
 int lineno_;
 
 static int flags_;
#if defined (ACE_HAS_THREADS)
 // flags_ needs a lock. 
 static ACE_Mutex lock_;
#endif /* ACE_HAS_THREADS */
};

// Static variables.
ACE_MT (ACE_Mutex Errno::lock_);
int Errno::flags_;

// This is our thread-specific error handler...
ACE_TSS<Errno> TSS_Error;

// This worker function is the entry point for each thread.

static void *
worker (void *c)
{
 int count = int (c);
 
 for (int i = 0; i < count; i++)
   {
    // Cause an error.
    ::read (-1, 0, 0);
    // The following two lines set the thread-specific state.
    TSS_Error->error (errno);
    TSS_Error->line (__LINE__);
    
    // This sets the static state (note how C++ makes it easy to do
    // both).
    TSS_Error->flags (ACE_Thread::self ());
    
    ::fprintf (stderr, "thread = %d, errno = %d, lineno = %d, flags = %d\n",
	       ACE_Thread::self (), TSS_Error->error (), TSS_Error->line (), TSS_Error->flags ());
    
    ACE_Thread::yield ();
  }
 
 return 0;
}

void 
winsock_init (void)
{
  WORD    wVersionRequested;
  WSADATA wsaData;
  int     err;

  wVersionRequested = MAKEWORD (1, 1);

  if ((err = ::WSAStartup (wVersionRequested, &wsaData)) != 0)
    {
      (void) ACE_OS::fprintf (stderr, "WSAStartup failed. WSAGetLastError returned %u.\n", err);
      ACE_OS::exit (0);
    }
}

int 
main (int argc, char *argv[])
{
  winsock_init ();

  u_short port_num = argc > 1 ? ACE_OS::atoi (argv[1]) : 12345;

  ACE_SOCK_Stream cli_stream;
  ACE_INET_Addr addr (port_num);
  ACE_INET_Addr cli_addr;

  ACE_SOCK_Acceptor acceptor (addr);

  while (acceptor.accept (cli_stream, &cli_addr) != -1)
    {
      // Send data to server (correctly handles "incomplete writes").
  
      ::printf ("connected from %s\n", cli_addr.get_host_name ());

		      
      for (;;)
	{
	  char buf[BUFSIZ];

	  ssize_t r_bytes = cli_stream.recv (buf, sizeof buf, 0);
      
	  if (r_bytes == -1)
	    ACE_OS::perror ("recv"), ACE_OS::exit (1);
	  else if (r_bytes == 0)
	    {
	      cli_stream.close ();
	      break;
	    }
	  else
	  {
		  buf[r_bytes] = '\0';
	    fputs (buf, stdout);
	  }
	}
    }

  return 0;
}
